/*
 *   IRC - Internet Relay Chat, src/channel.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: channel.c,v 1.16 2002/01/17 16:04:43 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "h.h"

aChannel *channel = NullChn;

static void sub1_from_channel(aChannel *);

void clean_channelname(unsigned char *);

/*
 * some buffers for rebuilding channel/nick lists with ,'s 
 */
static char nickbuf[BUFSIZE], buf[BUFSIZE];

/*
 * htm ... 
 */
extern int lifesux;

/*
 * externally defined function 
 */
extern Link *find_channel_link(Link *, aChannel *);	/* defined in list.c */

/*
 * return the length (>=0) of a chain of links. 
 */
static int list_length(Link *lp)
{
    int count = 0;

    for (; lp; lp = lp->next)
	count++;
    return count;
}

/*
 * find_chasing 
 *   Find the client structure for a nick name (user) using history 
 *   mechanism if necessary. If the client is not found, an error message 
 *   (NO SUCH NICK) is generated. If the client was found through the 
 *   history, chasing will be 1 and otherwise 0.
 */
aClient *find_chasing(aClient *sptr, char *user, int *chasing)
{
    aClient *who = find_client(user, (aClient *) NULL);

    if (chasing)
	*chasing = 0;
    if (who)
	return who;
    if (!(who = get_history(user, (long) KILLCHASETIMELIMIT))) {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, user);
	return ((aClient *) NULL);
    }
    if (chasing)
	*chasing = 1;
    return who;
}

/*
 * Fixes a string so that the first white space found becomes an end of
 * string marker (`\-`).  returns the 'fixed' string or "*" if the
 * string was NULL length or a NULL pointer.
 */
static char *check_string(char *s)
{
    static char star[2] = "*";
    char *str = s;

    if (BadPtr(s))
	return star;

    for (; *s; s++)
	if (isspace(*s)) {
	    *s = '\0';
	    break;
	}

    return (BadPtr(str)) ? star : str;
}

/*
 * create a string of form "foo!bar@fubar" given foo, bar and fubar as
 * the parameters.  If NULL, they become "*".
 */
static char *make_nick_user_host(char *nick, char *name, char *host)
{
    static char namebuf[NICKLEN + USERLEN + HOSTLEN + 6];
    int n;
    char *ptr1, *ptr2;

    ptr1 = namebuf;
    for (ptr2 = check_string(nick), n = NICKLEN; *ptr2 && n--;)
	*ptr1++ = *ptr2++;
    *ptr1++ = '!';
    for (ptr2 = check_string(name), n = USERLEN; *ptr2 && n--;)
	*ptr1++ = *ptr2++;
    *ptr1++ = '@';
    for (ptr2 = check_string(host), n = HOSTLEN; *ptr2 && n--;)
	*ptr1++ = *ptr2++;
    *ptr1 = '\0';
    return (namebuf);
}

/*
 * adds a user to a channel by adding another link to the channels
 * member chain.
 */
void add_user_to_channel(aChannel *chptr, aClient *who, int flags)
{
    Link *ptr;
    chanMember *cm;

    if (who->user) {
	cm = make_chanmember();
	cm->flags = flags;
	cm->cptr = who;
	cm->next = chptr->members;
	cm->bans = 0;
	chptr->members = cm;
	chptr->users++;

	ptr = make_link();
	ptr->value.chptr = chptr;
	ptr->next = who->user->channel;
	who->user->channel = ptr;
	who->user->joined++;
    }
}

void remove_user_from_channel(aClient *sptr, aChannel *chptr)
{
    chanMember **curr, *tmp;
    Link **lcurr, *ltmp;

    for (curr = &chptr->members; (tmp = *curr); curr = &tmp->next)
	if (tmp->cptr == sptr) {
	    *curr = tmp->next;
	    free_chanmember(tmp);
	    break;
	}

    for (lcurr = &sptr->user->channel; (ltmp = *lcurr); lcurr = &ltmp->next)
	if (ltmp->value.chptr == chptr) {
	    *lcurr = ltmp->next;
	    free_link(ltmp);
	    break;
	}
    sptr->user->joined--;
    sub1_from_channel(chptr);
}

int is_chan_owner(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;

    if (chptr)
	if ((cm = find_user_member(chptr->members, cptr)))
	    return (cm->flags & CHFL_OWNER);

    return 0;
}

int is_protected(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;

    if (chptr)
	if ((cm = find_user_member(chptr->members, cptr)))
	    return (cm->flags & CHFL_PROTECT);

    return 0;
}

int is_chan_op(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;

    if (chptr)
	if ((cm = find_user_member(chptr->members, cptr)))
	    return (cm->flags & CHFL_CHANOP);

    return 0;
}

int is_deopped(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;

    if (chptr)
	if ((cm = find_user_member(chptr->members, cptr)))
	    return (cm->flags & CHFL_DEOPPED);

    return 0;
}

int has_voice(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;

    if (chptr)
	if ((cm = find_user_member(chptr->members, cptr)))
	    return (cm->flags & CHFL_VOICE);

    return 0;
}

int can_send(aClient *cptr, aChannel *chptr, char *msg)
{
    chanMember *cm;
    int member;

    if (IsServer(cptr) || IsULine(cptr))
	return 0;
    if (IsChanAnon(chptr))
	return (MODE_MODERATED);
    if (IsAnOper(cptr))
	return 0;

    member = (cm = find_user_member(chptr->members, cptr)) ? 1 : 0;

    if (!member) {
	if (chptr->mode.mode & MODE_MODERATED)
	    return (MODE_MODERATED);
	if (chptr->mode.mode & MODE_NOPRIVMSGS)
	    return (MODE_NOPRIVMSGS);
	if (chptr->mode.mode & MODE_REGONLY && !IsRegNick(cptr))
	    return (ERR_NEEDREGGEDNICK);
	if ((chptr->mode.mode & MODE_NOCOLOR) && msg_has_colors(msg))
	    return (ERR_NOCOLORSONCHAN);
	if (MyClient(cptr) && is_banned(cptr, chptr)) {
	    if (has_banex(cptr, chptr))
		return 0;
	    else
		return (MODE_BAN);
	}
    } else {
	if (chptr->mode.mode & MODE_MODERATED && !(cm->flags & CHFL_SPEAKABLE))
	    return (MODE_MODERATED);
	if (cm->bans && !(cm->flags & CHFL_SPEAKABLE))
	    return (MODE_BAN);
	if (chptr->mode.mode & MODE_REGONLY && !IsRegNick(cptr)
	    && !(cm->flags & CHFL_SPEAKABLE))
	    return (ERR_NEEDREGGEDNICK);
	if (chptr->mode.mode & MODE_NOCOLOR && msg_has_colors(msg)
	    && !(cm->flags & CHFL_SPEAKABLE))
	    return (ERR_NOCOLORSONCHAN);
    }
    return 0;
}

/*
 * * Remove bells and commas from channel name
 */

void clean_channelname(unsigned char *cn)
{
    for (; *cn; cn++)
	/*
	 * All characters >33 are allowed, except commas, and the weird
	 * fake-space character mIRCers whine about -wd
	 */
	if (*cn < 33 || *cn == ',' || (*cn == 160)) {
	    *cn = '\0';
	    return;
	}
    return;
}

/*
 * we also tell the client if the channel is invalid. 
 */
int check_channelname(aClient *cptr, unsigned char *cn)
{

    aConfItem *aconf;

    if (!MyClient(cptr))
	return 1;

    /*
     * We bypass Jupiter Checks, if the client is an operator
     * * I personally accepted this idea, because, Ircops can also
     * * use quarantined nicknames, so they ought to be able to join
     * * jupitered channels
     */

    if (!IsAnOper(cptr)) {
	aconf = find_conf_name(cn, CONF_JUPECHAN);
	if (aconf) {
	    sendto_one(cptr, getreply(ERR_CHANISJUPE),
		       me.name, cptr->name, cn, aconf->name, aconf->passwd);
	    return 0;
	}
    }
    for (; *cn; cn++) {
	if (*cn < 33 || *cn == ',' || *cn == 160) {
	    sendto_one(cptr, getreply(ERR_BADCHANNAME), me.name, cptr->name,
		       cn);
	    return 0;
	}
    }
    return 1;
}

/*
 * *  Get Channel block for chname (and allocate a new channel *
 * block, if it didn't exist before).
 */
aChannel *get_channel(aClient *cptr, char *chname, int flag)
{
    aChannel *chptr;
    int len;

    if (BadPtr(chname))
	return NULL;

    len = strlen(chname);
    if (MyClient(cptr) && len > CHANNELLEN) {
	len = CHANNELLEN;
	*(chname + CHANNELLEN) = '\0';
    }
    if ((chptr = find_channel(chname, (aChannel *) NULL)))
	return (chptr);
    if (flag == CREATE) {
	chptr = make_channel();

	strncpyzt(chptr->chname, chname, len + 1);
	if (channel)
	    channel->prevch = chptr;
	chptr->prevch = NULL;
	chptr->nextch = channel;
	channel = chptr;
	chptr->channelts = timeofday;
	(void) add_to_channel_hash_table(chname, chptr);
	Count.chan++;
    }
    return chptr;
}

/*
 * *  Subtract one user from channel i (and free channel *  block, if
 * channel became empty).
 */
static void sub1_from_channel(aChannel *chptr)
{
    Link *tmp;
    aBan *bp, *bprem;

    if (--chptr->users <= 0) {
	/*
	 * Now, find all invite links from channel structure
	 */
	while ((tmp = chptr->invites))
	    del_invite(tmp->value.cptr, chptr);

	bp = chptr->banlist;
	while (bp) {
	    bprem = bp;
	    bp = bp->next;
	    MyFree(bprem->banstr);
	    MyFree(bprem->who);
	    MyFree(bprem);
	}
	if (chptr->prevch)
	    chptr->prevch->nextch = chptr->nextch;
	else
	    channel = chptr->nextch;
	if (chptr->nextch)
	    chptr->nextch->prevch = chptr->prevch;
	(void) del_from_channel_hash_table(chptr->chname, chptr);
#ifdef FLUD
	free_fluders(NULL, chptr);
#endif
	free_channel(chptr);
	Count.chan--;
    }
}

/*
 * * m_kick * parv[0] = sender prefix *       parv[1] = channel *
 * parv[2] = client to kick *   parv[3] = kick comment
 */
int m_kick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *who;
    aChannel *chptr;
    int chasing = 0;
    int user_count;		/* count nicks being kicked, only allow 4 */
    char *comment, *name, *p = NULL, *user, *p2 = NULL;
    int power = 0;

    if (parc < 3 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KICK");
	return 0;
    }
    if (IsServer(sptr) && !IsULine(sptr))
	sendto_ops("KICK from %s for %s %s", parv[0], parv[1], parv[2]);
    comment = (BadPtr(parv[3])) ? parv[0] : parv[3];
    if (strlen(comment) > (size_t) TOPICLEN)
	comment[TOPICLEN] = '\0';

    *nickbuf = *buf = '\0';
    name = strtoken(&p, parv[1], ",");

    while (name) {
	chptr = get_channel(sptr, name, !CREATE);
	if (!chptr) {
	    sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
		       name);
	    name = strtoken(&p, (char *) NULL, ",");
	    continue;
	}

	/*
	 * You either have chan op privs, or you don't -Dianora 
	 */

	if (!IsServer(sptr) && !is_chan_op(sptr, chptr) && !IsULine(sptr)) {
	    if (MyConnect(sptr)) {
		sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name,
			   parv[0], chptr->chname);
		name = strtoken(&p, (char *) NULL, ",");
		continue;
	    }
	    if (chptr->channelts == 0) {
		sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name,
			   parv[0], chptr->chname);
		name = strtoken(&p, (char *) NULL, ",");
		continue;
	    }
	    /*
	     * Its a user doing a kick, but is not showing as chanop
	     * locally its also not a user ON -my- server, and the channel
	     * has a TS. There are two cases we can get to this point
	     * then...
	     * 
	     * 1) connect burst is happening, and for some reason a legit op
	     * has sent a KICK, but the SJOIN hasn't happened yet or been
	     * seen. (who knows.. due to lag...)
	     * 
	     * 2) The channel is desynced. That can STILL happen with TS
	     * 
	     * Now, the old code roger wrote, would allow the KICK to go
	     * through. Thats quite legit, but lets weird things like
	     * KICKS by users who appear not to be chanopped happen, or
	     * even neater, they appear not to be on the channel. This
	     * fits every definition of a desync, doesn't it? ;-) So I
	     * will allow the KICK, otherwise, things are MUCH worse. But
	     * I will warn it as a possible desync.
	     * 
	     * -Dianora
	     */
	    /*
	     * sendto_one(sptr, err_str(ERR_DESYNC), me.name, parv[0],
	     * chptr->chname);
	     */
	    /*
	     * After more discussion with orabidoo...
	     * 
	     * The code was sound, however, what happens if we have +h (TS4)
	     * and some servers don't understand it yet? we will be seeing
	     * servers with users who appear to have no chanops at all,
	     * merrily kicking users.... -Dianora
	     * 
	     */
	}

	user = strtoken(&p2, parv[2], ",");
	user_count = 4;
	while (user && user_count) {
	    user_count--;
	    if (!(who = find_chasing(sptr, user, &chasing))) {
		user = strtoken(&p2, (char *) NULL, ",");
		continue;
	    }
	    power = ((is_chan_owner(who, chptr) || is_protected(who, chptr))
		     && !IsPrivileged(sptr));
	    if (IsMember(who, chptr) && !IsAnon(who) && !power) {
		sendto_channel_butserv(chptr, sptr, ":%s KICK %s %s :%s",
				       parv[0], name, who->name, comment);
		sendto_match_servs(chptr, cptr, ":%s KICK %s %s :%s",
				   parv[0], name, who->name, comment);
		sendto_service(SERVICE_SEE_KICKS | SERVICE_SEE_PARTS,
			       SERVICE_SEE_PREFIX, ":%s KICK %s %s :%s",
			       parv[0], name, who->name, comment);
		if (cptr->user)
		    sendto_service(SERVICE_SEE_KICKS | SERVICE_SEE_PARTS |
				   SERVICE_SEE_PREFIX, 0,
				   ":%s!%s@%s KICK %s %s :%s", parv[0], name,
				   cptr->user->username, who->name, comment);
		remove_user_from_channel(who, chptr);
	    } else if (power) {
		sendto_one(sptr, ":%s NOTICE %s :Permission denied", me.name,
			   parv[0]);
	    } else {
		sendto_one(sptr, err_str(ERR_USERNOTINCHANNEL), me.name,
			   parv[0], user, name);
	    }
	    user = strtoken(&p2, (char *) NULL, ",");
	}
	name = strtoken(&p, (char *) NULL, ",");
    }
    return (0);
}

int count_channels(aClient *sptr)
{
    aChannel *chptr;
    int count = 0;

    for (chptr = channel; chptr; chptr = chptr->nextch)
	count++;
    return (count);
}

void send_topic_burst(aClient *cptr)
{
    aChannel *chptr;
    aClient *acptr;

    for (chptr = channel; chptr; chptr = chptr->nextch) {
	if (chptr->topic[0] != '\0')
	    sendto_one(cptr, ":%s TOPIC %s %s %ld :%s", me.name,
		       chptr->chname, chptr->topic_nick, chptr->topic_time,
		       chptr->topic);
    }
    for (acptr = client; acptr; acptr = acptr->next) {
	if (!IsPerson(acptr) || acptr->from == cptr)
	    continue;
	if (acptr->user->away)
	    sendto_one(cptr, ":%s AWAY :%s", acptr->name, acptr->user->away);
    }
}

/*
 * m_topic 
 * parv[0] = sender prefix 
 * parv[1] = topic text
 */
int m_topic(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel *chptr = NullChn;
    char *topic = NULL, *name, *tnick = sptr->name;
    time_t ts = timeofday;
    int member;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "TOPIC");
	return 0;
    }
    name = parv[1];
    chptr = find_channel(name, NullChn);
    if (!chptr) {
	sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
	return 0;
    }
    member = IsMember(sptr, chptr);
    if (parc == 2) {		/* user is requesting a topic */
    char *namep = chptr->chname;
    char tempchname[CHANNELLEN + 2];

	if (!member && !(ShowChannel(sptr, chptr))) {
	    if (IsAdmin(sptr)) {
		tempchname[0] = '%';
		strcpy(&tempchname[1], chptr->chname);
		namep = tempchname;
	    } else {
		sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0],
			   name);
		return 0;
	    }
	}

	if (chptr->topic[0] == '\0')
	    sendto_one(sptr, rpl_str(RPL_NOTOPIC), me.name, parv[0], namep);
	else {
	    sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], namep,
		       chptr->topic);
	    sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME), me.name, parv[0], namep,
		       chptr->topic_nick, chptr->topic_time);
	}
	return 0;
    }

    if (!member && !IsServer(sptr) && !IsULine(sptr)) {
	sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], name);
	return 0;
    }

    if (parc > 3 && (!MyConnect(sptr) || IsULine(sptr) || IsServer(sptr))) {
	topic = (parc > 4 ? parv[4] : "");
	tnick = parv[2];
	ts = atoi(parv[3]);
    } else {
	topic = parv[2];
    }
    if (((!(chptr->mode.mode & MODE_TOPICLIMIT)
	  || is_chan_op(sptr, chptr))
	 || IsULine(sptr) || IsAnOper(sptr) || IsServer(sptr))) {

	/*
	 * local topic is newer than remote topic and we have a topic
	 * and we're in a synch (server setting topic) 
	 */

	if (IsServer(sptr) && !IsULine(sptr) && chptr->topic_time >= ts &&
	    chptr->topic[0])
	    return 0;

	strncpyzt(chptr->topic, topic, TOPICLEN + 1);
	strcpy(chptr->topic_nick, tnick);
	chptr->topic_time = ts;

	/*
	 * in this case I think it's better that we send all the info that df 
	 * * sends with the topic, so I changed everything to work like that. -wd 
	 */

	sendto_match_servs(chptr, cptr, ":%s TOPIC %s %s %lu :%s",
			   parv[0], chptr->chname, chptr->topic_nick,
			   chptr->topic_time, chptr->topic);
	sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s", parv[0],
			       chptr->chname, chptr->topic);
	sendto_service(SERVICE_SEE_TOPIC, SERVICE_SEE_PREFIX,
		       ":%s TOPIC %s %s %lu :%s", parv[0], chptr->chname,
		       chptr->topic_nick, chptr->topic_time, chptr->topic);
	if (cptr->user)
	    sendto_service(SERVICE_SEE_TOPIC | SERVICE_SEE_PREFIX, 0,
			   ":%s!%s@%s TOPIC %s %s %lu :%s", parv[0],
			   cptr->user->username, cptr->user->host,
			   chptr->chname, chptr->topic_nick, chptr->topic_time,
			   chptr->topic);
    } else {
	sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0],
		   chptr->chname);
    }
    return 0;
}

void add_invite(aClient *cptr, aChannel *chptr)
{
    Link *inv, **tmp;

    del_invite(cptr, chptr);
    /*
     * delete last link in chain if the list is max length
     */
    if (list_length(cptr->user->invited) >= MAXCHANNELSPERUSER) {
	/*
	 * This forgets the channel side of invitation     -Vesa inv =
	 * cptr->user->invited; cptr->user->invited = inv->next;
	 * free_link(inv);
	 */
	del_invite(cptr, cptr->user->invited->value.chptr);

    }
    /*
     * add client to channel invite list
     */
    inv = make_link();
    inv->value.cptr = cptr;
    inv->next = chptr->invites;
    chptr->invites = inv;
    /*
     * add channel to the end of the client invite list
     */
    for (tmp = &(cptr->user->invited); *tmp; tmp = &((*tmp)->next));
    inv = make_link();
    inv->value.chptr = chptr;
    inv->next = NULL;
    (*tmp) = inv;
}

/*
 * Delete Invite block from channel invite list and client invite list
 */
void del_invite(aClient *cptr, aChannel *chptr)
{
    Link **inv, *tmp;

    for (inv = &(chptr->invites); (tmp = *inv); inv = &tmp->next)
	if (tmp->value.cptr == cptr) {
	    *inv = tmp->next;
	    free_link(tmp);
	    break;
	}

    for (inv = &(cptr->user->invited); (tmp = *inv); inv = &tmp->next)
	if (tmp->value.chptr == chptr) {
	    *inv = tmp->next;
	    free_link(tmp);
	    break;
	}
}

/*
 * * m_invite *       parv[0] - sender prefix *       parv[1] - user to
 * invite *     parv[2] - channel number
 */
int m_invite(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    aChannel *chptr;

    if (parc < 3 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "INVITE");
	return -1;
    }

    if (!(acptr = find_person(parv[1], (aClient *) NULL))) {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
	return 0;
    }

    if (!check_channelname(sptr, (unsigned char *) parv[2]))
	return 0;

    if (!(chptr = find_channel(parv[2], NullChn))) {
	sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], parv[2]);
	return 0;
    }

    if (chptr && !IsMember(sptr, chptr) && !IsULine(sptr)) {
	sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], parv[2]);
	return -1;
    }

    if (IsMember(acptr, chptr)) {
	sendto_one(sptr, err_str(ERR_USERONCHANNEL), me.name, parv[0], parv[1],
		   parv[2]);
	return 0;
    }
    if (chptr && !IsULine(sptr)) {
	if (!is_chan_op(sptr, chptr)) {
	    sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0],
		       chptr->chname);
	    return -1;
	}
    }

    if (MyConnect(sptr)) {
	sendto_one(sptr, rpl_str(RPL_INVITING), me.name, parv[0],
		   acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
	if (acptr->user->away)
	    sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], acptr->name,
		       acptr->user->away);
    }
    if (MyConnect(acptr)) {
	if ((chptr && sptr->user && is_chan_op(sptr, chptr))
	    || IsULine(sptr)) {
	    add_invite(acptr, chptr);
	    sendto_channel_butone(NULL, CHFL_CHANOP, &me, chptr,
				  ":%s NOTICE @%s :%s invited %s into channel %s.",
				  me.name, chptr->chname, sptr->name,
				  acptr->name, chptr->chname);
	}
    }
    sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s", parv[0],
		      acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
    return 0;
}

/*
 * The function which sends the actual channel list back to the user.
 * Operates by stepping through the hashtable, sending the entries back if
 * they match the criteria.
 * cptr = Local client to send the output back to.
 * numsend = Number (roughly) of lines to send back. Once this number has
 * been exceeded, send_list will finish with the current hash bucket,
 * and record that number as the number to start next time send_list
 * is called for this user. So, this function will almost always send
 * back more lines than specified by numsend (though not by much,
 * assuming CH_MAX is was well picked). So be conservative in your choice
 * of numsend. -Rak
 */

void send_list(aClient *cptr, int numsend)
{
    aChannel *chptr;
    LOpts *lopt = cptr->user->lopt;
    int hashnum;

    for (hashnum = lopt->starthash; hashnum < CH_MAX; hashnum++) {
	if (numsend > 0)
	    for (chptr = (aChannel *) hash_get_chan_bucket(hashnum); chptr;
		 chptr = chptr->hnextch) {
		if (SecretChannel(chptr) && !IsMember(cptr, chptr))
		    continue;
		if ((!lopt->showall) &&
		    ((chptr->users < lopt->usermin) || ((lopt->usermax >= 0)
							&& (chptr->users >
							    lopt->usermax))
		     || ((chptr->channelts || 1) < lopt->chantimemin)
		     || (chptr->topic_time < lopt->topictimemin)
		     || (chptr->channelts > lopt->chantimemax)
		     || (chptr->topic_time > lopt->topictimemax)
		     || (lopt->nolist
			 && find_str_link(lopt->nolist, chptr->chname))
		     || (lopt->yeslist
			 && !find_str_link(lopt->yeslist, chptr->chname))))
		    continue;
		sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name,
			   ShowChannel(cptr, chptr) ? chptr->chname : "*",
			   chptr->users, ShowChannel(cptr,
						     chptr) ? chptr->
			   topic : "");
		numsend--;
	} else
	    break;
    }

    /*
     * All done 
     */
    if (hashnum == CH_MAX) {
    Link *lp, *next;

	sendto_one(cptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
	for (lp = lopt->yeslist; lp; lp = next) {
	    next = lp->next;
	    free_link(lp);
	}
	for (lp = lopt->nolist; lp; lp = next) {
	    next = lp->next;
	    free_link(lp);
	}

	MyFree(cptr->user->lopt);
	cptr->user->lopt = NULL;
	return;
    }

    /*
     * We've exceeded the limit on the number of channels to send back
     * at once.
     */
    lopt->starthash = hashnum;
    return;
}

/*
 * * m_list *      parv[0] = sender prefix *      parv[1] = channel
 */
int m_list(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel *chptr;
    time_t currenttime = time(NULL);
    char *name, *p = NULL;
    LOpts *lopt = NULL;
    Link *lp, *next;
    int usermax, usermin, error = 0, doall = 0;
    int x;
    time_t chantimemin, chantimemax;
    ts_val topictimemin, topictimemax;
    Link *yeslist = NULL, *nolist = NULL;

    static char *usage[] = {
	"   Usage: /raw LIST options (on mirc) or /quote LIST options (ircII)",
	"",
	"If you don't include any options, the default is to send you the",
	"entire unfiltered list of channels. Below are the options you can",
	"use, and what channels LIST will return when you use them.",
	">number  List channels with more than <number> people.",
	"<number  List channels with less than <number> people.",
	"C>number List channels created between now and <number> minutes ago.",
	"C<number List channels created earlier than <number> minutes ago.",
	"T>number List channels whose topics are older than <number> minutes",
	"         (Ie, they have not changed in the last <number> minutes.",
	"T<number List channels whose topics are not older than <number> minutes.",
	"*mask*   List channels that match *mask*",
	"!*mask*  List channels that do not match *mask*",
	NULL
    };

    /*
     * Some starting san checks -- No interserver lists allowed. 
     */
    if (cptr != sptr || !sptr->user)
	return 0;

    /*
     * If a /list is in progress, then another one will cancel it 
     */
    if ((lopt = sptr->user->lopt) != NULL) {
	sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, parv[0]);
	for (lp = lopt->yeslist; lp; lp = next) {
	    next = lp->next;
	    free_link(lp);
	}
	for (lp = lopt->nolist; lp; lp = next) {
	    next = lp->next;
	    free_link(lp);
	}
	MyFree(sptr->user->lopt);
	sptr->user->lopt = NULL;
	return 0;
    }

    /*
     * if HTM, drop this too 
     */
    if (lifesux) {
	sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, sptr->name);
	return 0;
    }

    if (parc < 2 || BadPtr(parv[1])) {

	sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, parv[0]);
	lopt = sptr->user->lopt = (LOpts *) MyMalloc(sizeof(LOpts));
	memset(lopt, '\0', sizeof(LOpts));

	lopt->showall = 1;

	if (DBufLength(&cptr->sendQ) < 2048)
	    send_list(cptr, 64);

	return 0;
    }

    if ((parc == 2) && (parv[1][0] == '?') && (parv[1][1] == '\0')) {
    char **ptr = usage;

	for (; *ptr; ptr++)
	    sendto_one(sptr, rpl_str(RPL_COMMANDSYNTAX), me.name, cptr->name,
		       *ptr);
	return 0;
    }

    sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, parv[0]);

    chantimemax = topictimemax = currenttime + 86400;
    chantimemin = topictimemin = 0;
    usermin = 2;		/* By default, set the minimum to 2 users */
    usermax = -1;		/* No maximum */

    for (name = strtoken(&p, parv[1], ","); name && !error;
	 name = strtoken(&p, (char *) NULL, ",")) {

	switch (*name) {
	    case '<':
		usermax = atoi(name + 1) - 1;
		doall = 1;
		break;
	    case '>':
		usermin = atoi(name + 1) + 1;
		doall = 1;
		break;
	    case 'C':
	    case 'c':		/* Channel TS time -- creation time? */
		++name;
		switch (*name++) {
		    case '<':
			chantimemax = currenttime - 60 * atoi(name);
			doall = 1;
			break;
		    case '>':
			chantimemin = currenttime - 60 * atoi(name);
			doall = 1;
			break;
		    default:
			sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name,
				   cptr->name);
			error = 1;
		}
		break;
	    case 'T':
	    case 't':
		++name;
		switch (*name++) {
		    case '<':
			topictimemax = currenttime - 60 * atoi(name);
			doall = 1;
			break;
		    case '>':
			topictimemin = currenttime - 60 * atoi(name);
			doall = 1;
			break;
		    default:
			sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name,
				   cptr->name);
			error = 1;
		}
		break;
	    default:		/* A channel, possibly with wildcards. Thought
				 * for the future: Consider turning wildcard
				 * processing on the fly. new syntax:
				 * !channelmask will tell ircd to ignore any
				 * channels matching that mask, and then
				 * channelmask will tell ircd to send us a list 
				 * * * * of channels only masking channelmask.
				 * Note: Specifying a channel without wildcards 
				 * * * will return that channel even if any of
				 * the !channelmask masks matches it. */
		if (*name == '!') {
		    doall = 1;
		    lp = make_link();
		    lp->next = nolist;
		    nolist = lp;
		    DupString(lp->value.cp, name + 1);
		} else if (strchr(name, '*') || strchr(name, '*')) {
		    doall = 1;
		    lp = make_link();
		    lp->next = yeslist;
		    yeslist = lp;
		    DupString(lp->value.cp, name);
		} else {	/* Just a normal channel */

		    chptr = find_channel(name, NullChn);
		    if (chptr && ((x = ShowChannel(sptr, chptr))
				  || IsAdmin(sptr))) {
    char *nameptr = name;
    char channame[CHANNELLEN + 2];

			if (!x && IsAdmin(sptr)) {
			    channame[0] = '%';
			    strcpy(&channame[1], chptr->chname);
			    nameptr = channame;
			}

			sendto_one(sptr, rpl_str(RPL_LIST), me.name, parv[0],
				   nameptr, chptr->users, chptr->topic);
		    }
		}
	}			/* switch */
    }				/* while */

    if (doall) {
	lopt = sptr->user->lopt = (LOpts *) MyMalloc(sizeof(LOpts));
	memset(lopt, '\0', sizeof(LOpts));
	lopt->usermin = usermin;
	lopt->usermax = usermax;
	lopt->topictimemax = topictimemax;
	lopt->topictimemin = topictimemin;
	lopt->chantimemax = chantimemax;
	lopt->chantimemin = chantimemin;
	lopt->nolist = nolist;
	lopt->yeslist = yeslist;

	if (DBufLength(&cptr->sendQ) < 2048)
	    send_list(cptr, 64);
	return 0;
    }

    sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, parv[0]);

    return 0;
}

/************************************************************************
 * m_names() - Added by Jto 27 Apr 1989
 * 12 Feb 2000 - geesh, time for a rewrite -lucas
 ************************************************************************/
/*
 * m_names 
 * parv[0] = sender prefix 
 * parv[1] = channel
 */

/*
 * maximum names para to show to opers when abuse occurs 
 */
#define TRUNCATED_NAMES 64

int m_names(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int mlen = strlen(me.name) + NICKLEN + 7;
    aChannel *chptr;
    aClient *acptr;
    int member;
    chanMember *cm;
    int idx, flag = 1, spos;
    char *s, *para = parv[1];

    if (parc < 2 || !MyConnect(sptr)) {
	sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");
	return 0;
    }

    for (s = para; *s; s++) {
	if (*s == ',') {
	    if (strlen(para) > TRUNCATED_NAMES)
		para[TRUNCATED_NAMES] = '\0';
	    sendto_realops("names abuser %s %s", get_client_name(sptr, FALSE),
			   para);
	    sendto_one(sptr, err_str(ERR_TOOMANYTARGETS), me.name, sptr->name,
		       "NAMES");
	    return 0;
	}
    }

    if (!check_channelname(sptr, (unsigned char *) para))
	return 0;

    chptr = find_channel(para, (aChannel *) NULL);

    if (!chptr || !ShowChannel(sptr, chptr)) {
	sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], para);
	return 0;
    }

    member = IsAnOper(sptr) ? 1 : (IsMember(sptr, chptr)
				   || IsService(sptr));

    if (PubChannel(chptr))
	buf[0] = '=';
    else if (SecretChannel(chptr))
	buf[0] = '@';
    else
	buf[0] = '*';

    idx = 1;
    buf[idx++] = ' ';
    for (s = chptr->chname; *s; s++)
	buf[idx++] = *s;
    buf[idx++] = ' ';
    buf[idx++] = ':';

    /*
     * If we go through the following loop and never add anything,
     * we need this to be empty, otherwise spurious things from the
     * LAST /names call get stuck in there.. - lucas 
     */
    buf[idx] = '\0';

    spos = idx;			/* starting point in buffer for names! */

    for (cm = chptr->members; cm; cm = cm->next) {
	acptr = cm->cptr;
	if (IsAnon(acptr))
	    continue;
	if (IsInvisible(acptr) && !member)
	    continue;
	if (IsChanAnon(chptr) && (cptr != acptr))
	    continue;
	if (cm->flags & CHFL_OWNER)
	    buf[idx++] = '.';
	else if (cm->flags & CHFL_CHANOP)
	    buf[idx++] = '@';
	else if (cm->flags & CHFL_VOICE)
	    buf[idx++] = '+';
	for (s = acptr->name; *s; s++)
	    buf[idx++] = *s;
	buf[idx++] = ' ';
	buf[idx] = '\0';
	flag = 1;
	if (mlen + idx + NICKLEN > BUFSIZE - 3) {
	    sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
	    idx = spos;
	    flag = 0;
	}
    }

    if (flag)
	sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);

    sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], para);

    return 0;
}

char *pretty_mask(char *mask)
{
    char *cp, *user, *host;

    if ((user = strchr((cp = mask), '!')))
	*user++ = '\0';
    if ((host = strrchr(user ? user : cp, '@'))) {
	*host++ = '\0';
	if (!user)
	    return make_nick_user_host(NULL, cp, host);
    } else if (!user && strchr(cp, '.'))
	return make_nick_user_host(NULL, NULL, cp);
    return make_nick_user_host(cp, user, host);
}
