/*
 *   IRC - Internet Relay Chat, src/m_sjoin.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "h.h"

#ifdef NO_CHANOPS_WHEN_SPLIT
#include "fdlist.h"
extern fdlist serv_fdlist;

int server_was_split = YES;
time_t server_split_time = 0;
int server_split_recovery_time = (MAX_SERVER_SPLIT_RECOVERY_TIME * 60);

#endif

#ifdef ORATIMING
struct timeval tsdnow, tsdthen;
unsigned long tsdms;
#endif

/*
 * externally defined function 
 */
extern Link *find_channel_link(Link *, aChannel *);	/* defined in list.c */

#ifdef ANTI_SPAMBOT
extern int spam_num;		/* defined in s_serv.c */
extern int spam_time;		/* defined in s_serv.c */
#endif

/*
 * number of seconds to add to all readings of time() when making TS's 
 */

static char *PartFmt = ":%s PART %s";
static char *PartFmt2 = ":%s PART %s :%s";

/*
 * old and new server SJOIN formats: old with dual TS 
 */
static char *oldSJOINFmt = ":%s SJOIN %ld %ld %s %s %s :%s";
static char *newSJOINFmt = ":%s SJOIN %ld %s %s %s :%s";

/*
 * NP means no paramaters, don't send the extra space there 
 */
static char *oldSJOINFmtNP = ":%s SJOIN %ld %ld %s %s :%s";
static char *newSJOINFmtNP = ":%s SJOIN %ld %s %s :%s";

/*
 * client sjoin.. old is the same as server, new is our new version 
 */
static char *oldCliSJOINFmt = ":%s SJOIN %ld %ld %s + :%s";
static char *newCliSJOINFmt = ":%s SJOIN %ld %s";

/*
 * some buffers for rebuilding channel/nick lists with ,'s 
 */
static char buf[BUFSIZE];
static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];

int can_join(aClient *sptr, aChannel *chptr, char *key)
{
    Link *lp;
    int invited = 0;

    for (lp = sptr->user->invited; lp; lp = lp->next) {
	if (lp->value.chptr == chptr) {
	    invited = 1;
	    break;
	}
    }
    if (invited || IsULine(sptr) || IsAnOper(sptr)
	|| has_invite(sptr, chptr))
	return 0;
    if (!IsOper(sptr)
	&& find_conf_name(chptr->chname, CONF_QUARANTINED_CHAN))
	return (ERR_NOPRIVILEGES);
    if (is_banned(sptr, chptr)) {
	if (has_banex(sptr, chptr))
	    return 0;
	else
	    return (ERR_BANNEDFROMCHAN);
    }
    if (chptr->mode.mode & MODE_INVITEONLY)
	return (ERR_INVITEONLYCHAN);
    if (chptr->mode.mode & MODE_REGONLY && !IsRegNick(sptr))
	return (ERR_NEEDREGGEDNICK);
    if (chptr->mode.mode & MODE_OPERONLY && !IsOper(sptr))
	return (ERR_NOPRIVILEGES);
    if (*chptr->mode.key && (BadPtr(key) || mycmp(chptr->mode.key, key)))
	return (ERR_BADCHANNELKEY);
    if (chptr->mode.limit && chptr->users >= chptr->mode.limit)
	return (ERR_CHANNELISFULL);
    return 0;
}

void send_user_joins(aClient *cptr, aClient *user)
{
    Link *lp;
    aChannel *chptr;
    int cnt = 0, len = 0, clen;
    char *mask;

    *buf = ':';
    (void) strcpy(buf + 1, user->name);
    (void) strcat(buf, " JOIN ");
    len = strlen(user->name) + 7;

    for (lp = user->user->channel; lp; lp = lp->next) {
	chptr = lp->value.chptr;
	if (*chptr->chname == '&')
	    continue;
	if ((mask = strchr(chptr->chname, ':')))
	    if (match(++mask, cptr->name))
		continue;
	clen = strlen(chptr->chname);
	if (clen > (size_t) BUFSIZE - 7 - len) {
	    if (cnt)
		sendto_one(cptr, "%s", buf);
	    *buf = ':';
	    (void) strcpy(buf + 1, user->name);
	    (void) strcat(buf, " JOIN ");
	    len = strlen(user->name) + 7;
	    cnt = 0;
	}
	(void) strcpy(buf + len, chptr->chname);
	cnt++;
	len += clen;
	if (lp->next) {
	    len++;
	    (void) strcat(buf, ",");
	}
    }
    if (*buf && cnt)
	sendto_one(cptr, "%s", buf);

    return;
}

/*
 * * m_join * parv[0] = sender prefix *       parv[1] = channel *
 * parv[2] = channel password (key) 
 */
int m_join(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static char jbuf[BUFSIZE];
    Link *lp;
    aChannel *chptr;
    char *name, *key = NULL;
    int i, flags = 0, chanlen = 0;
    int allow_op = YES;
    char *p = NULL, *p2 = NULL;

#ifdef ANTI_SPAMBOT
    int successful_join_count = 0;

    /*
     * Number of channels successfully joined 
     */
#endif

    if (!(sptr->user)) {
	/*
	 * something is *fucked* - bail 
	 */
	return 0;
    }

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "JOIN");
	return 0;
    }

    *jbuf = '\0';
    /*
     * * Rebuild list of channels joined to be the actual result of the *
     * JOIN.  Note that "JOIN 0" is the destructive problem.
     */
    for (i = 0, name = strtoken(&p, parv[1], ","); name;
	 name = strtoken(&p, (char *) NULL, ",")) {
	/*
	 * pathological case only on longest channel name. * If not dealt
	 * with here, causes desynced channel ops * since ChannelExists()
	 * doesn't see the same channel * as one being joined. cute bug.
	 * Oct 11 1997, Dianora/comstud
	 */
	if (!check_channelname(sptr, (unsigned char *) name))
	    continue;

	chanlen = strlen(name);
	if (chanlen > CHANNELLEN) {
	    name[CHANNELLEN] = '\0';
	    chanlen = CHANNELLEN;
	}
	if (*name == '&' && !MyConnect(sptr))
	    continue;
	if (*name == '0' && !atoi(name))
	    *jbuf = '\0';
	else if (!IsChannelName(name)) {
	    if (MyClient(sptr))
		sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
			   name);
	    continue;
	}
	if (*jbuf)
	    (void) strcat(jbuf, ",");
	(void) strncat(jbuf, name, sizeof(jbuf) - i - 1);
	i += chanlen + 1;
    }

    /*
     * (void)strcpy(parv[1], jbuf);
     */

    p = NULL;
    if (parv[2])
	key = strtoken(&p2, parv[2], ",");
    parv[2] = NULL;		/* for m_names call later, parv[parc] * must == 
				 * NULL */
    for (name = strtoken(&p, jbuf, ","); name;
	 key = (key) ? strtoken(&p2, NULL, ",") : NULL, name =
	 strtoken(&p, NULL, ",")) {
	/*
	 * * JOIN 0 sends out a part for all channels a user * has
	 * joined.
	 */
	if (*name == '0' && !atoi(name)) {
	    if (sptr->user->channel == NULL)
		continue;
	    while ((lp = sptr->user->channel)) {
		chptr = lp->value.chptr;
		sendto_channel_butserv(chptr, sptr, PartFmt, parv[0],
				       chptr->chname);
		remove_user_from_channel(sptr, chptr);
	    }

#ifdef ANTI_SPAMBOT

	    /*
	     * -Dianora 
	     */
	    if (MyConnect(sptr) && !IsAnOper(sptr)) {
		if (sptr->join_leave_count >= spam_num) {
		    sendto_ops_lev(SPAM_LEV,
				   "User %s (%s@%s) is a possible spambot",
				   sptr->name, sptr->user->username,
				   sptr->user->host);
		    sptr->oper_warn_count_down = OPER_SPAM_COUNTDOWN;
		} else {
    int t_delta;

		    if ((t_delta =
			 (NOW - sptr->last_leave_time)) >
			JOIN_LEAVE_COUNT_EXPIRE_TIME) {
    int decrement_count;

			decrement_count =
			    (t_delta / JOIN_LEAVE_COUNT_EXPIRE_TIME);
			if (decrement_count > sptr->join_leave_count)
			    sptr->join_leave_count = 0;
			else
			    sptr->join_leave_count -= decrement_count;
		    } else {
			if ((NOW - (sptr->last_join_time)) < spam_time)
			    sptr->join_leave_count++;
		    }
		    sptr->last_leave_time = NOW;
		}
	    }
#endif
	    sendto_match_servs(NULL, cptr, ":%s JOIN 0", parv[0]);
	    continue;
	}

	if (MyConnect(sptr)) {
	    /*
	     * * local client is first to enter previously nonexistent *
	     * channel so make them (rightfully) the Channel * Operator.
	     */
	    flags = (ChannelExists(name)) ? 0 : CHFL_CHANOP;
#ifdef NO_CHANOPS_WHEN_SPLIT
	    if (!IsAnOper(sptr) && server_was_split &&
		server_split_recovery_time) {
		if ((server_split_time + server_split_recovery_time) < NOW) {
		    if (serv_fdlist.last_entry)
			server_was_split = NO;
		    else {
			server_split_time = NOW;
			allow_op = NO;
		    }
		} else {
		    allow_op = NO;
		}
	    }
#endif

	    if ((sptr->user->joined >= MAXCHANNELSPERUSER) && (!IsAnOper(sptr)
							       || (sptr->user->
								   joined >=
								   MAXCHANNELSPERUSER
								   * 3))) {
		sendto_one(sptr, err_str(ERR_TOOMANYCHANNELS), me.name, parv[0],
			   name);
#ifdef ANTI_SPAMBOT
		if (successful_join_count)
		    sptr->last_join_time = NOW;
#endif
		return 0;
	    }
#ifdef ANTI_SPAMBOT
	    /*
	     * Dianora 
	     */
	    if (flags == 0)	/* if channel doesn't exist, don't * * * * * *
				 * * * * * * * * * penalize */
		successful_join_count++;

	    if ((sptr->join_leave_count >= spam_num)) {
		if (sptr->oper_warn_count_down > 0)
		    sptr->oper_warn_count_down--;
		else
		    sptr->oper_warn_count_down = 0;
		if (sptr->oper_warn_count_down == 0) {
		    sendto_ops_lev(SPAM_LEV,
				   "User %s (%s@%s) trying to join %s is a possible spambot",
				   sptr->name, sptr->user->username,
				   sptr->user->host, name);
		    sptr->oper_warn_count_down = OPER_SPAM_COUNTDOWN;
		}
# ifndef ANTI_SPAMBOT_WARN_ONLY
		return 0;
		/*
		 * Don't actually JOIN anything, but don't let spambot know that 
		 */
# endif
	    }
#endif
	} else {
	    if (!ChannelExists(name))
		sendto_realops("User on %s remotely JOINing new channel",
			       sptr->user->server);
	}

	chptr = get_channel(sptr, name, CREATE);

	if (!chptr || (MyConnect(sptr) && (i = can_join(sptr, chptr, key)))) {
	    sendto_one(sptr, getreply(i), me.name, parv[0], name);
#ifdef ANTI_SPAMBOT
	    if (successful_join_count > 0)
		successful_join_count--;
#endif
	    continue;
	}
	if (IsMember(sptr, chptr))
	    continue;

	/*
	 * only complain when the user can join the channel, the channel is being created by this user,
	 * and this user is not allowed to be an op. - lucas 
	 */

#ifdef NO_CHANOPS_WHEN_SPLIT
	if (flags && !allow_op)
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** Notice -- Due to a network split, you can not obtain"
		       "channel operator status at this time.", me.name,
		       sptr->name);
#endif

	if (allow_op)
	    add_user_to_channel(chptr, sptr, flags);
	else
	    add_user_to_channel(chptr, sptr, 0);

	if (!(*chptr->chname == '&')) {

	    if (MyClient(sptr) && flags == CHFL_CHANOP) {
		chptr->channelts = timeofday;

		/*
		 * we keep channel "creations" to the server sjoin format,
		 * so we can bounce modes and stuff if our ts is older. 
		 */

		if (allow_op) {
		    sendto_flag_serv_butone(cptr, 0, CAPAB_NSJOIN,
					    ":%s SJOIN %ld %ld %s +tn :@%s",
					    me.name, chptr->channelts,
					    chptr->channelts, name, parv[0]);
		    sendto_flag_serv_butone(cptr, CAPAB_NSJOIN, 0,
					    ":%s SJOIN %ld %s +tn :@%s",
					    me.name, chptr->channelts, name,
					    parv[0]);
		} else {
		    sendto_flag_serv_butone(cptr, 0, CAPAB_NSJOIN,
					    ":%s SJOIN %ld %ld %s +tn :%s",
					    me.name, chptr->channelts,
					    chptr->channelts, name, parv[0]);
		    sendto_flag_serv_butone(cptr, CAPAB_NSJOIN, 0,
					    ":%s SJOIN %ld %s +tn :%s",
					    me.name, chptr->channelts, name,
					    parv[0]);
		}
		sendto_service(SERVICE_SEE_JOINS, SERVICE_SEE_PREFIX,
			       ":%s JOIN :%s", parv[0], name);
		sendto_service(SERVICE_SEE_JOINS | SERVICE_SEE_PREFIX, 0,
			       ":%s!%s@%s JOIN :%s", parv[0],
			       sptr->user->username, sptr->user->host, name);
	    } else if (MyClient(sptr)) {
		sendto_flag_serv_butone(cptr, 0, CAPAB_NSJOIN,
					oldCliSJOINFmt, me.name,
					chptr->channelts, chptr->channelts,
					name, parv[0]);
		sendto_flag_serv_butone(cptr, CAPAB_NSJOIN, 0, newCliSJOINFmt,
					parv[0], chptr->channelts, name);
		sendto_service(SERVICE_SEE_JOINS, SERVICE_SEE_PREFIX,
			       ":%s JOIN :%s", parv[0], name);
		sendto_service(SERVICE_SEE_JOINS | SERVICE_SEE_PREFIX, 0,
			       ":%s!%s@%s JOIN :%s", parv[0],
			       sptr->user->username, sptr->user->host, name);
	    } else {
		sendto_match_servs(chptr, cptr, ":%s JOIN :%s", parv[0], name);
	    }
	}

	/*
	 * notify all other users on the new channel
	 */
	sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", parv[0], name);

	if (flags == CHFL_CHANOP && allow_op) {
	    sendto_channel_butserv(chptr, sptr, ":%s MODE %s +tn", me.name,
				   name);
	    chptr->mode.mode |= MODE_TOPICLIMIT;
	    chptr->mode.mode |= MODE_NOPRIVMSGS;
	}
	if (MyClient(sptr)) {
	    del_invite(sptr, chptr);
	    if (chptr->topic[0] != '\0') {
		sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], name,
			   chptr->topic);
		sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME), me.name, parv[0],
			   name, chptr->topic_nick, chptr->topic_time);
	    }
	    parv[1] = name;
	    (void) m_names(cptr, sptr, 2, parv);
	}
    }

#ifdef ANTI_SPAMBOT
    if (MyConnect(sptr) && successful_join_count)
	sptr->last_join_time = NOW;
#endif
    return 0;
}

/*
 * * m_part * parv[0] = sender prefix *       parv[1] = channel *
 * parv[2] = Optional part reason
 */
int m_part(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel *chptr;
    char *p, *name;
    char *reason = (parc > 2 && parv[2]) ? parv[2] : NULL;

    if (parc < 2 || parv[1][0] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "PART");
	return 0;
    }

    name = strtoken(&p, parv[1], ",");

#ifdef ANTI_SPAMBOT
    if (name && MyConnect(sptr) && !IsAnOper(sptr)) {
	if (sptr->join_leave_count >= spam_num) {
	    sendto_ops_lev(SPAM_LEV,
			   "User %s (%s@%s) is a possible spambot",
			   sptr->name, sptr->user->username, sptr->user->host);
	    sptr->oper_warn_count_down = OPER_SPAM_COUNTDOWN;
	} else {
    int t_delta;

	    if ((t_delta =
		 (NOW - sptr->last_leave_time)) >
		JOIN_LEAVE_COUNT_EXPIRE_TIME) {
    int decrement_count;

		decrement_count = (t_delta / JOIN_LEAVE_COUNT_EXPIRE_TIME);
		if (decrement_count > sptr->join_leave_count)
		    sptr->join_leave_count = 0;
		else
		    sptr->join_leave_count -= decrement_count;
	    } else {
		if ((NOW - (sptr->last_join_time)) < spam_time)
		    sptr->join_leave_count++;
	    }
	    sptr->last_leave_time = NOW;
	}
    }
#endif

    while (name) {
	chptr = get_channel(sptr, name, 0);
	if (!chptr) {
	    sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
		       name);
	    name = strtoken(&p, (char *) NULL, ",");
	    continue;
	}

	if (!IsMember(sptr, chptr)) {
	    sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], name);
	    name = strtoken(&p, (char *) NULL, ",");
	    continue;
	}
	if (parc < 3 || can_send(sptr, chptr, reason)) {
	    sendto_match_servs(chptr, cptr, PartFmt, parv[0], name);
	    sendto_channel_butserv(chptr, sptr, PartFmt, parv[0], name);
	    sendto_service(SERVICE_SEE_PARTS | SERVICE_SEE_JOINS,
			   SERVICE_SEE_PREFIX, ":%s PART %s", parv[0], name);
	    sendto_service(SERVICE_SEE_PARTS | SERVICE_SEE_JOINS |
			   SERVICE_SEE_PREFIX, 0, ":%s!%s@%s PART %s",
			   parv[0], sptr->user->username, sptr->user->host,
			   name);
	} else {
	    sendto_match_servs(chptr, cptr, PartFmt2, parv[0], name, reason);
	    sendto_channel_butserv(chptr, sptr, PartFmt2, parv[0], name,
				   reason);
	    sendto_service(SERVICE_SEE_PARTS | SERVICE_SEE_JOINS,
			   SERVICE_SEE_PREFIX, ":%s PART %s :%s", parv[0], name,
			   reason);
	    sendto_service(SERVICE_SEE_PARTS | SERVICE_SEE_JOINS |
			   SERVICE_SEE_PREFIX, 0, ":%s!%s@%s PART %s :%s",
			   parv[0], sptr->user->username, sptr->user->host,
			   name, reason);
	}
	remove_user_from_channel(sptr, chptr);
	name = strtoken(&p, (char *) NULL, ",");
    }
    return 0;
}

static inline void sjoin_sendit(aClient *cptr, aClient *sptr, aChannel *chptr,
				char *from)
{
    sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", from,
			   chptr->chname, modebuf, parabuf);
}

/*
 * m_sjoin
 * parv[0] - sender
 * parv[1] - TS
 * parv[2] - channel   
 * parv[3] - modes + n arguments (key and/or limit)
 * parv[4+n] - flags+nick list (all in one parameter)
 *
 *
 * process a SJOIN, taking the TS's into account to either ignore the
 * incoming modes or undo the existing ones or merge them, and JOIN all
 * the specified users while sending JOIN/MODEs to non-TS servers and
 * to clients
 */

#define INSERTSIGN(x,y) \
if (what != x) { \
*mbuf++=y; \
what = x; \
}

#define ADD_PARA(p) para = p; if(pbpos) parabuf[pbpos++] = ' '; \
                     while(*para) parabuf[pbpos++] = *para++;
#define ADD_SJBUF(p) para = p; if(sjbufpos) sjbuf[sjbufpos++] = ' '; \
                     while(*para) sjbuf[sjbufpos++] = *para++;

int m_sjoin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel *chptr;
    aClient *acptr;
    ts_val newts, oldts, tstosend;
    static Mode mode, *oldmode;
    chanMember *cm;
    int args = 0, haveops = 0, keepourmodes = 1, keepnewmodes =
	1, doesop = 0, what = 0, pargs = 0, fl, people = 0, isnew, clientjoin =
	0, pbpos, sjbufpos;

    char *s, *s0, *para;
    static char numeric[16], sjbuf[BUFSIZE];
    char keep_modebuf[REALMODEBUFLEN], keep_parabuf[REALMODEBUFLEN];
    char *mbuf = modebuf, *p;

    /*
     * if my client is SJOINing, it's just a local user being a dufus.
     * *  Ignore him.
     * * parc >= 5 (new serv<->serv SJOIN format)
     * * parc >= 6 (old serv<->serv SJOIN format)
     * * parc == 3 (new serv<->serv cliSJOIN format)
     */

    if (MyClient(sptr) || (parc < 5 && IsServer(sptr))
	|| (parc < 3 && IsPerson(sptr)))
	return 0;

    if (parc == 3 && IsPerson(sptr))
	clientjoin = 1;
    else if (isdigit(parv[2][0])) {
    int i;

	if (parc < 6)
	    return 0;

	for (i = 2; i < (parc - 1); i++)
	    parv[i] = parv[i + 1];

	parc--;
    }

    if (!IsChannelName(parv[2]))
	return 0;

    newts = atol(parv[1]);

    isnew = ChannelExists(parv[2]) ? 0 : 1;
    chptr = get_channel(sptr, parv[2], CREATE);
    oldts = chptr->channelts;

    if (chptr == find_channel(parv[2], (aChannel *) NULL)) {
	for (cm = chptr->members; cm; cm = cm->next)
	    if (cm->flags & MODE_CHANOP) {
		haveops++;
		break;
	    }
    } else
	return 0;

    if (clientjoin) {		/* we have a good old (new :) client sjoin,
				 * with timestamp */
	if (isnew)
	    chptr->channelts = tstosend = newts;
	else if (newts == 0 || oldts == 0)
	    chptr->channelts = tstosend = 0;
	else if (newts == oldts)
	    tstosend = oldts;
	else if (newts < oldts) {
	    if (haveops)
		tstosend = oldts;
	    else
		chptr->channelts = tstosend = newts;
	} else
	    tstosend = oldts;

	/*
	 * parv[0] is the client that is joining. parv[0] == sptr->name 
	 */

	if (!IsMember(sptr, chptr)) {
	    add_user_to_channel(chptr, sptr, 0);
	    sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", parv[0],
				   parv[2]);
	}
	if (!(*chptr->chname == '&')) {
	    sendto_flag_serv_butone(cptr, 0, CAPAB_NSJOIN, oldCliSJOINFmt,
				    me.name, tstosend, tstosend, parv[2],
				    parv[0]);
	    sendto_flag_serv_butone(cptr, CAPAB_NSJOIN, 0, newCliSJOINFmt,
				    parv[0], tstosend, parv[2]);
	    sendto_service(SERVICE_SEE_JOINS, SERVICE_SEE_PREFIX,
			   ":%s JOIN :%s", parv[0], parv[2]);
	    sendto_service(SERVICE_SEE_JOINS | SERVICE_SEE_PREFIX, 0,
			   ":%s!%s@%s JOIN :%s", parv[0], sptr->user->username,
			   sptr->user->host, parv[2]);
	}
	return 0;
    }

    memset((char *) &mode, '\0', sizeof(mode));
    doesop = (parv[4 + args][0] == '@' ||
	      parv[4 + args][1] == '@' ||
	      parv[4 + args][2] == '@' || parv[4 + args][3] == '@');

    s = parv[3];
    while (*s)
	switch (*(s++)) {
	    case 'i':
		mode.mode |= MODE_INVITEONLY;
		break;
	    case 'n':
		mode.mode |= MODE_NOPRIVMSGS;
		break;
	    case 'p':
		mode.mode |= MODE_PRIVATE;
		break;
	    case 's':
		mode.mode |= MODE_SECRET;
		break;
	    case 'm':
		mode.mode |= MODE_MODERATED;
		break;
	    case 't':
		mode.mode |= MODE_TOPICLIMIT;
		break;
	    case 'r':
		mode.mode |= MODE_REGISTERED;
		break;
	    case 'R':
		mode.mode |= MODE_REGONLY;
		break;
	    case 'c':
		mode.mode |= MODE_NOCOLOR;
		break;
	    case 'O':
		mode.mode |= MODE_OPERONLY;
		break;
	    case 'k':
		strncpyzt(mode.key, parv[4 + args], KEYLEN + 1);
		args++;
		if (parc < 5 + args)
		    return 0;
		break;
	    case 'l':
		mode.limit = atoi(parv[4 + args]);
		args++;
		if (parc < 5 + args)
		    return 0;
		break;
	}

    oldmode = &chptr->mode;

    if (isnew)
	chptr->channelts = tstosend = newts;
    else if (newts == 0 || oldts == 0)
	chptr->channelts = tstosend = 0;
    else if (newts == oldts)
	tstosend = oldts;
    else if (newts < oldts) {
	/*
	 * if remote ts is older, and they have ops, don't keep our modes. 
	 */
	if (doesop) {
	    kill_ban_list(sptr, chptr);
	    kill_banex_list(sptr, chptr);
	    kill_invites_list(sptr, chptr);
	    keepourmodes = 0;
	}
	if (haveops && !doesop)
	    tstosend = oldts;
	else
	    chptr->channelts = tstosend = newts;
    } else {
	/*
	 * if our TS is older, and we have ops, don't keep their modes 
	 */
	if (haveops)
	    keepnewmodes = 0;
	if (doesop && !haveops) {
	    chptr->channelts = tstosend = newts;
	    if (MyConnect(sptr) && !IsULine(sptr))
		sendto_realops("Hacked ops on opless channel: %s",
			       chptr->chname);
	} else {
	    tstosend = oldts;
	}
    }
    if (!keepnewmodes)
	mode = *oldmode;
    else if (keepourmodes) {
	mode.mode |= oldmode->mode;
	if (oldmode->limit > mode.limit)
	    mode.limit = oldmode->limit;
	if (*oldmode->key && *mode.key && strcmp(mode.key, oldmode->key) > 0) {
	    /*
	     * sketchy: keep the key that's lexographically greater
	     * if we both have a differing key. 
	     */
	    strcpy(mode.key, oldmode->key);
	}
    }

    pbpos = 0;

    /*
     * since the most common case is that the modes are exactly the same,
     *  this if will skip over the most common case... :)
     *  
     * this would look prettier in a for loop, but it's unrolled here
     *  so it's a bit faster.   - lucas
     *  
     * pass +: go through and add new modes that are in mode and not oldmode
     * pass -: go through and delete old modes that are in oldmode and not mode
     */

    if (mode.mode != oldmode->mode) {
	/*
	 * plus modes 
	 */
	if ((MODE_PRIVATE & mode.mode) && !(MODE_PRIVATE & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'p';
	}
	if ((MODE_SECRET & mode.mode) && !(MODE_SECRET & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 's';
	}
	if ((MODE_MODERATED & mode.mode)
	    && !(MODE_MODERATED & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'm';
	}
	if ((MODE_NOPRIVMSGS & mode.mode)
	    && !(MODE_NOPRIVMSGS & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'n';
	}
	if ((MODE_TOPICLIMIT & mode.mode)
	    && !(MODE_TOPICLIMIT & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 't';
	}
	if ((MODE_INVITEONLY & mode.mode)
	    && !(MODE_INVITEONLY & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'i';
	}
	if ((MODE_REGISTERED & mode.mode)
	    && !(MODE_REGISTERED & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'r';
	}
	if ((MODE_REGONLY & mode.mode)
	    && !(MODE_REGONLY & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'R';
	}
	if ((MODE_OPERONLY & mode.mode)
	    && !(MODE_OPERONLY & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'O';
	}
	if ((MODE_NOCOLOR & mode.mode)
	    && !(MODE_NOCOLOR & oldmode->mode)) {
	    INSERTSIGN(1, '+');
	    *mbuf++ = 'c';
	}

	/*
	 * minus modes 
	 */
	if ((MODE_PRIVATE & oldmode->mode) && !(MODE_PRIVATE & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'p';
	}
	if ((MODE_SECRET & oldmode->mode) && !(MODE_SECRET & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 's';
	}
	if ((MODE_MODERATED & oldmode->mode)
	    && !(MODE_MODERATED & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'm';
	}
	if ((MODE_NOPRIVMSGS & oldmode->mode)
	    && !(MODE_NOPRIVMSGS & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'n';
	}
	if ((MODE_TOPICLIMIT & oldmode->mode)
	    && !(MODE_TOPICLIMIT & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 't';
	}
	if ((MODE_INVITEONLY & oldmode->mode)
	    && !(MODE_INVITEONLY & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'i';
	}
	if ((MODE_REGISTERED & oldmode->mode)
	    && !(MODE_REGISTERED & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'r';
	}
	if ((MODE_REGONLY & oldmode->mode) && !(MODE_REGONLY & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'R';
	}
	if ((MODE_OPERONLY & oldmode->mode) && !(MODE_OPERONLY & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'O';
	}
	if ((MODE_NOCOLOR & oldmode->mode) && !(MODE_NOCOLOR & mode.mode)) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'c';
	}

	if (oldmode->limit && !mode.limit) {
	    INSERTSIGN(-1, '-');
	    *mbuf++ = 'l';
	}
    }

    if (oldmode->key[0] && !mode.key[0]) {
	INSERTSIGN(-1, '-');
	*mbuf++ = 'k';
	ADD_PARA(oldmode->key);
	pargs++;
    }

    if (mode.limit && oldmode->limit != mode.limit) {
	INSERTSIGN(1, '+');
	*mbuf++ = 'l';
	sprintf(numeric, "%-15d", mode.limit);
	if ((s = strchr(numeric, ' ')))
	    *s = '\0';
	ADD_PARA(numeric);
	pargs++;
    }

    if (mode.key[0] && strcmp(oldmode->key, mode.key)) {
	INSERTSIGN(1, '+');
	*mbuf++ = 'k';
	ADD_PARA(mode.key);
	pargs++;
    }

    chptr->mode = mode;

    if (!keepourmodes) {	/* deop and devoice everyone! */
	what = 0;
	for (cm = chptr->members; cm; cm = cm->next) {
	    if (cm->flags & MODE_OWNER) {
		INSERTSIGN(-1, '-');
		*mbuf++ = 'u';
		ADD_PARA(cm->cptr->name);
		pargs++;
		if (pargs >= MAXMODEPARAMS) {
		    *mbuf = '\0';
		    parabuf[pbpos] = '\0';
		    sjoin_sendit(cptr, sptr, chptr, parv[0]);
		    mbuf = modebuf;
		    *mbuf = '\0';
		    pargs = pbpos = what = 0;
		}
		cm->flags &= ~MODE_OWNER;
	    }
	    if (cm->flags & MODE_PROTECT) {
		INSERTSIGN(-1, '-');
		*mbuf++ = 'a';
		ADD_PARA(cm->cptr->name);
		pargs++;
		if (pargs >= MAXMODEPARAMS) {
		    *mbuf = '\0';
		    parabuf[pbpos] = '\0';
		    sjoin_sendit(cptr, sptr, chptr, parv[0]);
		    mbuf = modebuf;
		    *mbuf = '\0';
		    pargs = pbpos = what = 0;
		}
		cm->flags &= ~MODE_PROTECT;
	    }
	    if (cm->flags & MODE_CHANOP) {
		INSERTSIGN(-1, '-');
		*mbuf++ = 'o';
		ADD_PARA(cm->cptr->name);
		pargs++;
		if (pargs >= MAXMODEPARAMS) {
		    *mbuf = '\0';
		    parabuf[pbpos] = '\0';
		    sjoin_sendit(cptr, sptr, chptr, parv[0]);
		    mbuf = modebuf;
		    *mbuf = '\0';
		    pargs = pbpos = what = 0;
		}
		cm->flags &= ~MODE_CHANOP;
	    }
	    if (cm->flags & MODE_VOICE) {
		INSERTSIGN(-1, '-');
		*mbuf++ = 'v';
		ADD_PARA(cm->cptr->name);
		pargs++;
		if (pargs >= MAXMODEPARAMS) {
		    *mbuf = '\0';
		    parabuf[pbpos] = '\0';
		    sjoin_sendit(cptr, sptr, chptr, parv[0]);
		    mbuf = modebuf;
		    *mbuf = '\0';
		    pargs = pbpos = what = 0;
		}
		cm->flags &= ~MODE_VOICE;
	    }
	}
	sendto_channel_butserv(chptr, &me,
			       ":%s NOTICE %s :*** Notice -- TS for %s changed from %ld to %ld",
			       me.name, chptr->chname, chptr->chname, oldts,
			       newts);
    }

    if (mbuf != modebuf) {
	*mbuf = '\0';
	parabuf[pbpos] = '\0';
	sjoin_sendit(cptr, sptr, chptr, parv[0]);
    }

    *modebuf = '\0';
    parabuf[0] = '\0';
    if (parv[3][0] != '0' && keepnewmodes)
	channel_modes(sptr, modebuf, parabuf, chptr);
    else {
	modebuf[0] = '0';
	modebuf[1] = '\0';
    }

    /*
     * We do this down below now, so we can send out for two sjoin formats.
     * * sprintf(t, ":%s SJOIN %ld %ld %s %s %s :", parv[0], tstosend, tstosend,
     * *                     parv[2], modebuf, parabuf);
     * * t += strlen(t);
     * * the pointer "t" has been removed and is now replaced with an
     * * index into sjbuf for faster appending
     */

    strcpy(keep_modebuf, modebuf);
    strcpy(keep_parabuf, parabuf);

    sjbufpos = 0;
    mbuf = modebuf;
    pbpos = 0;
    pargs = 0;
    *mbuf++ = '+';

    for (s = s0 = strtoken(&p, parv[args + 4], " "); s;
	 s = s0 = strtoken(&p, (char *) NULL, " ")) {
	fl = 0;
	if (*s == '.' || s[1] == '.')
	    fl |= MODE_OWNER;
	if (*s == '~' || s[1] == '~')
	    fl |= MODE_PROTECT;
	if (*s == '@' || s[1] == '@')
	    fl |= MODE_CHANOP;
	if (*s == '+' || s[1] == '+')
	    fl |= MODE_VOICE;
	if (!keepnewmodes) {
	    if (fl & MODE_CHANOP)
		fl = MODE_DEOPPED;
	    else
		fl = 0;
	}
	while (*s == '.' || *s == '~' || *s == '@' || *s == '+')
	    s++;
	if (!(acptr = find_chasing(sptr, s, NULL)))
	    continue;
	if (acptr->from != cptr)
	    continue;
	people++;
	if (!IsMember(acptr, chptr)) {
	    add_user_to_channel(chptr, acptr, fl);
	    sendto_channel_butserv(chptr, acptr, ":%s JOIN :%s", s, parv[2]);
	    sendto_service(SERVICE_SEE_JOINS, SERVICE_SEE_PREFIX,
			   ":%s JOIN :%s", parv[0], parv[2]);
	    sendto_service(SERVICE_SEE_JOINS | SERVICE_SEE_PREFIX, 0,
			   ":%s!%s@%s JOIN :%s", parv[0], acptr->user->username,
			   acptr->user->host, parv[2]);
	}
	if (keepnewmodes) {
	    ADD_SJBUF(s0);
	} else {
	    ADD_SJBUF(s);
	}
	if (fl & MODE_OWNER) {
	    *mbuf++ = 'u';
	    ADD_PARA(s);
	    pargs++;
	    if (pargs >= MAXMODEPARAMS) {
		*mbuf = '\0';
		parabuf[pbpos] = '\0';
		sjoin_sendit(cptr, sptr, chptr, parv[0]);
		mbuf = modebuf;
		*mbuf++ = '+';
		pargs = pbpos = 0;
	    }
	}
	if (fl & MODE_PROTECT) {
	    *mbuf++ = 'a';
	    ADD_PARA(s);
	    pargs++;
	    if (pargs >= MAXMODEPARAMS) {
		*mbuf = '\0';
		parabuf[pbpos] = '\0';
		sjoin_sendit(cptr, sptr, chptr, parv[0]);
		mbuf = modebuf;
		*mbuf++ = '+';
		pargs = pbpos = 0;
	    }
	}
	if (fl & MODE_CHANOP) {
	    *mbuf++ = 'o';
	    ADD_PARA(s);
	    pargs++;
	    if (pargs >= MAXMODEPARAMS) {
		*mbuf = '\0';
		parabuf[pbpos] = '\0';
		sjoin_sendit(cptr, sptr, chptr, parv[0]);
		mbuf = modebuf;
		*mbuf++ = '+';
		pargs = pbpos = 0;
	    }
	}
	if (fl & MODE_VOICE) {
	    *mbuf++ = 'v';
	    ADD_PARA(s);
	    pargs++;
	    if (pargs >= MAXMODEPARAMS) {
		*mbuf = '\0';
		parabuf[pbpos] = '\0';
		sjoin_sendit(cptr, sptr, chptr, parv[0]);
		mbuf = modebuf;
		*mbuf++ = '+';
		pargs = pbpos = 0;
	    }
	}
    }

    parabuf[pbpos] = '\0';
    *mbuf = '\0';
    if (pargs)
	sjoin_sendit(cptr, sptr, chptr, parv[0]);
    if (people && !(*chptr->chname == '&')) {
	sjbuf[sjbufpos] = '\0';
	if (keep_parabuf[0] != '\0') {
	    sendto_flag_serv_butone(cptr, CAPAB_NSJOIN, 0, newSJOINFmt,
				    parv[0], tstosend, parv[2], keep_modebuf,
				    keep_parabuf, sjbuf);
	    sendto_flag_serv_butone(cptr, 0, CAPAB_NSJOIN, oldSJOINFmt, parv[0],
				    tstosend, tstosend, parv[2], keep_modebuf,
				    keep_parabuf, sjbuf);
	} else {
	    sendto_flag_serv_butone(cptr, CAPAB_NSJOIN, 0, newSJOINFmtNP,
				    parv[0], tstosend, parv[2], keep_modebuf,
				    sjbuf);
	    sendto_flag_serv_butone(cptr, 0, CAPAB_NSJOIN, oldSJOINFmtNP,
				    parv[0], tstosend, tstosend, parv[2],
				    keep_modebuf, sjbuf);
	}
    }
    return 0;
}

#undef INSERTSIGN
#undef ADD_PARA
#undef ADD_SJBUF
