/************************************************************************
 *   IRC - Internet Relay Chat, src/m_capab.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "msg.h"
#include "numeric.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
# include <strings.h>
# undef strchr
# define strchr index
#endif

static char buf[BUFSIZE], buf2[BUFSIZE];

#ifdef USE_FAKEHOST
extern int do_user(char *, aClient *, aClient *, char *, char *,
		   char *, char *, unsigned long, unsigned int, char *);
#else
extern int do_user(char *, aClient *, aClient *, char *, char *,
		   char *, unsigned long, unsigned int, char *);
#endif

/*
 * * m_user * parv[0] = sender prefix *       parv[1] = username
 * (login name, account) *      parv[2] = client host name (used only
 * from other servers) *        parv[3] = server host name (used only
 * from other servers) *        parv[4] = users real name info
 */
int m_user(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *username, *host, *server, *realname;
    aConfItem *aconf;

    if (parc > 2 && (username = (char *) strchr(parv[1], '@')))
	*username = '\0';
    if (parc < 5 || *parv[1] == '\0' || *parv[2] == '\0' || *parv[3] == '\0' ||
	*parv[4] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "USER");
	if (IsServer(cptr))
	    sendto_realops("bad USER param count for %s from %s",
			   parv[0], get_client_name(cptr, FALSE));
	else
	    return 0;
    }
    /*
     * Copy parameters into better documenting variables
     */
    username = (parc < 2 || BadPtr(parv[1])) ? "<bad-boy>" : parv[1];
    host = (parc < 3 || BadPtr(parv[2])) ? "<nohost>" : parv[2];
    server = (parc < 4 || BadPtr(parv[3])) ? "<noserver>" : parv[3];
    realname = (parc < 5 || BadPtr(parv[4])) ? "<bad-realname>" : parv[4];
    if ((aconf = find_conf_name(realname, CONF_GCOS))) {
	return exit_client(cptr, sptr, sptr, BadPtr(aconf->passwd) ?
			   "Bad GCOS: Reason unspecified" : aconf->passwd);
    }
#ifdef USE_FAKEHOST
    return do_user(parv[0], cptr, sptr, username, host, NULL, server, 0, 0,
		   realname);
#else
    return do_user(parv[0], cptr, sptr, username, host, server, 0, 0, realname);
#endif
}

/*
 * * m_quit * parv[0] = sender prefix *       parv[1] = comment
 */
int m_quit(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *reason = (parc > 1 && parv[1]) ? parv[1] : cptr->name;
    char comment[TOPICLEN + 1];

    if ((parc > 1) && (check_forbidden_words(reason)))
	strcpy(reason, DEFAULTQUITMSG);

    if (sptr->user)
	sendto_service(SERVICE_SEE_QUITS | SERVICE_SEE_PREFIX, 0,
		       ":%s!%s@%s QUIT :%s", sptr->name,
		       sptr->user->username, sptr->user->host, reason);
    else
	sendto_service(SERVICE_SEE_QUITS, 0, ":%s QUIT :%s", sptr->name,
		       reason);

    sptr->flags |= FLAGS_NORMALEX;
    if (!IsServer(cptr)) {
	strcpy(comment, "Quit: ");
	strncpy(comment + 6, reason, TOPICLEN - 6);
	comment[TOPICLEN] = 0;
	return exit_client(cptr, sptr, sptr, comment);
    } else
	return exit_client(cptr, sptr, sptr, reason);
}

/***********************************************************************
         * m_away() - Added 14 Dec 1988 by jto.
 *            Not currently really working, I don't like this
 *            call at all...
 *            
 *            ...trying to make it work. I don't like it either,
 *            but perhaps it's worth the load it causes to net. 
 *            This requires flooding of the whole net like NICK,
 *            USER, MODE, etc messages...  --msa
 *            
 *            Added FLUD-style limiting for those lame scripts out there.
 ***********************************************************************/
/*
 * * m_away * parv[0] = sender prefix *       parv[1] = away message
 */

int m_away(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *away, *awy2 = parv[1];

    /*
     * make sure the user exists
     */
    if (!(sptr->user)) {
	sendto_realops_lev(DEBUG_LEV,
			   "Got AWAY from nil user, from %s (%s)\n", cptr->name,
			   sptr->name);
	return 0;
    }

    away = sptr->user->away;

#ifdef NO_AWAY_FLUD
    if (MyClient(sptr)) {
	if ((sptr->alas + MAX_AWAY_TIME) < NOW)
	    sptr->acount = 0;
	sptr->alas = NOW;
	sptr->acount++;
    }
#endif

    if (parc < 2 || !*awy2) {
	if (away) {
	    MyFree(away);
	    sptr->user->away = NULL;
	    sendto_serv_butone(cptr, ":%s AWAY", parv[0]);
	}

	if (MyConnect(sptr))
	    sendto_one(sptr, rpl_str(RPL_UNAWAY), me.name, parv[0]);
	return 0;
    }
#ifdef NO_AWAY_FLUD
    if (MyClient(sptr) && (sptr->acount > MAX_AWAY_COUNT)
	&& !IsAnOper(sptr)) {
	sendto_one(sptr, err_str(ERR_TOOMANYAWAY), me.name, parv[0]);
	return 0;
    }
#endif
    if (strlen(awy2) > (size_t) TOPICLEN)
	awy2[TOPICLEN] = '\0';
    /*
     * some lamers scripts continually do a /away, hence making a lot of
     * unnecessary traffic. *sigh* so... as comstud has done, I've
     * commented out this sendto_serv_butone() call -Dianora
     * readded because of anti-flud stuffs -epi
     */

    sendto_serv_butone(cptr, ":%s AWAY :%s ", parv[0], parv[1]);

    if (away)
	MyFree(away);

    away = (char *) MyMalloc(strlen(awy2) + 1);
    strcpy(away, awy2);

    sptr->user->away = away;

    if (MyConnect(sptr))
	sendto_one(sptr, rpl_str(RPL_NOWAWAY), me.name, parv[0]);
    return 0;
}

/*
 * * m_oper * parv[0] = sender prefix *       parv[1] = oper name *
 * parv[2] = oper password
 */
int m_oper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;
    char *name, *password, *encr;

#ifdef CRYPT_OPER_PASSWORD
    extern char *crypt();

#endif

    name = parc > 1 ? parv[1] : (char *) NULL;
    password = parc > 2 ? parv[2] : (char *) NULL;

    if (!IsServer(cptr) && (BadPtr(name) || BadPtr(password))) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "OPER");
	return 0;
    }

    if ((IsServer(cptr) || IsMe(cptr)) && !IsOper(sptr)) {
#ifdef DEFAULT_HELP_MODE
	sptr->umode |= UMODE_o;
	sptr->umode |= UMODE_h;
	sendto_serv_butone(cptr, ":%s MODE %s :+oh", parv[0], parv[0]);
#else
	sptr->umode |= UMODE_o;
	sendto_serv_butone(cptr, ":%s MODE %s :+o", parv[0], parv[0]);
#endif
	Count.oper++;
	if (IsMe(cptr))
	    sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
	return 0;
    } else if (IsAnOper(sptr)) {
	if (MyConnect(sptr))
	    sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
	return 0;
    }
    if (!(sptr->user && sptr->user->real_oper_host)) {
	if (!
	    (aconf =
	     find_conf_exact(name, sptr->username, sptr->sockhost, CONF_OPS))
&& !(aconf = find_conf_exact(name, sptr->username, cptr->hostip, CONF_OPS))) {
	    sendto_one(sptr, err_str(ERR_NOOPERHOST), me.name, parv[0]);
	    sendto_realops("Failed OPER attempt by %s (%s@%s)", parv[0],
			   sptr->user->username, sptr->user->host);
	    return 0;
	}
    } else {
	if (!(aconf =
	      find_conf_exact(name, sptr->user->real_oper_username,
			      sptr->user->real_oper_host, CONF_OPS))
	    && !(aconf =
		 find_conf_exact(name, sptr->user->real_oper_username,
				 sptr->user->real_oper_ip, CONF_OPS))) {
	    sendto_one(sptr, err_str(ERR_NOOPERHOST), me.name, parv[0]);
	    sendto_realops("Failed OPER attempt by %s (%s@%s)", parv[0],
			   sptr->user->username, sptr->user->host);
	    return 0;
	}
    }
#ifdef CRYPT_OPER_PASSWORD
    /*
     * use first two chars of the password they send in as salt
     * passwd may be NULL pointer. Head it off at the pass...  
     */
    if (password && *aconf->passwd)
	encr = crypt(password, aconf->passwd);
    else
	encr = "";
#else
    encr = password;
#endif /* CRYPT_OPER_PASSWORD */

    if ((aconf->status & CONF_OPS) && StrEq(encr, aconf->passwd) &&
	!attach_conf(sptr, aconf)) {
    int old = (sptr->umode & ALL_UMODES);
    char *s;

	s = strchr(aconf->host, '@');
	if (s == (char *) NULL) {
	    sendto_one(sptr, err_str(ERR_NOOPERHOST), me.name, parv[0]);
	    sendto_realops("corrupt aconf->host = [%s]", aconf->host);
	    return 0;
	}
	*s++ = '\0';
	if (!(aconf->port & OFLAG_ISGLOBAL))
	    SetLocOp(sptr);
	else
	    SetOper(sptr);
#ifdef DEFAULT_HELP_MODE
	sptr->umode |= (UMODE_g | UMODE_w | UMODE_n | UMODE_h);
#else
	sptr->umode |= (UMODE_g | UMODE_w | UMODE_n);
#endif
#ifdef USE_FAKEHOST
	if (sptr->user->real_oper_host)
	    sptr->umode &= ~UMODE_x;
#endif
	sptr->oflag = aconf->port;
	if (OPIsSAdmin(sptr))
	    sptr->umode |= UMODE_a;
	if (OPIsAdmin(sptr))
	    sptr->umode |= UMODE_A;
	Count.oper++;
	*--s = '@';
	addto_fdlist(sptr->fd, &oper_fdlist);
	sendto_ops
	    ("%s (%s@%s) is now operator (%c) using host [%s] with uid [%s]",
	     parv[0], sptr->user->username, sptr->sockhost,
	     IsOper(sptr) ? 'O' : 'o', aconf->host, aconf->name);

	sendto_service(SERVICE_SEE_OPERS, SERVICE_SEE_PREFIX, ":%s OPER :%s",
		       parv[0], parv[0]);
	sendto_service(SERVICE_SEE_OPERS | SERVICE_SEE_PREFIX, 0,
		       ":%s!%s@%s OPER :%s", parv[0], sptr->user->username,
		       sptr->user->host, parv[0]);
	send_umode_out(cptr, sptr, old);
	sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
	sptr->pingval = get_client_ping(sptr);
#if !defined(CRYPT_OPER_PASSWORD) && (defined(FNAME_OPERLOG) ||\
    (defined(USE_SYSLOG) && defined(SYSLOG_OPER)))
	encr = "";
#endif
#if defined(USE_SYSLOG) && defined(SYSLOG_OPER)
	syslog(LOG_INFO, "OPER (%s) (%s) by (%s!%s@%s)",
	       name, encr, parv[0], sptr->user->username, sptr->sockhost);
#endif
#if defined(FNAME_OPERLOG)
	{
    int logfile;

	    /*
	     * This conditional makes the logfile active only after it's
	     * been created - thus logging can be turned off by removing
	     * the file.
	     * 
	     * stop NFS hangs...most systems should be able to open a file in
	     * 3 seconds. -avalon (curtesy of wumpus)
	     */
	    (void) alarm(3);
	    if (IsPerson(sptr) &&
		(logfile = open(FNAME_OPERLOG, O_WRONLY | O_APPEND)) != -1) {
		(void) alarm(0);
		(void) ircsprintf(buf, "%s OPER (%s) (%s) by (%s!%s@%s)\n",
				  myctime(timeofday), name, encr,
				  parv[0], sptr->user->username,
				  sptr->sockhost);
		(void) alarm(3);
		(void) write(logfile, buf, strlen(buf));
		(void) alarm(0);
		(void) close(logfile);
	    }
	    (void) alarm(0);
	    /*
	     * Modification by pjg
	     */
	}
#endif
    } else {
	(void) detach_conf(sptr, aconf);
	sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
#ifdef FAILED_OPER_NOTICE
	sendto_realops("Failed OPER attempt by %s (%s@%s)",
		       parv[0], sptr->user->username, sptr->sockhost);
#endif
    }
    return 0;
}

/*
 * m_userhost added by Darren Reed 13/8/91 to aid clients and reduce
 * the need for complicated requests like WHOIS. It returns user/host
 * information only (no spurious AWAY labels or channels).
 */
int m_userhost(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *p = NULL;
    aClient *acptr;
    char *s;
    int i, len;

    if (parc > 2)
	(void) m_userhost(cptr, sptr, parc - 1, parv + 1);

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "USERHOST");
	return 0;
    }

    (void) ircsprintf(buf, rpl_str(RPL_USERHOST), me.name, parv[0]);
    len = strlen(buf);
    *buf2 = '\0';

    for (i = 5, s = strtoken(&p, parv[1], " "); i && s;
	 s = strtoken(&p, (char *) NULL, " "), i--)
	if ((acptr = find_person(s, NULL))) {
	    if (*buf2)
		(void) strcat(buf, " ");
	    (void) ircsprintf(buf2, "%s%s=%c%s@%s",
			      acptr->name,
			      IsAnOper(acptr) ? "*" : "",
			      (acptr->user->away) ? '-' : '+',
			      acptr->user->username,
#ifdef USE_FAKEHOST
			      IsFake(acptr) ? acptr->user->fakehost : acptr->
			      user->host
#else
			      acptr->user->host
#endif
		);
	    (void) strncat(buf, buf2, sizeof(buf) - len);
	    len += strlen(buf2);
	}
    sendto_one(sptr, "%s", buf);
    return 0;
}

/*
 * m_ison added by Darren Reed 13/8/91 to act as an efficent user
 * indicator with respect to cpu/bandwidth used. Implemented for NOTIFY
 * feature in clients. Designed to reduce number of whois requests. Can
 * process nicknames in batches as long as the maximum buffer length.  
 * 
 * format: ISON :nicklist
 */

int m_ison(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *s, **pav = parv;
    char *p = (char *) NULL;
    int len;
    int len2;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ISON");
	return 0;
    }

    (void) ircsprintf(buf, rpl_str(RPL_ISON), me.name, *parv);
    len = strlen(buf);
    if (!IsOper(cptr))
	cptr->priority += 20;	/*
				 * this keeps it from moving to 'busy'
				 * * list
				 */
    for (s = strtoken(&p, *++pav, " "); s; s = strtoken(&p, (char *) NULL, " "))
	if ((acptr = find_person(s, NULL))) {
	    len2 = strlen(acptr->name);
	    if ((len + len2 + 5) < sizeof(buf)) {
		(void) strcat(buf, acptr->name);
		len += len2;
		(void) strcat(buf, " ");
		len++;
	    } else
		break;
	}
    sendto_one(sptr, "%s", buf);
    return 0;
}
