/************************************************************************
 *   IRC - Internet Relay Chat, src/s_serv.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: s_query.c,v 1.18 2001/12/05 18:31:50 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "zlink.h"

#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
# include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
  /*
   * older unices don't have strchr/strrchr .. help them out 
   */
# include <strings.h>
# undef strchr
# define strchr index
#endif
#include "dich_conf.h"
#include "fdlist.h"
#include "throttle.h"

extern void report_conf_links(aClient *, aConfList *, int, char);
extern void show_opers(aClient *, char *);
extern void show_servers(aClient *, char *);
extern void count_memory(aClient *, char *);

/*
 * Local function prototypes 
 */

int send_motd(aClient *, aClient *, int, char **);
void read_motd(char *);

#ifdef SHORT_MOTD
void read_shortmotd(char *);
#endif

char motd_last_changed_date[MAX_DATE_STRING];	/* enough room for date */

/* 
 * * hunt_server *
 * 
 *      Do the basic thing in delivering the message (command) *
 * across the relays to the specific server (server) for *
 * actions. *
 * 
 *      Note:   The command is a format string and *MUST* be *  
 * f prefixed style (e.g. ":%s COMMAND %s ..."). *              Command
 * can have only max 8 parameters. *
 * 
 *      server  parv[server] is the parameter identifying the *
 * arget server. *
 *
 *      *WARNING* *             parv[server] is replaced with the
 * pointer to the *             real servername from the matched client
 * (I'm lazy *          now --msa). *
 *
 *      returns: (see #defines)
 */
int hunt_server(aClient *cptr, aClient *sptr, char *command, int server,
		int parc, char *parv[])
{
    aClient *acptr;
    int wilds;
    if (parc <= server || BadPtr(parv[server]) ||
	match(me.name, parv[server]) == 0 || match(parv[server], me.name) == 0)
	return (HUNTED_ISME);
    /*
     * * These are to pickup matches that would cause the following *
     * message to go in the wrong direction while doing quick fast * 
     * non-matching lookups.
     */

    if ((acptr = find_client(parv[server], NULL)))
	if (acptr->from == sptr->from && !MyConnect(acptr))
	    acptr = NULL;
    if (!acptr && (acptr = find_server(parv[server], NULL)))
	if (acptr->from == sptr->from && !MyConnect(acptr))
	    acptr = NULL;

    (void) collapse(parv[server]);
    wilds = (strchr(parv[server], '?') || strchr(parv[server], '*'));
    /*
     * Again, if there are no wild cards involved in the server name,
     * use the hash lookup - Dianora
     */

    if (!acptr) {
	if (!wilds) {
	    acptr = find_name(parv[server], (aClient *) NULL);
	    if (!acptr || !IsRegistered(acptr) || !IsServer(acptr)) {
		sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0],
			   parv[server]);
		return (HUNTED_NOSUCH);
	    }
	} else {
	    for (acptr = client; (acptr = next_client(acptr, parv[server]));
		 acptr = acptr->next) {
		if (acptr->from == sptr->from && !MyConnect(acptr))
		    continue;
		/*
		 * Fix to prevent looping in case the parameter for some
		 * reason happens to match someone from the from link --jto
		 */
		if (IsRegistered(acptr) && (acptr != cptr))
		    break;
	    }
	}
    }

    if (acptr) {
	if (IsMe(acptr) || MyClient(acptr))
	    return HUNTED_ISME;
	if (match(acptr->name, parv[server]))
	    parv[server] = acptr->name;
	sendto_one(acptr, command, parv[0],
		   parv[1], parv[2], parv[3], parv[4], parv[5], parv[6],
		   parv[7], parv[8]);
	return (HUNTED_PASS);
    }
    sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[server]);
    return (HUNTED_NOSUCH);
}

/*
 * * m_version 
 *      parv[0] = sender prefix 
 *      parv[1] = remote server
 */
int m_version(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    extern char serveropts[];

    if (hunt_server(cptr, sptr, ":%s VERSION :%s", 1, parc, parv) ==
	HUNTED_ISME)
	sendto_one(sptr, rpl_str(RPL_VERSION), me.name, parv[0], version,
		   debugmode, me.name, serveropts);
    return 0;
}

/*
 * * m_info 
 *      parv[0] = sender prefix 
 *      parv[1] = servername
 */
int m_info(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char **text = infotext;
    char outstr[241];

    static time_t last_used = 0L;

    if (hunt_server(cptr, sptr, ":%s INFO :%s", 1, parc, parv) == HUNTED_ISME) {
	if (IsPerson(sptr))
	    sendto_realops_lev(SPY_LEV,
			       "INFO requested by %s (%s@%s) [%s]",
			       sptr->name, sptr->user->username,
			       sptr->user->host, sptr->user->server);

	if (!IsAnOper(sptr)) {
	    if (!MyConnect(sptr))
		return 0;
	    if ((last_used + MOTD_WAIT) > NOW)
		return 0;
	    else
		last_used = NOW;
	}
	while (*text)
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], *text++);

	sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], "");

	/*
	 * I am -definately- going to come up with a replacement for this! -Rak 
	 */

	if (IsAnOper(sptr)) {
#ifdef ANTI_NICK_FLOOD
	    strcpy(outstr, " ANTI_NICK_FLOOD=1");
#else
	    strcpy(outstr, " ANTI_NICK_FLOOD=0");
#endif
#ifdef ANTI_SPAMBOT
	    strcat(outstr, " ANTI_SPAMBOT=1");
#else
	    strcat(outstr, " ANTI_SPAMBOT=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef CLIENT_COUNT
	    strcpy(outstr, " CLIENT_COUNT=1");
#else
	    strcpy(outstr, " CLIENT_COUNT=0");
#endif
#ifdef CLIENT_FLOOD
	    strcat(outstr, " CLIENT_FLOOD=1");
#else
	    strcat(outstr, " CLIENT_FLOOD=0");
#endif
#ifdef CLIENT_SERVER
	    strcat(outstr, " CLIENT_SERVER=1");
#else
	    strcat(outstr, " CLIENT_SERVER=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef ZLINES_IN_KPATH
	    strcat(outstr, " ZLINES_IN_KPATH=1");
#else
	    strcat(outstr, " ZLINES_IN_KPATH=0");
#endif
#ifdef DNS_DEBUG
	    strcat(outstr, " DNS_DEBUG=1");
#else
	    strcat(outstr, " DNS_DEBUG=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef DO_IDENTD
	    strcpy(outstr, " DO_IDENTD=1");
#else
	    strcpy(outstr, " DO_IDENTD=0");
#endif
#ifdef E_LINES_OPER_ONLY
	    strcat(outstr, " E_LINES_OPER_ONLY=1");
#else
	    strcat(outstr, " E_LINES_OPER_ONLY=0");
#endif
#ifdef FAILED_OPER_NOTICE
	    strcat(outstr, " FAILED_OPER_NOTICE=1");
#else
	    strcat(outstr, " FAILED_OPER_NOTICE=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef FLUD
	    strcpy(outstr, " FLUD=1");
#else
	    strcpy(outstr, " FLUD=0");
#endif
#ifdef SHORT_MOTD
	    strcat(outstr, " SHORT_MOTD=1");
#else
	    strcat(outstr, " SHORT_MOTD=0");
#endif
#ifdef F_LINES_OPER_ONLY
	    strcat(outstr, " F_LINES_OPER_ONLY=1");
#else
	    strcat(outstr, " F_LINES_OPER_ONLY=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef HIGHEST_CONNECTION
	    strcpy(outstr, " HIGHEST_CONNECTION=1");
#else
	    strcpy(outstr, " HIGHEST_CONNECTION=0");
#endif
#ifdef HUB
	    strcat(outstr, " HUB=1");
#else
	    strcat(outstr, " HUB=0");
#endif
#ifdef IDENTD_COMPLAIN
	    strcat(outstr, " IDENTD_COMPLAIN=1");
#else
	    strcat(outstr, " IDENTD_COMPLAIN=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef IGNORE_FIRST_CHAR
	    strcpy(outstr, " IGNORE_FIRST_CHAR=1");
#else
	    strcpy(outstr, " IGNORE_FIRST_CHAR=0");
#endif
#ifdef KPATH
	    strcat(outstr, " KPATH=1");
#else
	    strcat(outstr, " KPATH=0");
#endif
#ifdef LOCKFILE
	    strcat(outstr, " LOCKFILE=1");
#else
	    strcat(outstr, " LOCKFILE=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef MAXBUFFERS
	    strcpy(outstr, " MAXBUFFERS=1");
#else
	    strcpy(outstr, " MAXBUFFERS=0");
#endif
#ifdef NON_REDUNDANT_KLINES
	    strcat(outstr, " NON_REDUNDANT_KLINES=1");
#else
	    strcat(outstr, " NON_REDUNDANT_KLINES=0");
#endif
#ifdef NO_CHANOPS_WHEN_SPLIT
	    strcat(outstr, " NO_CHANOPS_WHEN_SPLIT=1");
#else
	    strcat(outstr, " NO_CHANOPS_WHEN_SPLIT=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef NO_MIXED_CASE
	    strcat(outstr, " NO_MIXED_CASE=1");
#else
	    strcat(outstr, " NO_MIXED_CASE=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef NO_OPER_FLOOD
	    strcpy(outstr, " NO_OPER_FLOOD=1");
#else
	    strcpy(outstr, " NO_OPER_FLOOD=0");
#endif
#ifdef NO_PRIORITY
	    strcat(outstr, " NO_PRIORITY=1");
#else
	    strcat(outstr, " NO_PRIORITY=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef OLD_Y_LIMIT
	    strcpy(outstr, " OLD_Y_LIMIT=1");
#else
	    strcpy(outstr, " OLD_Y_LIMIT=0");
#endif
#ifdef RFC1035_ANAL
	    strcat(outstr, " RFC1035_ANAL=1");
#else
	    strcat(outstr, " RFC1035_ANAL=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
	    strcpy(outstr, "SEPARATE_QUOTE_KLINES_BY_DATE=1");
#else
	    strcpy(outstr, "SEPARATE_QUOTE_KLINES_BY_DATE=0");
#endif
#ifdef SHORT_MOTD
	    strcat(outstr, " SHORT_MOTD=1");
#else
	    strcat(outstr, " SHORT_MOTD=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef SHOW_HEADERS
	    strcpy(outstr, " SHOW_HEADERS=1");
#else
	    strcpy(outstr, " SHOW_HEADERS=0");
#endif
#ifdef SHOW_INVISIBLE_LUSERS
	    strcat(outstr, " SHOW_INVISIBLE_LUSERS=1");
#else
	    strcat(outstr, " SHOW_INVISIBLE_LUSERS=0");
#endif
#ifdef SHOW_UH
	    strcat(outstr, " SHOW_UH=1");
#else
	    strcat(outstr, " SHOW_UH=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef SUNDBE
	    strcat(outstr, " SUNDBE=1");
#else
	    strcat(outstr, " SUNDBE=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef USERNAMES_IN_TRACE
	    strcpy(outstr, " USERNAMES_IN_TRACE=1");
#else
	    strcpy(outstr, " USERNAMES_IN_TRACE=0");
#endif
#ifdef USE_FAST_FD_ISSET
	    strcat(outstr, " USE_FAST_FD_ISSET=1");
#else
	    strcat(outstr, " USE_FAST_FD_ISSET=0");
#endif
#ifdef USE_SYSLOG
	    strcat(outstr, " USE_SYSLOG=1");
#else
	    strcat(outstr, " USE_SYSLOG=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef WARN_NO_NLINE
	    strcpy(outstr, " WARN_NO_NLINE=1");
#else
	    strcpy(outstr, " WARN_NO_NLINE=0");
#endif
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);

#ifdef FLUD
	    ircsprintf(outstr, "FLUD_BLOCK=%d FLUD_NUM=%d FLUD_TIME=%d",
		       FLUD_BLOCK, FLUD_NUM, FLUD_TIME);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
#endif

	    ircsprintf(outstr, " HARD_FDLIMIT_=%d HYBRID_SOMAXCONN=%d",
		       HARD_FDLIMIT_, HYBRID_SOMAXCONN);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
	    ircsprintf(outstr,
		       " MOTD_WAIT=%d MOTD_MAX=%d MAXIMUM_LINKS=%d",
		       MOTD_WAIT, MOTD_MAX, MAXIMUM_LINKS);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
	    ircsprintf(outstr, " MAX_NICK_CHANGES=%d MAX_NICK_TIME=%d",
		       MAX_NICK_CHANGES, MAX_NICK_TIME);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
	    ircsprintf(outstr, " NICKNAMEHISTORYLENGTH=%d NOISY_HTM=%d",
		       NICKNAMEHISTORYLENGTH, NOISY_HTM);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
	    ircsprintf(outstr,
		       " MAXSENDQLENGTH=%d BUFFERPOOL=%d INIT_MAXCLIENTS=%d",
		       MAXSENDQLENGTH, BUFFERPOOL, INIT_MAXCLIENTS);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
	    ircsprintf(outstr, " TS_MAX_DELTA=%d TS_WARN_DELTA=%d",
		       TS_MAX_DELTA, TS_WARN_DELTA);
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], outstr);
	}

	sendto_one(sptr,
		   ":%s %d %s :Birth Date: %s, compile # %s",
		   me.name, RPL_INFO, parv[0], creation, generation);
	sendto_one(sptr, ":%s %d %s :On-line since %s",
		   me.name, RPL_INFO, parv[0], myctime(me.firsttime));
	sendto_one(sptr, rpl_str(RPL_ENDOFINFO), me.name, parv[0]);
    }

    return 0;
}

/*
 * * m_links 
 *      parv[0] = sender prefix 
 *      parv[1] = servername mask 
 * or 
 *      parv[0] = sender prefix 
 *      parv[1] = server to query 
 *      parv[2] = servername mask
 */
int m_links(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *mask;
    aClient *acptr;
    char clean_mask[(2 * HOSTLEN) + 1];
    char *s;
    char *d;
    char *tmpname = NULL;
    char out[HOSTLEN];
    int n;

    if (parc > 2) {
	if (hunt_server(cptr, sptr, ":%s LINKS %s :%s", 1, parc, parv) !=
	    HUNTED_ISME)
	    return 0;
	mask = parv[2];

    } else
	mask = parc < 2 ? NULL : parv[1];

    /*
     * reject non-local requests 
     */
    if (!IsAnOper(sptr) && !MyConnect(sptr))
	return 0;

    /*
     * * sigh* Before the kiddies find this new and exciting way of
     * * annoying opers, lets clean up what is sent to all opers
     * * -Dianora
     */

    if (mask) {			/* only necessary if there is a mask */
	s = mask;
	d = clean_mask;
	n = (2 * HOSTLEN) - 2;
	while (*s && n > 0) {
	    if ((unsigned char) *s < (unsigned char) ' ') {
		/* Is it a control character ? */
		*d++ = '^';
		*d++ = (char) ((unsigned char) *s + 0x40);
		/* turn it into a printable */
		s++;
		n -= 2;
	    } else if ((unsigned char) *s > (unsigned char) '~') {
		*d++ = '.';
		s++;
		n--;
	    } else {
		*d++ = *s++;
		n--;
	    }
	}
	*d = '\0';
    }
    if (MyConnect(sptr) && IsPerson(sptr))
	sendto_realops_lev(SPY_LEV,
			   "LINKS %s requested by %s (%s@%s) [%s]",
			   mask ? clean_mask : "all",
			   sptr->name, sptr->user->username, sptr->user->host,
			   sptr->user->server);

    tmpname = stealth_server(me.name, out);

    if (bootopt & BOOT_HIDELINKS) {
	sendto_one(sptr, rpl_str(RPL_LINKS),
		   me.name, parv[0], tmpname, tmpname, 0, me.info);
    } else {

	for (acptr = client, (void) collapse(mask); acptr; acptr = acptr->next) {
	    if (!IsServer(acptr) && !IsMe(acptr))
		continue;
	    if (!BadPtr(mask) && match(mask, acptr->name))
		continue;
#ifdef HIDEULINEDSERVS
	    if (!IsOper(sptr) && IsULine(acptr))
		continue;
#endif
	    sendto_one(sptr, rpl_str(RPL_LINKS),
		       me.name, parv[0], acptr->name, acptr->serv->up,
		       acptr->hopcount,
		       (acptr->info[0] ? acptr->info : "(Unknown Location)"));
	}
    }

    sendto_one(sptr, rpl_str(RPL_ENDOFLINKS), me.name, parv[0],
	       BadPtr(mask) ? "*" : clean_mask);
    return 0;
}

static int report_array[14][3] = {
    {CONF_CONNECT_SERVER, RPL_STATSCLINE, 'C'},
    {CONF_NOCONNECT_SERVER, RPL_STATSNLINE, 'N'},
    {CONF_GCOS, RPL_STATSGLINE, 'G'},
    {CONF_CLIENT, RPL_STATSILINE, 'I'},
    {CONF_KILL, RPL_STATSKLINE, 'K'},
    {CONF_OPERATOR, RPL_STATSOLINE, 'O'},
    {CONF_HUB, RPL_STATSHLINE, 'H'},
    {CONF_LOCOP, RPL_STATSOLINE, 'o'},
    {CONF_QUARANTINE, RPL_STATSQLINE, 'Q'},
    {CONF_ULINE, RPL_STATSULINE, 'U'},
    {CONF_JUPECHAN, RPL_STATSJLINE, 'J'},
    {CONF_SERVICE, RPL_STATSULINE, 'S'},
    {0, 0}
};

static void report_configured_links(aClient *sptr, int mask)
{
    static char null[] = "<NULL>";
    aConfItem *tmp;
    int *p, port;
    char c, *host, *pass, *name;

    for (tmp = conf; tmp; tmp = tmp->next)
	if (tmp->status & mask) {
	    for (p = &report_array[0][0]; *p; p += 3)
		if (*p & tmp->status)
		    break;
	    if (!*p)
		continue;
	    c = (char) *(p + 2);
	    host = BadPtr(tmp->host) ? null : tmp->host;
	    pass = BadPtr(tmp->passwd) ? null : tmp->passwd;
	    name = BadPtr(tmp->name) ? null : tmp->name;
	    port = (int) tmp->port;

	    /*
	     * block out IP addresses of U: lined servers in 
	     * * a C/N request - lucas/xpsycho
	     * * modified on request by taz, hide all
	     * * ips to other than routing staff or admins
	     * * -epi
	     * * yet again, make this just hide _all_ ips. how sad.
	     * * - lucas
	     */

	    if (tmp->status & CONF_SERVER_MASK)
		host = "*";

	    if (tmp->status == CONF_KILL)
		sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host,
			   name, pass);
	    else if (tmp->status & CONF_QUARANTINE)
		sendto_one(sptr, rpl_str(p[1]), me.name,
			   sptr->name, c, pass, name, port,
			   get_conf_class(tmp));
	    else if (tmp->status & CONF_JUPECHAN)
		sendto_one(sptr, rpl_str(p[1]), me.name,
			   sptr->name, c, pass, name, port,
			   get_conf_class(tmp));
	    else if (tmp->status & CONF_GCOS)
		sendto_one(sptr, rpl_str(p[1]), me.name,
			   sptr->name, c, pass, name, port,
			   get_conf_class(tmp));
	    else {
		if (!IsAnOper(sptr))
		    sendto_one(sptr, rpl_str(p[1]), me.name,
			       sptr->name, c, "*", name, port,
			       get_conf_class(tmp));
		else
		    sendto_one(sptr, rpl_str(p[1]), me.name,
			       sptr->name, c, host, name, port,
			       get_conf_class(tmp));
	    }
	}
    return;
}

/*
 * 
 * m_stats 
 *      parv[0] = sender prefix 
 *      parv[1] = statistics selector (defaults to Message frequency) 
 *      parv[2] = server name (current server defaulted, if omitted) 
 */
/*
 * 
 * m_stats/stats_conf 
 *    Report N/C-configuration lines from this server. This could 
 *    report other configuration lines too, but converting the 
 *    status back to "char" is a bit akward--not worth the code 
 *    it needs...
 * 
 *    Note:   The info is reported in the order the server uses 
 *    it--not reversed as in ircd.conf!
 */

int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static char Lformat[] = ":%s %d %s %s %u %u %u %u %u :%u %u %s";
    static char Sformat[] =
	":%s %d %s Name SendQ SendM SendBytes RcveM RcveBytes :Open_since Idle TS";

    struct Message *mptr;
    aClient *acptr;
    char stat = parc > 1 ? parv[1][0] : '\0';
    int i;
    int doall = 0, wilds = 0;
    char *name;
    time_t sincetime;

    static time_t last_used = 0L;

    if (hunt_server(cptr, sptr, ":%s STATS %s :%s", 2, parc, parv) !=
	HUNTED_ISME)
	return 0;

    if (!IsAnOper(sptr) && !IsULine(sptr)) {
	/*
	 * allow remote stats p l ? u 
	 */
	if (!((stat == 'p') || (stat == 'P') || (stat == '?')
	      || (stat == 'u') || (stat == 'l') || (stat == 'L'))
	    && !MyConnect(sptr))
	    return 0;

	if ((last_used + MOTD_WAIT) > NOW)
	    return 0;
	else
	    last_used = NOW;
    }

    if (parc > 2) {
	name = parv[2];
	if (!mycmp(name, me.name))
	    doall = 2;
	else if (match(name, me.name) == 0)
	    doall = 1;
	if (strchr(name, '*') || strchr(name, '?'))
	    wilds = 1;
    } else
	name = me.name;

    if (stat != (char) 0 && !IsULine(sptr))
	if (IsPerson(sptr))
	    sendto_realops_lev(SPY_LEV,
			       "STATS %c requested by %s (%s@%s) [%s]",
			       stat, sptr->name, sptr->user->username,
			       sptr->user->host, sptr->user->server);
    switch (stat) {
	case 'L':
	case 'l':
	    sendto_one(sptr, Sformat, me.name, RPL_STATSLINKINFO, parv[0]);
	    if ((parc > 2) && !(doall || wilds)) {	/* Single client lookup 
							 */
		if (!IsAnOper(sptr)) {
		    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,
			       parv[0]);
		    break;
		}
		if (!(acptr = find_person(name, NULL)))
		    break;
		/*
		 * sincetime might be greater than timeofday,
		 * store a new value here to avoid sending 
		 * negative since-times. -Rak
		 */
		sincetime =
		    (acptr->since > timeofday) ? 0 : timeofday - acptr->since;
		sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO, parv[0],
			   get_client_name(acptr, TRUE),
			   (int) DBufLength(&acptr->sendQ), (int) acptr->sendM,
			   (int) acptr->sendK, (int) acptr->receiveM,
			   (int) acptr->receiveK, timeofday - acptr->firsttime,
			   sincetime,
			   IsServer(acptr) ? (DoesTS(acptr) ? "TS" : "NoTS") :
			   "-");
	    } else {
		for (i = 0; i <= highest_fd; i++) {
		    if (!(acptr = local[i]))
			continue;
		    if (!IsServer(acptr))
			continue;	/* nothing but servers */
#ifdef HIDEULINEDSERVS
		    if (IsULine(acptr) && !IsAnOper(sptr))
			continue;
#endif
		    sincetime =
			(acptr->since >
			 timeofday) ? 0 : timeofday - acptr->since;
		    sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO,
			       parv[0], get_client_name(acptr, HIDEME),
			       (int) DBufLength(&acptr->sendQ),
			       (int) acptr->sendM, (int) acptr->sendK,
			       (int) acptr->receiveM, (int) acptr->receiveK,
			       timeofday - acptr->firsttime, sincetime,
			       IsServer(acptr) ? (DoesTS(acptr) ? "TS" : "NoTS")
			       : "-");
		}
	    }
	    break;
	case 'C':
	case 'c':
#ifdef HIDEULINEDSERVS
	    if (!IsAnOper(sptr))
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    else
#endif
		report_configured_links(sptr,
					CONF_CONNECT_SERVER |
					CONF_NOCONNECT_SERVER);
	    break;

	case 'D':
	    if (!IsAnOper(sptr))
		break;
	    report_conf_links(sptr, &ZList1, RPL_STATSZLINE, 'Z');
	    break;
	case 'd':
	    if (!IsAnOper(sptr))
		break;
	    report_szlines(sptr);
	    break;
	case 'E':
	    if (IsAnOper(sptr))
		report_temp_excludes(sptr);
	    break;
	case 'e':
#ifdef E_LINES_OPER_ONLY
	    if (!IsAnOper(sptr))
		break;
#endif
	    report_conf_links(sptr, &EList1, RPL_STATSELINE, 'e');
	    report_conf_links(sptr, &EList2, RPL_STATSELINE, 'e');
	    report_conf_links(sptr, &EList3, RPL_STATSELINE, 'e');
	    break;

	case 'F':
	case 'f':
#ifdef F_LINES_OPER_ONLY
	    if (!IsAnOper(sptr))
		break;
#endif
	    report_conf_links(sptr, &FList1, RPL_STATSFLINE, 'F');
	    report_conf_links(sptr, &FList2, RPL_STATSFLINE, 'F');
	    report_conf_links(sptr, &FList3, RPL_STATSFLINE, 'F');
	    break;

	case 'G':
	case 'g':
	    report_configured_links(sptr, CONF_GCOS);
	    break;

	case 'H':
	case 'h':
#ifdef HIDEULINEDSERVS
	    if (!IsOper(sptr))
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    else
#endif
		report_configured_links(sptr, CONF_HUB);
	    break;

	case 'I':
	case 'i':
	    report_configured_links(sptr, CONF_CLIENT);
	    break;

	case 'k':
	    if (IsAnOper(sptr))
		report_temp_klines(sptr);

	case 'K':
	    if (parc > 3)
		report_matching_host_klines(sptr, parv[3]);
	    else if (IsAnOper(sptr)) {
		report_conf_links(sptr, &KList1, RPL_STATSKLINE, 'K');
		report_conf_links(sptr, &KList2, RPL_STATSKLINE, 'K');
		report_conf_links(sptr, &KList3, RPL_STATSKLINE, 'K');
	    } else if (sptr->user)
		report_matching_host_klines(sptr, sptr->user->host);
	    break;

	case 'M':
	case 'm':
	    /*
	     * original behaviour was not to report the command, if
	     * the command hadn't been used. I'm going to always
	     * report the command instead -Dianora
	     */
	    for (mptr = msgtab; mptr->cmd; mptr++)
		/*
		 * * if (mptr->count) 
		 */
		sendto_one(sptr, rpl_str(RPL_STATSCOMMANDS),
			   me.name, parv[0], mptr->cmd, mptr->count,
			   mptr->bytes);
	    break;

	case 'N':
	case 'n':
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
		       "User Connects Today: ", Count.today);
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
		       "User Connects past week: ", Count.weekly);
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
		       "User Connects past month: ", Count.monthly);
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
		       "User Conects past year: ", Count.yearly);
	    break;
	case 'o':
	case 'O':
	    report_configured_links(sptr, CONF_OPS);
	    break;

	case 'p':
	case 'P':
	    show_opers(sptr, parv[0]);
	    break;

	case 'Q':
	case 'q':
	    report_configured_links(sptr, CONF_QUARANTINE);
	    break;

	case 'R':
	case 'r':
#ifdef DEBUGMODE
	    send_usage(sptr, parv[0]);
#endif
	    break;

	case 'J':
	case 'j':
	    report_configured_links(sptr, CONF_JUPECHAN);
	    break;

	case 'S':
	    if (!IsOper(sptr))
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    else
		report_configured_links(sptr, CONF_SERVICE);
	    break;

	case 's':
	    if (IsAnOper(sptr))
		list_scache(cptr, sptr, parc, parv);
	    else
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    break;

	case 'T':
	    if (!IsAnOper(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		break;
	    }
	    throttle_stats(sptr, parv[0]);
	    break;
	case 't':
	    if (!IsAnOper(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		break;
	    }
	    tstats(sptr, parv[0]);
	    break;

	case 'U':
#ifdef HIDEULINEDSERVS
	    if (!IsOper(sptr))
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    else
#endif

		report_configured_links(sptr, CONF_ULINE);
	    break;

	case 'u':
	    {
    time_t now;

		now = timeofday - me.since;
		sendto_one(sptr, rpl_str(RPL_STATSUPTIME), me.name, parv[0],
			   now / 86400, (now / 3600) % 24, (now / 60) % 60,
			   now % 60);
#ifdef HIGHEST_CONNECTION
		sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0],
			   max_connection_count, max_client_count);
#endif
		break;
	    }

	case 'v':
	case 'V':
	    show_servers(sptr, parv[0]);
	    break;

	case 'Y':
	case 'y':
	    report_classes(sptr);
	    break;

	case 'Z':
	case 'z':
	    if (IsAnOper(sptr))
		count_memory(sptr, parv[0]);
	    else
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    break;

	case '?':
	    serv_info(sptr, parv[0]);

	default:
	    stat = '*';
	    break;
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, parv[0], stat);
    return 0;
}

/*
 * * m_users 
 *        parv[0] = sender prefix 
 *        parv[1] = servername
 */
int m_users(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (hunt_server(cptr, sptr, ":%s USERS :%s", 1, parc, parv) == HUNTED_ISME) {
	/*
	 * No one uses this any more... so lets remap it..   -Taner 
	 */
	sendto_one(sptr, rpl_str(RPL_LOCALUSERS), me.name, parv[0], Count.local,
		   Count.max_loc);
	sendto_one(sptr, rpl_str(RPL_GLOBALUSERS), me.name, parv[0],
		   Count.total, Count.max_tot);
    }
    return 0;
}

/*
 * * m_help 
 * parv[0] = sender prefix
 */
int m_help(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;
    aMotd *helpfile_ptr;

    static time_t last_used = 0L;

    if (!IsAnOper(sptr)) {
	/*
	 * reject non local requests 
	 */
	if ((last_used + MOTD_WAIT) > NOW)
	    return 0;
	else
	    last_used = NOW;

    }

    if (!IsAnOper(sptr) || (helpfile == (aMotd *) NULL)) {
	for (i = 0; msgtab[i].cmd; i++)
	    sendto_one(sptr, ":%s NOTICE %s :%s", me.name, parv[0],
		       msgtab[i].cmd);
	return 0;
    }

    helpfile_ptr = helpfile;
    while (helpfile_ptr) {
	sendto_one(sptr, ":%s NOTICE %s :%s", me.name, parv[0],
		   helpfile_ptr->line);
	helpfile_ptr = helpfile_ptr->next;
    }

    return 0;
}

/*
 * parv[0] = sender parv[1] = host/server mask. 
 * parv[2] = server to query
 * 
 * 199970918 JRL hacked to ignore parv[1] completely and require parc > 3
 * to cause a force
 */
int m_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int send_lusers(aClient *, aClient *, int, char **);

    if (parc > 2) {
	if (hunt_server(cptr, sptr, ":%s LUSERS %s :%s", 2, parc, parv) !=
	    HUNTED_ISME)
	    return 0;
    }
    return send_lusers(cptr, sptr, parc, parv);
}

/*
 * send_lusers
 *     parv[0] = sender
 *     parv[1] = host/server mask.
 *     parv[2] = server to query
 */
int send_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#define LUSERS_CACHE_TIME 180
    static long last_time = 0;
    static int s_count = 0, c_count = 0, u_count = 0, i_count = 0;
    static int o_count = 0, m_client = 0, m_server = 0, m_ulined = 0;
    static int sv_count = 0, m_service = 0;
    int forced;
    aClient *acptr;

    forced = (IsAnOper(sptr) && (parc > 3));

    Count.unknown = 0;
    m_server = Count.myserver;
    m_ulined = Count.myulined;
    m_client = Count.local;
    m_service = Count.myservice;
    i_count = Count.invisi;
    u_count = Count.unknown;
    c_count = Count.total - Count.invisi;
    s_count = Count.server;
    sv_count = Count.service;
    o_count = Count.oper;
    if (forced || (timeofday > last_time + LUSERS_CACHE_TIME)) {
	last_time = timeofday;
	/*
	 * only recount if more than a second has passed since last request
	 * * use LUSERS_CACHE_TIME instead... 
	 */
	s_count = 0;
	c_count = 0;
	u_count = 0;
	i_count = 0;
	o_count = 0;
	m_client = 0;
	m_server = 0;
	m_ulined = 0;
	sv_count = 0;
	m_service = 0;

	for (acptr = client; acptr; acptr = acptr->next) {
	    switch (acptr->status) {
		case STAT_SERVER:
		    if (MyConnect(acptr)) {
			m_server++;
			if (IsULine(acptr))
			    m_ulined++;
		    }
		case STAT_ME:
		    s_count++;
		    break;
		case STAT_CLIENT:
		    if (IsAnOper(acptr))
			o_count++;
#ifdef	SHOW_INVISIBLE_LUSERS
		    if (MyConnect(acptr))
			m_client++;
		    if (!IsInvisible(acptr))
			c_count++;
		    else
			i_count++;
#else
		    if (MyConnect(acptr)) {
			if (IsInvisible(acptr)) {
			    if (IsAnOper(sptr))
				m_client++;
			} else
			    m_client++;
		    }
		    if (!IsInvisible(acptr))
			c_count++;
		    else
			i_count++;
#endif
		    break;
		case STAT_SERVICE:
		    if (MyConnect(acptr))
			m_service++;
		    sv_count++;
		    break;
		default:
		    u_count++;
		    break;
	    }
	}
	/*
	 * We only want to reassign the global counts if the recount time
	 * has expired, and NOT when it was forced, since someone may
	 * supply a mask which will only count part of the userbase
	 * -Taner
	 */
	if (!forced) {
	    if (m_server != Count.myserver) {
		sendto_realops_lev(DEBUG_LEV,
				   "Local server count off by %d",
				   Count.myserver - m_server);
		Count.myserver = m_server;
	    }
	    if (m_service != Count.myservice) {
		sendto_realops_lev(DEBUG_LEV,
				   "Local service count off by %d",
				   Count.myservice - m_service);
		Count.myservice = m_service;
	    }
	    if (m_ulined != Count.myulined) {
		sendto_realops_lev(DEBUG_LEV,
				   "Local ulinedserver count off by %d",
				   Count.myulined - m_ulined);
		Count.myulined = m_ulined;
	    }
	    if (s_count != Count.server) {
		sendto_realops_lev(DEBUG_LEV, "Server count off by %d",
				   Count.server - s_count);
		Count.server = s_count;
	    }
	    if (i_count != Count.invisi) {
		sendto_realops_lev(DEBUG_LEV,
				   "Invisible client count off by %d",
				   Count.invisi - i_count);
		Count.invisi = i_count;
	    }
	    if ((c_count + i_count) != Count.total) {
		sendto_realops_lev(DEBUG_LEV,
				   "Total client count off by %d",
				   Count.total - (c_count + i_count));
		Count.total = c_count + i_count;
	    }
	    if (m_client != Count.local) {
		sendto_realops_lev(DEBUG_LEV,
				   "Local client count off by %d",
				   Count.local - m_client);
		Count.local = m_client;
	    }
	    if (o_count != Count.oper) {
		sendto_realops_lev(DEBUG_LEV, "Oper count off by %d",
				   Count.oper - o_count);
		Count.oper = o_count;
	    }
	    Count.unknown = u_count;
	}			/* Complain & reset loop */
    }
    /*
     * Recount loop 
     */
#ifdef SAVE_MAXCLIENT_STATS
    if ((timeofday - last_stat_save) > SAVE_MAXCLIENT_STATS_TIME) {	/* save 
									 * stats 
									 */
    FILE *fp;

	last_stat_save = timeofday;

	fp = fopen(DPATH "/.maxclients", "w");
	if (fp != NULL) {
	    fprintf(fp, "%d %d %li %li %li %ld %ld %ld %ld", Count.max_loc,
		    Count.max_tot, Count.weekly, Count.monthly,
		    Count.yearly, Count.start, Count.week, Count.month,
		    Count.year);
	    fclose(fp);
	    sendto_realops_lev(DEBUG_LEV, "Saved maxclient statistics");
	}
    }
#endif

#ifndef	SHOW_INVISIBLE_LUSERS
    if (IsAnOper(sptr) && i_count)
#endif
	sendto_one(sptr, rpl_str(RPL_LUSERCLIENT), me.name, parv[0], c_count,
		   i_count, s_count);
#ifndef	SHOW_INVISIBLE_LUSERS
    else
	sendto_one(sptr,
		   ":%s %d %s :There are %d users on %d servers", me.name,
		   RPL_LUSERCLIENT, parv[0], c_count, s_count);
#endif
    if (sv_count && m_service)
	sendto_one(sptr,
		   ":%s %d %s :There are %d services online, I have %d services.",
		   me.name, RPL_SERVICE, parv[0], sv_count, m_service);
    else if (sv_count)
	sendto_one(sptr, ":%s %d %s :There are %d services online.",
		   me.name, RPL_SERVICE, parv[0], sv_count);
    if (o_count)
	sendto_one(sptr, rpl_str(RPL_LUSEROP), me.name, parv[0], o_count);
    if (u_count > 0)
	sendto_one(sptr, rpl_str(RPL_LUSERUNKNOWN), me.name, parv[0], u_count);

    /*
     * This should be ok 
     */
    if (Count.chan > 0)
	sendto_one(sptr, rpl_str(RPL_LUSERCHANNELS), me.name, parv[0],
		   Count.chan);
    sendto_one(sptr, rpl_str(RPL_LUSERME),
#ifdef HIDEULINEDSERVS
	       me.name, parv[0], m_client,
	       IsOper(sptr) ? m_server : m_server - m_ulined);
#else
	       me.name, parv[0], m_client, m_server);
#endif
#ifdef CLIENT_COUNT
    sendto_one(sptr, rpl_str(RPL_LOCALUSERS), me.name, parv[0], Count.local,
	       Count.max_loc);
    sendto_one(sptr, rpl_str(RPL_GLOBALUSERS), me.name, parv[0], Count.total,
	       Count.max_tot);
#else
# ifdef HIGHEST_CONNECTION
    sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0],
	       max_connection_count, max_client_count);
    if (m_client > max_client_count)
	max_client_count = m_client;
    if ((m_client + m_server) > max_connection_count) {
	max_connection_count = m_client + m_server;
	if (max_connection_count % 10 == 0)
	    sendto_ops("New highest connections: %d (%d clients)",
		       max_connection_count, max_client_count);
    }
# endif	/* HIGHEST_CONNECTION */
#endif /* CLIENT_COUNT */
    return 0;
}

/*
 * * m_time 
 *       parv[0] = sender prefix 
 *       parv[1] = servername
 */
int m_time(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (hunt_server(cptr, sptr, ":%s TIME :%s", 1, parc, parv) == HUNTED_ISME)
	sendto_one(sptr, rpl_str(RPL_TIME), me.name, parv[0], me.name,
		   date((long) 0));
    return 0;
}

/*
 * * m_admin 
 *        parv[0] = sender prefix 
 *        parv[1] = servername
 */
int m_admin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;

    if (hunt_server(cptr, sptr, ":%s ADMIN :%s", 1, parc, parv) != HUNTED_ISME)
	return 0;

    if (IsPerson(sptr))
	sendto_realops_lev(SPY_LEV, "ADMIN requested by %s (%s@%s) [%s]",
			   sptr->name, sptr->user->username, sptr->user->host,
			   sptr->user->server);

    if ((aconf = find_admin())) {
	sendto_one(sptr, rpl_str(RPL_ADMINME), me.name, parv[0], me.name);
	sendto_one(sptr, rpl_str(RPL_ADMINLOC1), me.name, parv[0],
		   aconf->host ? aconf->host : "");
	sendto_one(sptr, rpl_str(RPL_ADMINLOC2), me.name, parv[0],
		   aconf->passwd ? aconf->passwd : "");
	sendto_one(sptr, rpl_str(RPL_ADMINEMAIL), me.name, parv[0],
		   aconf->name ? aconf->name : "");
    } else
	sendto_one(sptr, err_str(ERR_NOADMININFO), me.name, parv[0], me.name);
    return 0;
}

/*
 * * m_trace 
 *        parv[0] = sender prefix 
 *        parv[1] = servername
 */
int m_trace(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;
    aClient *acptr = NULL;
    aClass *cltmp;
    char *tname;
    int doall, link_s[MAXCONNECTIONS], link_u[MAXCONNECTIONS];
    int cnt = 0, wilds = 0, dow = 0;

    tname = (parc > 1) ? parv[1] : me.name;

#ifdef HIDEULINEDSERVS
    if ((acptr = next_client_double(client, tname))) {
	if (!(IsAnOper(sptr)) && IsULine(acptr)) {
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	acptr = NULL;		/* shrug, we borrowed it, reset it just in case 
				 */
    }
#endif

    if (!IsAnOper(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc > 2)
	if (hunt_server(cptr, sptr, ":%s TRACE %s :%s", 2, parc, parv))
	    return 0;

    switch (hunt_server(cptr, sptr, ":%s TRACE :%s", 1, parc, parv)) {
	case HUNTED_PASS:	/* note: gets here only if parv[1] * * * * * *
				 * exists */
	    {
    aClient *ac2ptr = next_client_double(client, tname);

		if (ac2ptr)
		    sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name,
			       parv[0], version, debugmode, tname,
			       ac2ptr->from->name);
		else
		    sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name,
			       parv[0], version, debugmode, tname,
			       "ac2ptr_is_NULL!!");
		return 0;
	    }
	case HUNTED_ISME:
	    break;
	default:
	    return 0;
    }
    if (!IsAnOper(sptr)) {
	if (parv[1] && !strchr(parv[1], '.') && (strchr(parv[1], '*')
						 || strchr(parv[1], '?')))
	    /*
	     * bzzzt, no wildcard nicks for nonopers 
	     */
	{
	    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0],
		       parv[1]);
	    return 0;
	}
    }
    if (IsPerson(sptr))
	sendto_realops_lev(SPY_LEV, "TRACE requested by %s (%s@%s) [%s]",
			   sptr->name, sptr->user->username, sptr->user->host,
			   sptr->user->server);

    doall = (parv[1] && (parc > 1)) ? !match(tname, me.name) : TRUE;
    wilds = !parv[1] || strchr(tname, '*') || strchr(tname, '?');
    dow = wilds || doall;
    if (!IsAnOper(sptr) || !dow) {	/* non-oper traces must be full nicks */
	/*
	 * lets also do this for opers tracing nicks 
	 */
    char *name;
    int class;

	acptr = hash_find_client(tname, (aClient *) NULL);
	if (!acptr || !IsPerson(acptr)) {
	    /*
	     * this should only be reached if the matching
	     * target is this server 
	     */
	    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	    return 0;

	}
	name = get_client_name(acptr, FALSE);
	class = get_client_class(acptr);
	if (IsAnOper(acptr)) {
	    sendto_one(sptr, rpl_str(RPL_TRACEOPERATOR),
		       me.name, parv[0], class, name,
		       timeofday - acptr->lasttime);
	} else {
	    sendto_one(sptr, rpl_str(RPL_TRACEUSER),
		       me.name, parv[0], class, name,
		       timeofday - acptr->lasttime);
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	return 0;
    }
    if (dow && lifesux && !IsOper(sptr)) {
	sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
	return 0;
    }

    memset((char *) link_s, '\0', sizeof(link_s));
    memset((char *) link_u, '\0', sizeof(link_u));
    /*
     * Count up all the servers and clients in a downlink.
     */
    if (doall)
	for (acptr = client; acptr; acptr = acptr->next) {
	    if (IsPerson(acptr) && (!IsInvisible(acptr) || IsAnOper(sptr)))
		link_u[acptr->from->fd]++;
	    else if (IsServer(acptr))
#ifdef HIDEULINEDSERVS
		if (IsOper(sptr) || !IsULine(acptr))
#endif
		    link_s[acptr->from->fd]++;
	}

    /*
     * report all direct connections 
     */

    for (i = 0; i <= highest_fd; i++) {
    char *name;
    int class;

	if (!(acptr = local[i]))	/* Local Connection? */
	    continue;
#ifdef HIDEULINEDSERVS
	if (!IsOper(sptr) && IsULine(acptr))
	    continue;
#endif
	if (IsInvisible(acptr) && dow &&
	    !(MyConnect(sptr) && IsAnOper(sptr)) && !IsAnOper(acptr) &&
	    (acptr != sptr))
	    continue;
	if (!doall && wilds && match(tname, acptr->name))
	    continue;
	if (!dow && mycmp(tname, acptr->name))
	    continue;
	if (IsAnOper(sptr))
	    name = get_client_name(acptr, FALSE);
	else
	    name = get_client_name(acptr, HIDEME);
	class = get_client_class(acptr);

	switch (acptr->status) {
	    case STAT_CONNECTING:
		sendto_one(sptr, rpl_str(RPL_TRACECONNECTING), me.name, parv[0],
			   class, name);
		cnt++;
		break;
	    case STAT_HANDSHAKE:
		sendto_one(sptr, rpl_str(RPL_TRACEHANDSHAKE), me.name, parv[0],
			   class, name);
		cnt++;
		break;
	    case STAT_ME:
		break;
	    case STAT_UNKNOWN:
		/*
		 * added time -Taner 
		 */
		sendto_one(sptr, rpl_str(RPL_TRACEUNKNOWN),
			   me.name, parv[0], class, name,
			   acptr->firsttime ? timeofday -
			   acptr->firsttime : -1);
		cnt++;
		break;
	    case STAT_CLIENT:
		/*
		 * Only opers see users if there is a wildcard but
		 * anyone can see all the opers.
		 */
		if (((IsAnOper(sptr) && (MyClient(sptr))) ||
		     !(dow && IsInvisible(acptr)))
		    || !dow || IsAnOper(acptr)) {
		    if (IsAnOper(acptr))
			sendto_one(sptr,
				   rpl_str(RPL_TRACEOPERATOR),
				   me.name, parv[0], class, name,
				   timeofday - acptr->lasttime);
		}
		break;
	    case STAT_SERVER:
		sendto_one(sptr, rpl_str(RPL_TRACESERVER),
			   me.name, parv[0], class, link_s[i],
			   link_u[i], name,
			   *(acptr->serv->bynick) ? acptr->serv->
			   bynick : "*",
			   *(acptr->serv->byuser) ? acptr->serv->
			   byuser : "*",
			   *(acptr->serv->byhost) ? acptr->serv->byhost : me.
			   name);
		cnt++;
		break;
	    case STAT_SERVICE:
		sendto_one(sptr, rpl_str(RPL_TRACESERVICE), me.name,
			   parv[0], class, name, acptr->service->enable,
			   acptr->service->sflags, timeofday - acptr->lasttime,
			   acptr->info);
		cnt++;
		break;
	    case STAT_LOG:
		sendto_one(sptr, rpl_str(RPL_TRACELOG), me.name, parv[0],
			   LOGFILE, acptr->port);
		cnt++;
		break;
	    default:		/* 
				 * ...we actually shouldn't come
				 * * here... --msa 
				 */
		sendto_one(sptr, rpl_str(RPL_TRACENEWTYPE), me.name, parv[0],
			   name);
		cnt++;
		break;
	}
    }
    /*
     * Add these lines to summarize the above which can get rather long
     * and messy when done remotely - Avalon
     */
    if (!SendWallops(sptr) || !cnt) {
	if (cnt) {
	    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	    return 0;
	}
	/*
	 * let the user have some idea that its at the end of the trace 
	 */
	sendto_one(sptr, rpl_str(RPL_TRACESERVER),
		   me.name, parv[0], 0, link_s[me.fd],
		   link_u[me.fd], me.name, "*", "*", me.name,
		   timeofday - acptr->lasttime);
	sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	return 0;
    }
#ifdef HIDEULINEDSERVS
    if (IsOper(sptr))
#endif
	for (cltmp = FirstClass(); doall && cltmp; cltmp = NextClass(cltmp))
	    if (Links(cltmp) > 0)
    sendto_one(sptr, rpl_str(RPL_TRACECLASS), me.name, parv[0], Class (cltmp),
	       Links(cltmp));

    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
    return 0;
}

/*
 * * m_motd 
 *       parv[0] = sender prefix 
 *       parv[1] = servername
 */
int m_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static time_t last_used = 0L;

    if (hunt_server(cptr, sptr, ":%s MOTD :%s", 1, parc, parv) != HUNTED_ISME)
	return 0;
    if (!IsAnOper(sptr)) {
	if ((last_used + MOTD_WAIT) > NOW)
	    return 0;
	else
	    last_used = NOW;

    }
    if (IsPerson(sptr))
	sendto_realops_lev(SPY_LEV, "MOTD requested by %s (%s@%s) [%s]",
			   sptr->name, sptr->user->username, sptr->user->host,
			   sptr->user->server);
    send_motd(cptr, sptr, parc, parv);
    return 0;
}

/*
 ** send_motd
 **  parv[0] = sender prefix
 **  parv[1] = servername
 **
 ** This function split off so a server notice could be generated on a
 ** user requested motd, but not on each connecting client.
 ** -Dianora
 */
int send_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aMotd *temp;
    struct tm *tm;

    tm = motd_tm;
    if (motd == (aMotd *) NULL) {
	sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0]);
	return 0;
    }
    sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name);

    if (tm)
	sendto_one(sptr,
		   ":%s %d %s :- %d/%d/%d %d:%02d", me.name, RPL_MOTD,
		   parv[0], tm->tm_mday, tm->tm_mon + 1,
		   1900 + tm->tm_year, tm->tm_hour, tm->tm_min);

    temp = motd;
    while (temp) {
	sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0], temp->line);
	temp = temp->next;
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0]);
    return 0;
}

/*
 * read_motd() - From CoMSTuD, added Aug 29, 1996
 */
void read_motd(char *filename)
{
    aMotd *temp, *last;
    struct stat sb;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /*
     * Clear out the old MOTD 
     */

    while (motd) {
	temp = motd->next;
	MyFree(motd);
	motd = temp;
    }
    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;
    fstat(fd, &sb);
    motd_tm = localtime(&sb.st_mtime);
    last = (aMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!motd)
	    motd = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);

    if (motd_tm)
	(void) sprintf(motd_last_changed_date,
		       "%d/%d/%d %d:%02d",
		       motd_tm->tm_mday,
		       motd_tm->tm_mon + 1,
		       1900 + motd_tm->tm_year, motd_tm->tm_hour,
		       motd_tm->tm_min);
}

#ifdef SHORT_MOTD
void read_shortmotd(char *filename)
{
    aMotd *temp, *last;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /*
     * Clear out the old MOTD 
     */

    while (shortmotd) {
	temp = shortmotd->next;
	MyFree(shortmotd);
	shortmotd = temp;
    }
    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;

    last = (aMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!shortmotd)
	    shortmotd = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);
}
#endif

/*
 * read_help() - modified from from CoMSTuD's read_motd added Aug 29,
 * 1996 modifed  Aug 31 1997 - Dianora
 * 
 * Use the same idea for the oper helpfile
 */
void read_help(char *filename)
{
    aMotd *temp, *last;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /*
     * Clear out the old HELPFILE 
     */

    while (helpfile) {
	temp = helpfile->next;
	MyFree(helpfile);
	helpfile = temp;
    }

    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;

    last = (aMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!helpfile)
	    helpfile = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);
}
