/************************************************************************
 *   IRC - Internet Relay Chat, src/s_serv.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: s_serv.c,v 1.13 2001/12/05 18:31:50 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "zlink.h"

#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
# include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
  /*
   * older unices don't have strchr/strrchr .. help them out 
   */
# include <strings.h>
# undef strchr
# define strchr index
#endif
#include "dich_conf.h"
#include "fdlist.h"
#include "throttle.h"

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif

#ifdef HTM_LOCK_ON_NETBURST
extern int HTMLOCK;
#endif

/*
 * * m_functions execute protocol messages on this server: *
 * 
 * cptr: 
 ** always NON-NULL, pointing to a *LOCAL* client
 ** structure (with an open socket connected!). This 
 ** is the physical socket where the message originated (or
 ** which caused the m_function to be executed--some
 ** m_functions may call others...). 
 * 
 * sptr:
 ** the source of the message, defined by the
 ** prefix part of the message if present. If not or
 ** prefix not found, then sptr==cptr. 
 * 
 *      *Always* true (if 'parse' and others are working correct): 
 * 
 *      1)      sptr->from == cptr  (note: cptr->from == cptr) 
 * 
 *      2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr 
 * annot* be a local connection, unless it's actually
 * cptr!). 
 *
 * MyConnect(x) should probably  be defined as (x == x->from) --msa
 * 
 * parc:    
 ** number of variable parameter strings (if zero, 
 ** parv is allowed to be NULL)
 * 
 * parv:    
 ** a NULL terminated list of parameter pointers,
 *** parv[0], sender (prefix string), if not present his points to 
 *** an empty string.
 *
 ** [parc-1]:
 *** pointers to additional parameters 
 *** parv[parc] == NULL, *always* 
 * 
 * note:   it is guaranteed that parv[0]..parv[parc-1] are all
 *         non-NULL pointers.
 */

/*
 * m_squit
 * there are two types of squits: those going downstream (to the target server)
 * and those going back upstream (from the target server).
 * previously, it wasn't necessary to distinguish between these two types of 
 * squits because they neatly echoed back all of the QUIT messages during an squit.
 * This, however, is no longer practical.
 * 
 * To clarify here, DOWNSTREAM signifies an SQUIT heading towards the target server
 * UPSTREAM signifies an SQUIT which has successfully completed, heading out everywhere.
 *
 * acptr is the server being squitted.
 * a DOWNSTREAM squit is where the notice did not come from acptr->from.
 * an UPSTREAM squit is where the notice DID come from acptr->from.
 *
 *        parv[0] = sender prefix 
 *        parv[1] = server name 
 *        parv[2] = comment
 */

int m_squit(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;
    char *server;
    aClient *acptr;
    char *comment = (parc > 2) ? parv[2] : sptr->name;

    if (!IsPrivileged(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc > 1) {
	server = parv[1];
	/*
	 * To accomodate host masking, a squit for a masked server
	 * * name is expanded if the incoming mask is the same as the
	 * * server name for that link to the name of link.
	 */
	while ((*server == '*') && IsServer(cptr)) {
	    aconf = cptr->serv->nline;
	    if (!aconf)
		break;
	    if (!mycmp(server, my_name_for_link(me.name, aconf)))
		server = cptr->name;
	    break;		/* WARNING is normal here */
	    /*
	     * NOTREACHED 
	     */
	}
	/*
	 * * The following allows wild cards in SQUIT. Only useful * when
	 * the command is issued by an oper.
	 */
	for (acptr = client; (acptr = next_client(acptr, server));
	     acptr = acptr->next)
	    if (IsServer(acptr) || IsMe(acptr))
		break;
	if (acptr && IsMe(acptr)) {
	    acptr = cptr;
	    server = cptr->sockhost;
	}
    } else {
	/*
	 * This is actually protocol error. But, well, closing the
	 * * link is very proper answer to that...
	 */
	server = cptr->sockhost;
	acptr = cptr;
    }

    if (!acptr) {
	sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], server);
	return 0;
    }

    if (MyClient(sptr) && ((!OPCanGRoute(sptr) && !MyConnect(acptr)) ||
			   (!OPCanLRoute(sptr) && MyConnect(acptr)))) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    /*
     * If the server is mine, we don't care about upstream or downstream,
     * just kill it and do the notice. 
     */

    if (MyConnect(acptr)) {
	sendto_gnotice("from %s: Received SQUIT %s from %s (%s)",
		       me.name, acptr->name, get_client_name(sptr, HIDEME),
		       comment);
	sendto_serv_butone(NULL, ":%s GNOTICE :Received SQUIT %s from %s (%s)",
			   me.name, server, get_client_name(sptr, HIDEME),
			   comment);

	sendto_service(SERVICE_SEE_SQUITS, 0, ":%s SQUIT %s :%s", sptr->name,
		       server, comment);

#if defined(USE_SYSLOG) && defined(SYSLOG_SQUIT)
	syslog(LOG_DEBUG, "SQUIT From %s : %s (%s)", parv[0], server, comment);
#endif
	/*
	 * I am originating this squit! Not cptr! 
	 */
	/*
	 * ack, but if cptr is squitting itself.. 
	 */
	if (cptr == sptr) {
	    exit_client(&me, acptr, sptr, comment);
	    return FLUSH_BUFFER;	/* kludge */
	}
	return exit_client(&me, acptr, sptr, comment);
    }

    /*
     * the server is not connected to me. Determine whether this is an upstream
     * or downstream squit 
     */

    if (sptr->from == acptr->from) {	/* upstream */
	sendto_ops_lev(DEBUG_LEV,
		       "Exiting server %s due to upstream squit by %s [%s]",
		       acptr->name, sptr->name, comment);
	sendto_service(SERVICE_SEE_SQUITS, 0, ":%s SQUIT %s :%s", sptr->name,
		       acptr->name, comment);
	return exit_client(cptr, acptr, sptr, comment);
    }

    /*
     * fallthrough: downstream 
     */

    if (!(IsUnconnect(acptr->from))) {	/* downstream not unconnect capable */
	sendto_ops_lev(DEBUG_LEV,
		       "Exiting server %s due to non-unconnect server %s [%s]",
		       acptr->name, acptr->from->name, comment);
	sendto_service(SERVICE_SEE_SQUITS, 0, ":%s SQUIT %s :%s", sptr->name,
		       acptr->name, comment);
	return exit_client(&me, acptr, sptr, comment);
    }

    sendto_ops_lev(DEBUG_LEV, "Passing along SQUIT for %s by %s [%s]",
		   acptr->name, sptr->name, comment);
    sendto_one(acptr->from, ":%s SQUIT %s :%s", parv[0], acptr->name, comment);

    sendto_service(SERVICE_SEE_SQUITS, 0, ":%s SQUIT %s :%s", parv[0],
		   acptr->name, comment);

    return 0;
}

/*
 * * ts_servcount *   returns the number of TS servers that are
 * connected to us
 */
int ts_servcount()
{
int i;
aClient *acptr;
int r = 0;

    for (i = 0; i <= highest_fd; i++)
	if ((acptr = local[i]) && IsServer(acptr) && DoesTS(acptr))
	    r++;
    return r;
}

/*
 * * m_svinfo 
 *       parv[0] = sender prefix 
 *       parv[1] = TS_CURRENT for the server 
 *       parv[2] = TS_MIN for the server 
 *       parv[3] = server is standalone or connected to non-TS only 
 *       parv[4] = server's idea of UTC time
 */
int m_svinfo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    time_t deltat, tmptime, theirtime;

    if (!IsServer(sptr) || !MyConnect(sptr))
	return 0;

#ifdef USE_FAKEHOST
    if (!IsDTone(sptr)) {
	sendto_one(cptr, "ERROR :No Access (DT1 required)");
	sendto_ops("Access denied to %s (DT1 required)", sptr->name);
	return exit_client(cptr, cptr, cptr, "DT1 required");
    }
#else
    if (IsDTone(sptr)) {
	sendto_one(cptr, "ERROR :No Access (Your side uses DT1 - We do not!)");
	sendto_ops("Access denied to %s (Version mismatch - Remote uses DT1)",
		   sptr->name);
	return exit_client(cptr, cptr, cptr, "Version mismatch");
    }
#endif

    if (parc == 2) {
	if (mycmp(parv[1], "ZIP") == 0) {
	    SetZipIn(sptr);
	    sptr->serv->zip_in = zip_create_input_session();
	    sendto_gnotice("from %s: Input from %s is now compressed",
			   me.name, get_client_name(sptr, HIDEME));
	    sendto_serv_butone(sptr,
			       ":%s GNOTICE :Input from %s is now compressed",
			       me.name, get_client_name(sptr, HIDEME));
	    return ZIP_NEXT_BUFFER;

	} else if (mycmp(parv[1], "HIDEME") == 0) {
	    SetHidename(sptr);
	    sendto_gnotice("from %s: Servername %s will now be masked",
			   me.name, get_client_name(sptr, HIDEME));
	    return 0;
	}
    }

    if (parc < 5 || !DoesTS(sptr))
	return 0;

    if (TS_CURRENT < atoi(parv[2]) || atoi(parv[1]) < TS_MIN) {
	/*
	 * a server with the wrong TS version connected; since we're
	 * * TS_ONLY we can't fall back to the non-TS protocol so we drop
	 * * the link  -orabidoo
	 */
	sendto_ops("Link %s dropped, wrong TS protocol version (%s,%s)",
		   get_client_name(sptr, HIDEME), parv[1], parv[2]);
	return exit_client(sptr, sptr, sptr, "Incompatible TS version");
    }

    tmptime = time(NULL);
    theirtime = atol(parv[4]);
    deltat = abs(theirtime - tmptime);

    if (deltat > TS_MAX_DELTA) {
	sendto_gnotice
	    ("from %s: Link %s dropped, excessive TS delta (my TS=%d, their TS=%d, delta=%d)",
	     me.name, get_client_name(sptr, HIDEME), tmptime, theirtime,
	     deltat);
	sendto_serv_butone(sptr,
			   ":%s GNOTICE :Link %s dropped, excessive TS delta (delta=%d)",
			   me.name, get_client_name(sptr, HIDEME), deltat);
	return exit_client(sptr, sptr, sptr, "Excessive TS delta");
    }

    if (deltat > TS_WARN_DELTA) {
	sendto_realops
	    ("Link %s notable TS delta (my TS=%d, their TS=%d, delta=%d)",
	     get_client_name(sptr, HIDEME), tmptime, theirtime, deltat);
    }

    return 0;
}

/*
 * * m_server *       parv[0] = sender prefix *       parv[1] = servername *
 * parv[2] = serverinfo/hopcount *      parv[3] = serverinfo/hideinfo
 * parv[4] = serverinfo
 */
int m_server(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;
    char info[REALLEN + 1], *inpath, *host;
    aClient *acptr, *bcptr;
    aConfItem *aconf;
    int hop;
    char hide = 0;		/* server wants hidden names */
    char nbuf[HOSTLEN * 2 + USERLEN + 5];	/* same size as in s_misc.c */

    info[0] = '\0';
    inpath = get_client_name(cptr, HIDEME);

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(cptr, "ERROR :No servername");
	return 0;
    }

    hop = 0;
    host = parv[1];
    if (parc > 4 && atoi(parv[2])) {
	hop = atoi(parv[2]);
	hide = parv[3][0];
	strncpyzt(info, parv[4], REALLEN);
    } else if (parc > 3) {
	hop = atoi(parv[2]);
	strncpyzt(info, parv[3], REALLEN);
    } else if (parc > 2) {
	strncpyzt(info, parv[2], REALLEN);
	if ((parc > 3) && ((i = strlen(info)) < (REALLEN - 2))) {
	    (void) strcat(info, " ");
	    (void) strncat(info, parv[3], REALLEN - i - 2);
	    info[REALLEN] = '\0';
	}
    }
    /*
     * * July 5, 1997 
     * Rewritten to throw away server cruft from users, 
     * combined the hostname validity test with cleanup of host name,
     * so a cleaned up hostname can be returned as an error if
     * necessary. - Dianora
     */

    /*
     * yes, the if(strlen) below is really needed!! 
     */
    if (strlen(host) > HOSTLEN)
	host[HOSTLEN] = '\0';

    if (IsPerson(cptr)) {
	/*
	 * A local link that has been identified as a USER tries
	 * * something fishy... ;-)
	 */
	sendto_one(cptr, err_str(ERR_UNKNOWNCOMMAND), me.name, parv[0],
		   "SERVER");

	return 0;
    } else {
	/*
	 * Lets check for bogus names and clean them up we don't bother
	 * cleaning up ones from users, becasuse we will never see them
	 * any more - Dianora
	 */

    int bogus_server = 0;
    int found_dot = 0;
    char clean_host[(2 * HOSTLEN) + 1];
    char *s;
    char *d;
    int n;

	s = host;
	d = clean_host;
	n = (2 * HOSTLEN) - 2;

	while (*s && n > 0) {
	    if ((unsigned char) *s < (unsigned char) ' ') {	/* Is it a * *
								 * * control *
								 * * *
								 * character? */
		bogus_server = 1;
		*d++ = '^';
		*d++ = (char) ((unsigned char) *s + 0x40);	/* turn it into 
								 * a printable */
		n -= 2;
	    } else if ((unsigned char) *s > (unsigned char) '~') {
		bogus_server = 1;
		*d++ = '.';
		n--;
	    } else {
		if (*s == '.')
		    found_dot = 1;
		*d++ = *s;
		n--;
	    }
	    s++;
	}
	*d = '\0';

	if ((!found_dot) || bogus_server) {
	    sendto_one(sptr, "ERROR :Bogus server name (%s)", clean_host);
	    return exit_client(cptr, cptr, cptr, "Bogus server name");
	}
    }

    /*
     * check to see this host even has an N line before bothering 
     * anyone about it. Its only a quick sanity test to stop the
     * conference room and win95 ircd dorks. Sure, it will be
     * redundantly checked again in m_server_estab() *sigh* yes there
     * will be wasted CPU as the conf list will be scanned twice. But
     * how often will this happen? - Dianora
     * 
     * This should (will be) be recoded to check the IP is valid as well,
     * with a pointer to the valid N line conf kept for later, saving an
     * extra lookup.. *sigh* - Dianora
     */
    if (!IsServer(cptr)) {
	if (find_conf_name(host, CONF_NOCONNECT_SERVER) == NULL) {

#ifdef WARN_NO_NLINE
	    sendto_realops("Link %s dropped, no N: line",
			   get_client_name(cptr, TRUE));
#endif

	    return exit_client(cptr, cptr, cptr, "NO N line");
	}
    }

    if ((acptr = find_name(host, NULL))) {
	/*
	 * * This link is trying feed me a server that I already have 
	 * access through another path -- multiple paths not accepted 
	 * currently, kill this link immediately!! 
	 * 
	 * Rather than KILL the link which introduced it, KILL the 
	 * youngest of the two links. -avalon
	 */

	bcptr = (cptr->firsttime > acptr->from->firsttime) ? cptr : acptr->from;
	sendto_one(bcptr, "ERROR :Server %s already exists", host);
	if (bcptr == cptr) {
	    sendto_gnotice
		("from %s: Link %s cancelled, server %s already exists",
		 me.name, get_client_name(bcptr, HIDEME), host);
	    sendto_serv_butone(bcptr,
			       ":%s GNOTICE :Link %s cancelled, server %s already exists",
			       me.name, get_client_name(bcptr, HIDEME), host);
	    return exit_client(bcptr, bcptr, &me, "Server Exists");
	}
	/*
	 * inform all those who care (set +n) -epi 
	 */
	strcpy(nbuf, get_client_name(bcptr, HIDEME));
	sendto_gnotice
	    ("from %s: Link %s cancelled, server %s reintroduced by %s",
	     me.name, nbuf, host, get_client_name(cptr, HIDEME));
	sendto_serv_butone(bcptr,
			   ":%s GNOTICE :Link %s cancelled, server %s reintroduced by %s",
			   me.name, nbuf, host, get_client_name(cptr, HIDEME));
	(void) exit_client(bcptr, bcptr, &me, "Server Exists");
    }
    /*
     * The following if statement would be nice to remove since user
     * nicks never have '.' in them and servers must always have '.' in
     * them. There should never be a server/nick name collision, but it
     * is possible a capricious server admin could deliberately do
     * something strange.
     * 
     * -Dianora
     */

    if ((acptr = find_client(host, NULL)) && acptr != cptr) {
	/*
	 * * Server trying to use the same name as a person. Would 
	 * cause a fair bit of confusion. Enough to make it hellish for
	 * a while and servers to send stuff to the wrong place.
	 */
	sendto_one(cptr, "ERROR :Nickname %s already exists!", host);
	strcpy(nbuf, get_client_name(cptr, HIDEME));
	sendto_gnotice("from %s: Link %s cancelled, servername/nick collision",
		       me.name, nbuf);
	sendto_serv_butone(cptr,
			   ":%s GNOTICE :Link %s cancelled, servername/nick collision",
			   me.name, nbuf);
	return exit_client(cptr, cptr, cptr, "Nick as Server");
    }

    if (IsServer(cptr)) {
	/*
	 * * Server is informing about a new server behind this link.
	 * Create REMOTE server structure, add it to list and propagate
	 * word to my other server links...
	 */
	if (parc == 1 || info[0] == '\0') {
	    sendto_one(cptr, "ERROR :No server info specified for %s", host);
	    return 0;
	}

	if (!(aconf = find_conf_host(cptr->confs, host, CONF_HUB)) ||
	    (aconf->port && (hop > aconf->port))) {
	    sendto_gnotice("from %s: Non-Hub link %s introduced %s(%s).",
			   me.name, get_client_name(cptr, HIDEME), host,
			   aconf ? (aconf->host ? aconf->host : "*") : "!");
	    sendto_serv_butone(cptr,
			       ":%s GNOTICE :Non-Hub link %s introduced %s(%s).",
			       me.name, get_client_name(cptr, HIDEME),
			       host,
			       aconf ? (aconf->host ? aconf->host : "*") : "!");
	    sendto_one(cptr, "ERROR :%s has no H: line for %s.",
		       get_client_name(cptr, HIDEME), host);
	    return exit_client(cptr, cptr, cptr, "Too many servers");
	}

	acptr = make_client(cptr, sptr);
	(void) make_server(acptr);
	acptr->hopcount = hop;
	strncpyzt(acptr->name, host, sizeof(acptr->name));
	strncpyzt(acptr->info, info, REALLEN);
	acptr->serv->up = find_or_add(parv[0]);
	if (hide == 'H')
	    SetHidename(acptr);

	SetServer(acptr);

	/*
	 * if this server is behind a U-lined server, make it U-lined as well. - lucas 
	 */

	if (IsULine(sptr) || (find_uline(cptr->confs, acptr->name))) {
	    acptr->flags |= FLAGS_ULINE;
	    sendto_ops("%s introducing U:lined server %s", cptr->name,
		       acptr->name);
	}

	Count.server++;

	add_client_to_list(acptr);
	(void) add_to_client_hash_table(acptr->name, acptr);
	/*
	 * * Old sendto_serv_but_one() call removed because we now need
	 * to send different names to different servers (domain name matching)
	 */
	for (i = 0; i <= highest_fd; i++) {
	    if (!(bcptr = local[i]) || !IsServer(bcptr) || bcptr == cptr ||
		IsMe(bcptr))
		continue;
	    if (!(aconf = bcptr->serv->nline)) {
		sendto_gnotice("from %s: Lost N-line for %s on %s. Closing",
			       me.name, get_client_name(cptr, HIDEME), host);
		sendto_serv_butone(cptr,
				   ":%s GNOTICE :Lost N-line for %s on %s. Closing",
				   me.name, get_client_name(cptr, HIDEME),
				   host);
		return exit_client(cptr, cptr, cptr, "Lost N line");
	    }
	    if (match(my_name_for_link(me.name, aconf), acptr->name) == 0)
		continue;
	    if (CanHide(bcptr))
		sendto_one(bcptr, ":%s SERVER %s %d %c :%s", parv[0],
			   acptr->name, hop + 1, (WillHide(acptr) ? 'H' : 'N'),
			   acptr->info);
	    else
		sendto_one(bcptr, ":%s SERVER %s %d :%s", parv[0], acptr->name,
			   hop + 1, acptr->info);
	}
	return 0;
    }

    if (!IsUnknown(cptr) && !IsHandshake(cptr))
	return 0;
    /*
     * * A local link that is still in undefined state wants to be a
     * SERVER. Check if this is allowed and change status
     * accordingly...
     * Reject a direct nonTS server connection if we're TS_ONLY
     * -orabidoo
     */

    strncpyzt(cptr->name, host, sizeof(cptr->name));
    strncpyzt(cptr->info, info[0] ? info : me.name, REALLEN);
    cptr->hopcount = hop;
    if (hide == 'H')
	SetHidename(cptr);	/* is that necessary ? -TimeMr14C */

    switch (check_server_init(cptr)) {
	case 0:
	    return m_server_estab(cptr);
	case 1:
	    sendto_ops("Access check for %s in progress",
		       get_client_name(cptr, HIDEME));
	    return 1;
	default:
	    ircstp->is_ref++;
	    sendto_ops("Received unauthorized connection from %s.",
		       get_client_host(cptr));
	    return exit_client(cptr, cptr, cptr, "No C/N conf lines");
    }

}

static void sendnick_TS(aClient *cptr, aClient *acptr)
{
    static char ubuf[12];

    if (IsPerson(acptr)) {
	send_umode(NULL, acptr, 0, SEND_UMODES, ubuf);
	if (!*ubuf) {		/* trivial optimization - Dianora */
	    ubuf[0] = '+';
	    ubuf[1] = '\0';
	}
#ifdef USE_FAKEHOST
	if (IsNICKIP(cptr)) {
	    sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %s %lu %lu :%s",
		       acptr->name, acptr->hopcount + 1, acptr->tsinfo,
		       ubuf, acptr->user->username, acptr->user->host,
		       acptr->user->fakehost, acptr->user->server,
		       acptr->user->servicestamp, htonl(acptr->ip.s_addr),
		       acptr->info);
	} else {
	    sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %s %lu :%s",
		       acptr->name, acptr->hopcount + 1, acptr->tsinfo,
		       ubuf, acptr->user->username, acptr->user->host,
		       acptr->user->fakehost, acptr->user->server,
		       acptr->user->servicestamp, acptr->info);
	}
#else
	if (IsNICKIP(cptr)) {
	    sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %lu %lu :%s",
		       acptr->name, acptr->hopcount + 1, acptr->tsinfo,
		       ubuf, acptr->user->username, acptr->user->host,
		       acptr->user->server, acptr->user->servicestamp,
		       htonl(acptr->ip.s_addr), acptr->info);
	} else {
	    sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %lu :%s",
		       acptr->name, acptr->hopcount + 1, acptr->tsinfo,
		       ubuf, acptr->user->username, acptr->user->host,
		       acptr->user->server, acptr->user->servicestamp,
		       acptr->info);
	}
#endif
    }
}

int do_server_estab(aClient *cptr)
{
    aClient *acptr;
    aConfItem *aconf;
    aChannel *chptr;
    int i;
    char *inpath = get_client_name(cptr, HIDEME);	/* "refresh" inpath
							 * with host */

    SetServer(cptr);

    Count.server++;
    Count.myserver++;

    if (CanHide(cptr) && (bootopt & BOOT_HIDELINKS))
	sendto_one(cptr, "SVINFO HIDEME");
    if (IsZipCapable(cptr) && DoZipThis(cptr)) {
	sendto_one(cptr, "SVINFO ZIP");
	SetZipOut(cptr);
	cptr->serv->zip_out = zip_create_output_session();
    }
#ifdef MAXBUFFERS
    /*
     * let's try to bump up server sock_opts... -Taner 
     */
    reset_sock_opts(cptr->fd, 1);
#endif

    addto_fdlist(cptr->fd, &serv_fdlist);

#ifndef NO_PRIORITY
    /*
     * this causes the server to be marked as "busy" 
     */
    check_fdlists();
#endif

    cptr->pingval = get_client_ping(cptr);
    cptr->sendqlen = get_sendq(cptr);

    /*
     * error, error, error! if a server is U:'d, and it connects to us, 
     * * we need to figure that out! So, do it here. - lucas 
     */

    if ((find_uline(cptr->confs, cptr->name))) {
	Count.myulined++;
	cptr->flags |= FLAGS_ULINE;
    }

    sendto_gnotice("from %s: Link with %s established, states:%s%s%s%s",
		   me.name, inpath,
		   ZipOut(cptr) ? " Output-compressed" : "",
		   RC4EncLink(cptr) ? " encrypted" : "",
		   IsULine(cptr) ? " ULined" : "",
		   DoesTS(cptr) ? " TS" : " Non-TS");

    /*
     * Notify everyone of the fact that this has just linked: the entire network should get two
     * of these, one explaining the link between me->serv and the other between serv->me 
     */

    sendto_serv_butone(NULL, ":%s GNOTICE :Link with %s established: %s",
		       me.name, inpath,
		       DoesTS(cptr) ? "TS link" : "Non-TS link!");
    (void) add_to_client_hash_table(cptr->name, cptr);

    /*
     * add it to scache 
     */

    (void) find_or_add(cptr->name);

    /*
     * * Old sendto_serv_but_one() call removed because we now need to
     * send different names to different servers (domain name
     * matching) Send new server to other servers.
     */
    for (i = 0; i <= highest_fd; i++) {
	if (!(acptr = local[i]) || !IsServer(acptr) || acptr == cptr ||
	    IsMe(acptr))
	    continue;
	if ((aconf = acptr->serv->nline)
	    && !match(my_name_for_link(me.name, aconf), cptr->name))
	    continue;
	if (CanHide(acptr))
	    sendto_one(acptr, ":%s SERVER %s 2 %c :%s", me.name, cptr->name,
		       (WillHide(cptr) ? 'H' : 'N'), cptr->info);
	else
	    sendto_one(acptr, ":%s SERVER %s 2 :%s", me.name, cptr->name,
		       cptr->info);
    }

    sendto_service(SERVICE_SEE_SERVERS, 0, ":%s SERVER %s 2 :%s", me.name,
		   cptr->name, cptr->info);

    /*
     * Pass on my client information to the new server
     * 
     * First, pass only servers (idea is that if the link gets 
     * cancelled beacause the server was already there, there are no
     * NICK's to be cancelled...). Of course, if cancellation occurs,
     * all this info is sent anyway, and I guess the link dies when a
     * read is attempted...? --msa 
     * 
     * Note: Link cancellation to occur at this point means that at
     * least two servers from my fragment are building up connection
     * this other fragment at the same time, it's a race condition,
     * not the normal way of operation... 
     * 
     * ALSO NOTE: using the get_client_name for server names-- see
     * previous *WARNING*!!! (Also, original inpath is
     * destroyed...)
     */

    aconf = cptr->serv->nline;
    for (acptr = &me; acptr; acptr = acptr->prev) {
	/*
	 * acptr->from == acptr for acptr == cptr 
	 */
	if (acptr->from == cptr)
	    continue;
	if (IsServer(acptr)) {
	    if (match(my_name_for_link(me.name, aconf), acptr->name) == 0)
		continue;
	    if (CanHide(cptr))
		sendto_one(cptr, ":%s SERVER %s %d %c :%s",
			   acptr->serv->up, acptr->name, acptr->hopcount + 1,
			   (WillHide(acptr) ? 'H' : 'N'), acptr->info);
	    else
		sendto_one(cptr, ":%s SERVER %s %d :%s",
			   acptr->serv->up, acptr->name, acptr->hopcount + 1,
			   acptr->info);
	}
    }

    /*
     * send out our SQLINES and SGLINES too 
     */
    for (aconf = conf; aconf; aconf = aconf->next) {
	if ((aconf->status & CONF_QUARANTINE)
	    && (aconf->status & CONF_SQLINE))
	    sendto_one(cptr, ":%s SQLINE %s :%s", me.name, aconf->name,
		       aconf->passwd);
	if ((aconf->status & CONF_GCOS) && (aconf->status & CONF_SGLINE))
	    sendto_one(cptr, ":%s SGLINE %d :%s:%s", me.name,
		       strlen(aconf->name), aconf->name, aconf->passwd);
	if ((aconf->status & CONF_JUPECHAN)
	    && (aconf->status & CONF_SJUPECHAN))
	    sendto_one(cptr, ":%s JUPITER %s :%s", me.name, aconf->name,
		       aconf->passwd);

    }

    /*
     * Lets do szlines 
     */
    send_szlines(cptr);

    /*
     * we also need to inform about our services -TimeMr14C 
     */

    send_services(cptr);

    /*
     * Bursts are about to start.. send a BURST 
     */
    if (IsBurst(cptr))
	sendto_one(cptr, "BURST");

    /*
     * * Send it in the shortened format with the TS, if it's a TS
     * server; walk the list of channels, sending all the nicks that
     * haven't been sent yet for each channel, then send the channel
     * itself -- it's less obvious than sending all nicks first, but
     * on the receiving side memory will be allocated more nicely 
     * saving a few seconds in the handling of a split -orabidoo
     */
    {
    chanMember *cm;
    static char nickissent = 1;

	nickissent = 3 - nickissent;
	/*
	 * flag used for each nick to check if we've sent it yet - must
	 * be different each time and !=0, so we alternate between 1 and
	 * 2 -orabidoo
	 */
	for (chptr = channel; chptr; chptr = chptr->nextch) {
	    for (cm = chptr->members; cm; cm = cm->next) {
		acptr = cm->cptr;
		if (acptr->nicksent != nickissent) {
		    acptr->nicksent = nickissent;
		    if (acptr->from != cptr)
			sendnick_TS(cptr, acptr);
		}
	    }
	    send_channel_modes(cptr, chptr);
	}
	/*
	 * also send out those that are not on any channel 
	 */
	for (acptr = &me; acptr; acptr = acptr->prev)
	    if (acptr->nicksent != nickissent) {
		acptr->nicksent = nickissent;
		if (acptr->from != cptr)
		    sendnick_TS(cptr, acptr);
	    }
    }

    if (ZipOut(cptr)) {
    unsigned long inb, outb;
    double rat;

	zip_out_get_stats(cptr->serv->zip_out, &inb, &outb, &rat);

	if (inb) {
	    sendto_gnotice
		("from %s: Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
		 me.name, get_client_name(cptr, HIDEME), inb, outb, rat);
	    sendto_serv_butone(cptr,
			       ":%s GNOTICE :Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
			       me.name, get_client_name(cptr, HIDEME), inb,
			       outb, rat);
	}
    }

    /*
     * stuff a PING at the end of this burst so we can figure out when
     * the other side has finished processing it. 
     */
    cptr->flags |= FLAGS_BURST | FLAGS_PINGSENT;
    if (IsBurst(cptr))
	cptr->flags |= FLAGS_SOBSENT;
    sendto_one(cptr, "PING :%s", me.name);

    return 0;
}

int m_server_estab(aClient *cptr)
{
    aConfItem *aconf, *bconf;

    char *inpath, *host, *s, *encr;
    int split;

#ifndef HUB
    int i;
#endif

    inpath = get_client_name(cptr, HIDEME);	/* "refresh" inpath with host */
    split = mycmp(cptr->name, cptr->sockhost);
    host = cptr->name;

    if (!(aconf = find_conf(cptr->confs, host, CONF_NOCONNECT_SERVER))) {
	ircstp->is_ref++;
	sendto_one(cptr, "ERROR :Access denied. No N line for server %s",
		   inpath);
	sendto_ops("Access denied. No N line for server %s", inpath);
	return exit_client(cptr, cptr, cptr, "No N line for server");
    }
    if (!(bconf = find_conf(cptr->confs, host, CONF_CONNECT_SERVER))) {
	ircstp->is_ref++;
	sendto_one(cptr, "ERROR :Only N (no C) field for server %s", inpath);
	sendto_ops("Only N (no C) field for server %s", inpath);
	return exit_client(cptr, cptr, cptr, "No C line for server");
    }

    encr = cptr->passwd;
    if (*aconf->passwd && !StrEq(aconf->passwd, encr)) {
	ircstp->is_ref++;
	sendto_one(cptr, "ERROR :No Access (passwd mismatch) %s", inpath);
	sendto_ops("Access denied (passwd mismatch) %s", inpath);
	return exit_client(cptr, cptr, cptr, "Bad Password");
    }
    memset(cptr->passwd, '\0', sizeof(cptr->passwd));

#ifndef	HUB
    for (i = 0; i <= highest_fd; i++)
	if (local[i] && IsServer(local[i])) {
	    ircstp->is_ref++;
	    sendto_one(cptr, "ERROR :I'm a leaf not a hub");
	    return exit_client(cptr, cptr, cptr, "I'm a leaf");
	}
#endif

    /*
     * aconf->port is a CAPAB field, kind-of. kludge. mm, mm. 
     */
    cptr->capabilities |= aconf->port;
    if (IsUnknown(cptr)) {
    int dontwantflags = 0;
	if (bconf->passwd[0])
	    sendto_one(cptr, "PASS %s :TS", bconf->passwd);

#ifdef HAVE_ENCRYPTION_ON
	dontwantflags |= WantDKEY(cptr) ? 0 : CAPAB_DKEY;
#endif

	send_capab_to(cptr, dontwantflags);
	sendto_one(cptr, "SERVER %s 1 :%s",
		   my_name_for_link(me.name, aconf),
		   (me.info[0]) ? (me.info) : "IRCers United");
    } else {
	s = (char *) strchr(aconf->host, '@');
	*s = '\0';		/* should never be NULL -- wanna bet? -Dianora */

	Debug((DEBUG_INFO, "Check Usernames [%s]vs[%s]", aconf->host,
	       cptr->username));
	if (match(aconf->host, cptr->username)) {
	    *s = '@';
	    ircstp->is_ref++;
	    sendto_ops("Username mismatch [%s]v[%s] : %s",
		       aconf->host, cptr->username, get_client_name(cptr,
								    HIDEME));
	    sendto_one(cptr, "ERROR :No Username Match");
	    return exit_client(cptr, cptr, cptr, "Bad User");
	}
	*s = '@';
    }

    /*
     * send routing notice, this should never happen anymore 
     */
    if (!DoesTS(cptr)) {
	sendto_gnotice("from %s: Warning: %s linked, non-TS server",
		       me.name, get_client_name(cptr, TRUE));
	sendto_serv_butone(cptr,
			   ":%s GNOTICE :Warning: %s linked, non-TS server",
			   me.name, get_client_name(cptr, TRUE));
    }

    sendto_one(cptr, "SVINFO %d %d 0 :%ld", TS_CURRENT, TS_MIN,
	       (ts_val) timeofday);

    /*
     * sendto one(cptr, "CAPAB ...."); moved to after PASS but before SERVER
     * * now in two places.. up above and in s_bsd.c. - lucas
     * * This is to make sure we pass on our capabilities before we establish a server connection 
     */

    det_confs_butmask(cptr, CONF_HUB | CONF_NOCONNECT_SERVER | CONF_ULINE);
    /*
     * * *WARNING* 
     *   In the following code in place of plain
     * server's name we send what is returned by
     * get_client_name which may add the "sockhost" after the name.
     * It's *very* *important* that there is a SPACE between 
     * the name and sockhost (if present). The receiving server
     * will start the information field from this first blank and
     * thus puts the sockhost into info. ...a bit tricky, but
     * you have been warned, besides code is more neat this way...
     * --msa
     */

    /*
     * doesnt duplicate cptr->serv if allocated this struct already 
     */

    (void) make_server(cptr);
    cptr->serv->up = me.name;
    cptr->serv->nline = aconf;

    /*
     * now fill out the servers info so nobody knows dink about it. 
     */
    memset((char *) &cptr->ip, '\0', sizeof(struct in_addr));
    strcpy(cptr->hostip, "127.0.0.1");
    strcpy(cptr->sockhost, "localhost");

#ifdef HAVE_ENCRYPTION_ON
    if (!CanDoDKEY(cptr) || !WantDKEY(cptr))
	return do_server_estab(cptr);
    else {
	SetNegoServer(cptr);	/* VERY IMPORTANT THAT THIS IS HERE */
	sendto_one(cptr, "DKEY START");
    }
#else
    return do_server_estab(cptr);
#endif

    return 0;
}

/*
 * m_burst
 *      parv[0] = sender prefix
 *      parv[1] = SendQ if an EOB
 */
int m_burst(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    if (!IsServer(sptr) || sptr != cptr || parc > 2 || !IsBurst(sptr))
	return 0;
    if (parc == 2) {		/* This is an EOB */
	sptr->flags &= ~(FLAGS_EOBRECV);
	if (sptr->flags & (FLAGS_SOBSENT | FLAGS_BURST))
	    return 0;

	/*
	 * we've already sent our EOB.. we synched first
	 * * no need to check IsBurst because we shouldn't receive a BURST if they're not BURST capab 
	 */

#ifdef HTM_LOCK_ON_NETBURST
	HTMLOCK = NO;
#endif
	sendto_gnotice("from %s: synch to %s in %d %s at %s sendq",
		       me.name, *parv, (timeofday - sptr->firsttime),
		       (timeofday - sptr->firsttime) == 1 ? "sec" : "secs",
		       parv[1]);
	sendto_serv_butone(NULL,
			   ":%s GNOTICE :synch to %s in %d %s at %s sendq",
			   me.name, sptr->name, (timeofday - sptr->firsttime),
			   (timeofday - sptr->firsttime) == 1 ? "sec" : "secs",
			   parv[1]);

    } else {			/* SOB.. lock HTM if defined by admin */
#ifdef HTM_LOCK_ON_NETBURST
	HTMLOCK = YES;
#endif
	sptr->flags |= FLAGS_EOBRECV;
    }
    return 0;
}

/*
 * * Note: At least at protocol level ERROR has only one parameter, 
 * although this is called internally from other functions  --msa 
 *
 *      parv[0] = sender prefix 
 *      parv[*] = parameters
 */
int m_error(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *para;

    para = (parc > 1 && *parv[1] != '\0') ? parv[1] : "<>";

    Debug((DEBUG_ERROR, "Received ERROR message from %s: %s", sptr->name,
	   para));
    /*
     * * Ignore error messages generated by normal user clients 
     * (because ill-behaving user clients would flood opers screen
     * otherwise). Pass ERROR's from other sources to the local
     * operator...
     */
    if (IsPerson(cptr) || IsUnknown(cptr))
	return 0;
    if (cptr == sptr)
	sendto_ops("ERROR :from %s -- %s", get_client_name(cptr, HIDEME), para);
    else
	sendto_ops("ERROR :from %s via %s -- %s", sptr->name,
		   get_client_name(cptr, HIDEME), para);
    return 0;
}
