/************************************************************************
 *   IRC - Internet Relay Chat, src/s_service.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * -TimeMr14C
 *
 * 06.01.2001
 *
 * This file contains every newer function used in the ircd
 * of the IRCTR network
 * Any laws and rules applicable to other source files of
 * the Bahamut ircd are applicable to this file, too.
 * This file contains the core sources for the service extension
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"

#if defined( HAVE_STRING_H)
#include <string.h>
#else
#include <strings.h>
#endif

extern void reset_sock_opts();

static char buf[BUFSIZE];
static char buf2[BUFSIZE];

int service_info[] = {
    SMODE_H, 'H',
    SMODE_U, 'U',
    SMODE_A, 'A',
    0, 0
};

int service_modes[] = {
    SERVICE_SEE_PREFIX, 'P',
    SERVICE_SEE_OPERS, 'O',
    SERVICE_SEE_NICKS, 'N',
    SERVICE_SEE_QUITS, 'Q',
    SERVICE_SEE_KILLS, 'K',
    SERVICE_SEE_SERVERS, 'S',
    SERVICE_SEE_SERVICES, 'Z',
    SERVICE_SEE_SQUITS, 'L',
    SERVICE_SEE_RNICKS, 'R',
    SERVICE_SEE_UMODES, 'U',
    SERVICE_SEE_JOINS, 'j',
    SERVICE_SEE_KICKS, 'k',
    SERVICE_SEE_PARTS, 'p',
    SERVICE_SEE_MODES, 'm',
    SERVICE_SEE_TOPIC, 't',
    0, 0
};

aService *firstservice = NULL;

aService *make_service(aClient *cptr)
{

    aService *svc = cptr->service;

    if (!svc) {
	cptr->service = svc = (aService *) MyMalloc(sizeof(aService));
	memset(svc, 0, sizeof(aService));

	svc->scptr = cptr;
	if (firstservice)
	    firstservice->prev = svc;
	svc->next = firstservice;
	svc->prev = NULL;
	firstservice = svc;
    }
    return svc;
}

void free_service(aClient *cptr)
{

    aService *svc = cptr->service;

    if (svc) {

	if (svc->next)
	    svc->next->prev = svc->prev;
	if (svc->prev)
	    svc->prev->next = svc->next;
	if (firstservice == svc)
	    firstservice = svc->next;
	if (svc->server)
	    MyFree(svc->server);

	MyFree((char *) svc);
	cptr->service = NULL;

    }

}

static char *service_modes_to_char(aService * svc)
{

char *m;
int e;
int *s;

    m = buf2;
    *m = '\0';
    for (s = service_modes; (e = *s); s += 2)
	if (svc->enable & e)
	    *m++ = *(s + 1);
    *m = '\0';

    return buf2;
}

static char *service_info_to_char(aService * svc)
{

char *m;
int f;
int *s;

    m = buf;
    *m = '\0';
    for (s = service_info; (f = *s); s += 2)
	if (svc->sflags & f)
	    *m++ = *(s + 1);
    *m = '\0';

    return buf;

}

/*
 * -TimeMr14C
 * the m_service function
 * This is the function, where a connection defines itself
 * as a service instead of a user or server
 * it will also be parsed, if a remote server
 * introduces a service to us
 *
 *      parv[0] = sender prefix
 *      parv[1] = service name
 *      parv[2] = ident of the service
 *      parv[3] = service description
 *
 * The upper one should only be accepted if arriving from a
 * local client connection, else ignored.
 * 
 *      parv[0] = sender prefix
 *      parv[1] = service name
 *      parv[2] = timestamp of service introduction
 *      parv[3] = hopcount of the service
 *      parv[4] = flags of the service
 *      parv[5] = ident of the service
 *      parv[6] = hostname of the service
 *      parv[7] = server of the service
 *      parv[8] = description of the service
 *
 * This formation results from the SERVICE command arriving from a
 * server. It should be ignored if arriving from a user connection.
 *
 */

int m_service(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;
    aService *svc;
    aConfItem *aconf;

    char *m;
    char name[NICKLEN + 2];
    int *s;
    int flag = 0;
    int flags = 0;
    int modes = 0;

    if ((IsServer(cptr) && (parc < 8))
	|| (IsUnknown(cptr) && (parc < 4))) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "SERVICE");
	return 0;
    }

    if (IsServer(cptr) && HasServices(cptr)) {

	if (!isdigit(parv[2][0]) || !isdigit(parv[3][0])) {
	    sendto_realops_lev(SERVICE_LEV,
			       "Remote server %s tried to introduce non-ts service %s",
			       cptr->name, parv[1]);
	    return 0;
	}

	strncpyzt(name, parv[1], NICKLEN + 1);

	if ((acptr = hash_find_client(parv[1], (aClient *) NULL)) != NULL) {
	    return 0;
	}

	acptr = find_server(parv[7], NULL);

	if (!acptr) {
	    sendto_realops_lev(SERVICE_LEV,
			       "Remote SERVICE %s on unknown server %s", name,
			       parv[7]);
	    return 0;
	}

	sptr = make_client(cptr, acptr);

	sptr->hopcount = atoi(parv[2]);
	sptr->tsinfo = atol(parv[3]);

	m = &parv[4][0];
	while (*m) {
	    for (s = service_info; (flag = *s); s += 2)
		if (*m == *(s + 1)) {
		    flags |= flag;
		    break;
		}
	    m++;
	}

	strcpy(sptr->name, name);

	strncpyzt(sptr->username, parv[5], USERLEN + 1);
	strncpyzt(sptr->sockhost, parv[6], HOSTLEN + 1);
	strncpyzt(sptr->info, parv[8], REALLEN + 1);

	sptr->status = STAT_SERVICE;

	svc = make_service(sptr);
	svc->sflags |= flags;
	svc->enable = modes;
	svc->server = strdup(parv[7]);

	add_client_to_list(sptr);
	add_to_client_hash_table(sptr->name, sptr);

	sendto_realops_lev(SERVICE_LEV,
			   "Created service %s (%s@%s) from %s. [%s]",
			   sptr->name, sptr->username, sptr->sockhost,
			   svc->server, sptr->info);

	Count.service++;

	m = buf2;
	*m = '\0';
	for (s = service_info; (flag = *s); s += 2)
	    if (svc->sflags & flag)
		*m++ = *(s + 1);
	*m = '\0';

	sendto_realops_lev(SERVICE_LEV,
			   "Flags for Service %s are: %s%s%s", sptr->name,
			   (strchr(buf2, 'U') ? "U:Lined " : ""),
			   (strchr(buf2, 'H') ? "Routing " : ""),
			   (strchr(buf2, 'A') ? "Akill " : ""));

	if (svc->sflags & SMODE_U)
	    svc->scptr->flags |= FLAGS_ULINE;

	sendto_flag_serv_butone(cptr, CAPAB_SERVICES, 0,
				"SERVICE %s %d %ld %s %s %s %s :%s",
				sptr->name, (sptr->hopcount + 1),
				sptr->tsinfo, service_info_to_char(svc),
				sptr->username, sptr->sockhost, svc->server,
				sptr->info);

	sendto_service(SERVICE_SEE_SERVICES, 0,
		       ":%s SERVICE %s %d %ld %s %s %s %s :%s",
		       cptr->name, sptr->name, (sptr->hopcount + 1),
		       sptr->tsinfo, service_info_to_char(svc),
		       sptr->username, sptr->sockhost, svc->server, sptr->info);

    }

    else if (!IsServer(cptr) && !IsPerson(cptr) && MyConnect(cptr)) {

    aClient *acptr;

	strncpyzt(name, parv[1], NICKLEN + 1);

	if ((aconf = find_conf_name(name, CONF_SERVICE)) == NULL) {
	    return exit_client(cptr, sptr, &me, "Service Does Not Exist");
	}

	if (!BadPtr(aconf->passwd) && !StrEq(aconf->passwd, cptr->passwd)) {
	    sendto_realops_lev(REJ_LEV,
			       "Access denied (password mismatch) to %s (%s@%s)",
			       name, parv[2], get_client_name(cptr, TRUE));
	    return exit_client(cptr, sptr, &me, "Password incorrect");
	}

	if ((acptr = hash_find_client(parv[1], (aClient *) NULL)) != NULL) {
	    sendto_realops_lev(REJ_LEV, "Service %s already exists",
			       acptr->name);
	    return exit_client(cptr, sptr, &me, "Service already exists");
	}

	strcpy(cptr->name, name);

	// if ((cptr->flags & FLAGS_DOID) && !(cptr->flags &
	// FLAGS_GOTID))
	strncpyzt(cptr->username, parv[2], USERLEN + 1);

	strncpyzt(cptr->info, parv[3], REALLEN + 1);

	cptr->tsinfo = timeofday;

	attach_conf(cptr, aconf);

	cptr->status = STAT_SERVICE;

	cptr->pingval = get_client_ping(cptr);
	cptr->sendqlen = get_sendq(cptr);

#ifdef MAXBUFFERS
	reset_sock_opts(cptr->fd, 0);
#endif

	memset(cptr->passwd, '\0', PASSWDLEN);

	svc = make_service(cptr);
	svc->sflags |= aconf->port;
	svc->enable = 0;
	svc->server = strdup(me.name);

	sendto_one(cptr,
		   ":%s NOTICE AUTH :SERVICE %s with %d %d %s %s %d :%s",
		   me.name, cptr->name, cptr->tsinfo, cptr->flags,
		   cptr->username, svc->server, svc->sflags, cptr->info);

	add_to_client_hash_table(cptr->name, cptr);

	sendto_one(cptr, rpl_str(RPL_YOURESERVICE), me.name, cptr->name);
	sendto_one(cptr, rpl_str(RPL_YOURHOST), me.name, cptr->name,
		   get_client_name(&me, TRUE), version);

	sendto_one(cptr, rpl_str(RPL_CREATED), me.name, cptr->name, creation);
	sendto_one(cptr, rpl_str(RPL_MYINFO), me.name, cptr->name, me.name,
		   version);
	sendto_one(sptr, rpl_str(RPL_PROTOCTL), me.name, parv[0], MAXWATCH,
		   MAXMODEPARAMS, MAXCHANNELSPERUSER, MAXBANS, NICKLEN,
		   TOPICLEN, TOPICLEN, NETWORKNAME, MAXSILES);

	sendto_realops_lev(SERVICE_LEV,
			   "Service %s (%s@%s) connected. [%s]",
			   cptr->name, cptr->username, cptr->sockhost,
			   cptr->info);

	Count.myservice++;
	Count.service++;

	m = buf2;
	*m = '\0';
	for (s = service_info; (flag = *s); s += 2)
	    if (svc->sflags & flag)
		*m++ = *(s + 1);
	*m = '\0';

	sendto_realops_lev(SERVICE_LEV,
			   "Flags for Service %s are: %s%s%s", cptr->name,
			   (strchr(buf2, 'U') ? "U:Lined " : ""),
			   (strchr(buf2, 'H') ? "Routing " : ""),
			   (strchr(buf2, 'A') ? "Akill " : ""));

	if (svc->sflags & SMODE_U)
	    cptr->flags |= FLAGS_ULINE;

	sendto_flag_serv_butone(cptr, CAPAB_SERVICES, 0,
				"SERVICE %s 1 %ld %s %s %s %s :%s",
				cptr->name, cptr->tsinfo,
				service_info_to_char(svc), cptr->username,
				cptr->sockhost, me.name, cptr->info);

	sendto_service(SERVICE_SEE_SERVICES, 0,
		       ":%s SERVICE 1 %ld %s %s %s %s :%s", me.name,
		       cptr->name, cptr->tsinfo,
		       service_info_to_char(svc), cptr->username,
		       cptr->sockhost, me.name, cptr->info);
    }

    else if (IsServer(sptr) && !HasServices(sptr)) {

	sendto_realops
	    ("Non-Service capable server %s is trying to introduce service",
	     sptr->name);
    }

    else {

	/*
	 * Someone is trying something fishy ;-P 
	 */
	/*
	 * -TimeMr14C 
	 */

	sendto_one(sptr, err_str(ERR_UNKNOWNCOMMAND), me.name, sptr->name,
		   "SERVICE");
    }

    return 0;
}

/*
 * Returns list of all matching services.
 */

int m_servlist(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aService *svc;
    aClient *acptr;
    char out[HOSTLEN];

    for (svc = firstservice; svc; svc = svc->next) {
	if ((acptr = svc->scptr) != NULL)
	    sendto_one(sptr, rpl_str(RPL_SERVLIST), me.name, parv[0],
		       acptr->name, acptr->username,
		       (IsOper(sptr) ? acptr->sockhost : "<hidden>"),
		       (WillHide(acptr) ? stealth_server(svc->server, out)
			: svc->server), acptr->hopcount, acptr->info);
    }
    sendto_one(sptr, rpl_str(RPL_SERVLISTEND), me.name, parv[0]);
    return 2;
}

void send_services(aClient *to)
{

    aService *svc;
    aClient *acptr;

    if (!HasServices(to))
	return;

    for (svc = firstservice; svc; svc = svc->next) {
	if (((acptr = svc->scptr) != NULL) && (svc->scptr != to)) {

	    sendto_one(to, "SERVICE %s 1 %ld %s %s %s %s :%s",
		       acptr->name, acptr->tsinfo,
		       service_info_to_char(svc), acptr->username,
		       acptr->sockhost, svc->server, acptr->info);
	}
    }
    return;

}

/*
 * -TimeMr14C
 *
 * Until there is a solution how we can make a service watch
 * many many channels (hundreds) The option to see a specific
 * channel will stay disabled. The SERVSET command will ignore
 * additional parametres.
 *
 * 15/06/2001
 */

/*
 * static int does_checkchan(aService *svcptr, char *chan) {
 *
 *        Link *lp;
 *
 *        if (!(svcptr) || !(lp = svcptr->channel))
 *           return 0;
 *
 *        for ( ; lp; lp = lp->next) {
 *           if (!mycmp(lp->value.cp, chan)) 
 *               return 1;
 *        }
 *        return 0;
 * }
 *
 *
 * static char *get_checkchan(aService *svcptr, char *list) {
 *
 *        Link *lp;
 *       char chanarray[329];         
 *       int first;
 *
 *       if (!(svcptr) || !(lp = svcptr->channel)) {
 *          list = " ";
 *          return list;
 *       }
 *                 
 *        for (first = 0; lp; lp = lp->next) {
 *          if (!first)
 *              strcpy(chanarray, "+");
 *          else {
 *              strcat(chanarray, " ");
 *              strcat(chanarray, "+");
 *          }
 *          strcat(chanarray, lp->value.cp);
 *          first++;
 *       }
 *       strcpy(list, chanarray);
 *       return list;
 * }
 *
 * int del_checkchan(aService *svcptr, char *chan) {
 *       Link **lp, *tmp;
 *                         
 *       for (lp = &(svcptr->channel); *lp; lp=&((*lp)->next))
 *           if (mycmp(chan, (*lp)->value.cp)==0) {
 *                 tmp = *lp;
 *                 *lp = tmp->next;
 *                 MyFree(tmp->value.cp);
 *                 free_link(tmp);
 *                 return 0;
 *           }
 *       return 1;
 * }
 *
 * static int add_checkchan(aService *svcptr, char *chan) {
 *       Link *lp;
 *       int cnt=0, len=0;
 *
 *       for (lp = svcptr->channel; lp; lp=lp->next) {
 *
 *           len += strlen(lp->value.cp);
 *
 *           if ((++cnt >= MAXSERVICECHAN)) {
 *                 sendto_one(svcptr->scptr, err_str(ERR_NOCHANTOWATCH), 
 *                              me.name, svcptr->scptr->name, chan);
 *                 return -1;
 *           } else {
 *                 if (!mycmp(lp->value.cp, chan)) 
 *                     return -1;
 *           }
 *       }
 *       lp = make_link();
 *       lp->next = svcptr->channel;
 *       lp->value.cp = (char *)MyMalloc(strlen(chan)+1);
 *       (void)strcpy(lp->value.cp, chan);
 *       svcptr->channel = lp;
 *       return 0;
 * }
 */
/*
 * sets the modes for the service, respecting the defines
 * about what the service is going to be able to see.
 * -TimeMr14C
 *
 *      parv[0] = sender prefix
 *      parv[1] = service name
 *      parv[2] = +/-modes
 *      parv[3] = possible list of channels - currently only one.
 *
 */

int m_servset(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    int flag;
    int *s;
    char *m;
    aClient *acptr;
    int what = 0;

    /*
     * char list[329]; 
     */

    if (parc < 3) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "SERVSET");
	return 0;
    }

    if ((acptr = hash_find_client(parv[1], (aClient *) NULL)) == NULL) {
	sendto_realops_lev(SERVICE_LEV,
			   "Received SERVSET for non existant service %s",
			   parv[1]);
	return 0;
    }

    if (!IsService(acptr))
	return 0;

    for (m = parv[2]; *m; m++) {
	switch (*m) {

	    case '+':
		what = MODE_ADD;
		break;
	    case '-':
		what = MODE_DEL;
		break;
	    case ' ':
	    case '\r':
	    case '\n':
	    case '\t':
		break;
	    default:
		for (s = service_modes; (flag = *s); s += 2) {
		    if (*m == (char) (*(s + 1))) {
			if (what == MODE_ADD)
			    acptr->service->enable |= flag;
			else
			    acptr->service->enable &= ~flag;
			break;
		    }
		}
		break;
	}
    }

    /*
     *  if (parv[3]) {
     *
     *      char *r = NULL;
     *      char *chkptr;
     *      int chknum;
     *      int len;
     *      int amount;
     *
     *       * This parameter of the servset contains names like +#test1 +#test2 -#test3
     *       * We do need to check the first char of every word, whether it is a + or -
     *       * Drop any word, that is none of them, + means add to the checkchan list, 
     *       * - means remove from that list.
     *
     *      for (amount = 0, chkptr = strtoken(&r, parv[3], " "); chkptr; 
     *                       chkptr = strtoken(&r, NULL, " ")) {
     *      
     *          if (*chkptr == '+') 
     *              chknum = 1;
     *          else if (*chkptr == '-')
     *              chknum = -1;
     *          else
     *              continue;
     *         
     *          chkptr++;
     *
     *          if (!check_channelname(acptr, (unsigned char *) chkptr))
     *              continue;
     *
     *          len = strlen(chkptr);
     *          if (len > CHANNELLEN)
     *              chkptr[CHANNELLEN] = '\0';
     *
     *          if (!IsChannelName(chkptr) || (amount > 9)) {
     *              if (MyClient(acptr))
     *                  sendto_one(acptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], chkptr);
     *               continue;
     *          }
     *
     *          if (chknum > 0) 
     *              add_checkchan(acptr->service, chkptr);
     *          else if (chknum < 0) 
     *              del_checkchan(acptr->service, chkptr);
     *          amount++;
     *
     *      }       
     *
     *  } 
     *
     *  if (MyConnect(acptr)) 
     *      sendto_one(acptr, ":%s %d %s +%s :%s", me.name, RPL_UMODEIS, parv[0],
     *              service_modes_to_char(acptr->service), get_checkchan(acptr->service, list));
     *
     */

    if (MyConnect(acptr))
	sendto_one(acptr, ":%s %d %s +%s", me.name, RPL_UMODEIS, parv[0],
		   service_modes_to_char(acptr->service));

    return 0;
}

/*
 * Send query to service.
 * parv[1] - string to match name against
 * parv[2] - string to send to service
 */

int m_squery(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NORECIPIENT), me.name, parv[0], "SQUERY");
	return 1;
    }
    if (parc == 2) {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return 2;
    }

    if (((acptr = hash_find_client(parv[1], (aClient *) NULL)) != NULL)
	&& (acptr->service)) {
	if (acptr->service->enable & SERVICE_SEE_PREFIX)
	    sendto_one(acptr, ":%s!%s@%s SQUERY %s :%s", parv[0],
		       sptr->user->username, sptr->user->host, acptr->name,
		       parv[2]);
	else
	    sendto_one(acptr, ":%s SQUERY %s :%s", parv[0], acptr->name,
		       parv[2]);
    } else {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
    }
    return 0;
}

void sendto_service(int include, int exclude, char *pattern, ...)
{
    aService *svc;
    va_list vl;

    va_start(vl, pattern);

    for (svc = firstservice; svc; svc = svc->next) {
	if ((svc->scptr == NULL) || (!MyConnect(svc->scptr)) ||
	    (include && !((svc->enable & include) == include)) ||
	    (exclude && (svc->enable & exclude) == exclude))
	    continue;
	vsendto_one(svc->scptr, pattern, vl);
    }
    va_end(vl);
    return;
}

/*
 * void sendto_service_with_chan(char *chan, int include, int exclude, char *pattern, ...) {
 *
 *  aService *svc;
 *  int includs;
 *  va_list vl;
 *
 *   va_start(vl, pattern);
 *
 *  includs = (include | SERVICE_SEE_CHANNEL);
 *
 *  for (svc = firstservice; svc; svc = svc->next) {
 *     if (
 *      (svc->scptr == NULL) || 
 *      (!MyConnect(svc->scptr)) ||
 *      (includs && !((svc->enable & includs) == includs)) ||
 *      (exclude &&  ((svc->enable & exclude) == exclude)) ||
 *      (!does_checkchan(svc, chan))
 *     ) continue;
 *     vsendto_one(svc->scptr, pattern, vl);
 *  }
 *  va_end(vl);
 *  return;
 *
 * }
 */
