/***********************************************************************
 *   IRC - Internet Relay Chat, src/send.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: send.c,v 1.9 2001/12/05 18:31:50 tr-ircd Exp $ 
 */

#include "struct.h"
#include "channel.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include "numeric.h"
#include "dh.h"
#include "zlink.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendbuf[2048];

#ifdef HAVE_ENCRYPTION_ON
static char rc4buf[768];
#endif

#ifdef HTM_LOCK_ON_NETBURST
extern int HTMLOCK;
#endif

static int sentalong[MAXCONNECTIONS];
static int sent_serial;

void init_send()
{
    memset(sentalong, 0, sizeof(int) * MAXCONNECTIONS);
    sent_serial = 0;
}

/*
 * This routine increments our serial number so it will
 * * be unique from anything in sentalong, no need for a memset
 * * except for every MAXINT calls - lucas
 */

/*
 * This should work on any OS where an int is 32 bit, I hope.. 
 */

#define HIGHEST_SERIAL INT_MAX

#define INC_SERIAL if(sent_serial == HIGHEST_SERIAL) \
   { memset(sentalong, 0, sizeof(sentalong)); sent_serial = 0; } \
   sent_serial++;

/*
 * dead_link
 *
 * somewhere along the lines of sending out, there was an error.
 * we can't close it from the send loop, so mark it as dead
 * and close it from the main loop.
 *
 * if this link is a server, tell routing people.
 */

static int dead_link(aClient *to, char *notice, int sockerr)
{

    int errtmp = errno;		/* so we don't munge this later */

    to->flags |= FLAGS_DEADSOCKET;
    /*
     * If because of BUFFERPOOL problem then clean dbuf's now so that
     * notices don't hurt operators below.
     */
    DBufClear(&to->recvQ);
    DBufClear(&to->sendQ);
    /*
     * Ok, if the link we're dropping is a server, send a routing
     * * notice..
     */
    if (IsServer(to) && !(to->flags & FLAGS_CLOSING)) {
    char fbuf[512];

	ircsprintf(fbuf, "from %s: %s", me.name, notice);
	sendto_gnotice(fbuf, get_client_name(to, HIDEME), strerror(errtmp));
	ircsprintf(fbuf, ":%s GNOTICE :%s", me.name, notice);
	sendto_serv_butone(to, fbuf, get_client_name(to, HIDEME),
			   strerror(errtmp));
    }

    return -1;
}

/*
 * * send_message *   Internal utility which delivers one message
 * buffer to the *      socket. Takes care of the error handling and
 * buffering, if *      needed.
 */
int send_message(aClient *to, char *msg, int len)
{
    static int SQinK;

#ifdef DUMP_DEBUG
    fprintf(dumpfp, "-> %s: %s" \ n, (to->name ? to->name : "*"), msg);
#endif

    if (to->from)
	to = to->from;		/* shouldn't be necessary */

    if (IsServer(to) || IsNegoServer(to)) {
	if (len > 510) {
	    msg[511] = '\n';
	    msg[512] = '\0';
	    len = 512;
	} else {
	    msg[len] = '\n';
	    msg[len + 1] = '\0';
	    len++;
	}
    } else {
	if (len > 509) {
	    msg[510] = '\r';
	    msg[511] = '\n';
	    msg[512] = '\0';
	    len = 512;
	} else {
	    msg[len] = '\r';
	    msg[len + 1] = '\n';
	    msg[len + 2] = '\0';
	    len += 2;
	}
    }

    if (IsMe(to)) {
	sendto_ops("Trying to send to myself! [%s]", msg);
	return 0;
    }

    if (IsDead(to))
	return 0;

    if (DBufLength(&to->sendQ) > to->sendqlen) {
	/*
	 * this would be a duplicate notice, but it contains some useful information that
	 * would be spamming the rest of the network. Kept in. - lucas 
	 */
	if (IsServer(to))
	    sendto_ops("Max SendQ limit exceeded for %s: %d > %d",
		       get_client_name(to, HIDEME), DBufLength(&to->sendQ),
		       get_sendq(to));
	to->flags |= FLAGS_SENDQEX;
	return dead_link(to, "Max Sendq exceeded for %s, closing link", 0);
    }

    if (ZipOut(to)) {
    int ldata = (to->flags & FLAGS_BURST);

	msg = zip_output(to->serv->zip_out, msg, &len, 0, &ldata);
	if (len == -1) {
	    sendto_realops("Zipout error for %s: (%d) %s\n", to->name, ldata,
			   msg);
	    return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
	}

	if (len == 0)
	    return 0;
    }
#ifdef HAVE_ENCRYPTION_ON
    if (IsRC4OUT(to)) {
	/*
	 * don't destroy the data in 'msg' 
	 */
	rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
	msg = rc4buf;
    }
#endif

    if (dbuf_put(&to->sendQ, msg, len) < 0)
	return dead_link(to, "Buffer allocation error for %s, closing link",
			 IRCERR_BUFALLOC);
    /*
     * * Update statistics. The following is slightly incorrect *
     * because it counts messages even if queued, but bytes * only
     * really sent. Queued bytes get updated in SendQueued.
     */
    to->sendM += 1;
    me.sendM += 1;
    if (to->acpt != &me)
	to->acpt->sendM += 1;
    /*
     * * This little bit is to stop the sendQ from growing too large
     * when * there is no need for it to. Thus we call send_queued()
     * every time * 2k has been added to the queue since the last
     * non-fatal write. * Also stops us from deliberately building a
     * large sendQ and then * trying to flood that link with data
     * (possible during the net * relinking done by servers with a large
     * load).
     */
    /*
     * Well, let's try every 4k for clients, and immediately for servers
     * -Taner
     */

#ifdef ALWAYS_SEND_DURING_SPLIT
    if (currently_processing_netsplit && !(to->flags & FLAGS_BLOCKED)) {
	send_queued(to);
	return 0;
    }
#endif

    SQinK = (DBufLength(&to->sendQ) >> 10);
    if (IsServer(to)) {
	if (SQinK > to->lastsq)
	    send_queued(to);
    } else {
	if (SQinK > (to->lastsq + 4))
	    send_queued(to);
    }
    return 0;
}

/*
 * * send_queued *    This function is called from the main
 * select-loop (or whatever) *  when there is a chance the some output
 * would be possible. This *    attempts to empty the send queue as far
 * as possible...
 */
int send_queued(aClient *to)
{
    char *msg;
    int len, rlen;
    int more_data = 0;		/* the hybrid approach.. */

    /*
     * * Once socket is marked dead, we cannot start writing to it, *
     * even if the error is removed...
     */
    if (IsDead(to)) {
	/*
	 * * Actually, we should *NEVER* get here--something is * not
	 * working correct if send_queued is called for a * dead
	 * socket... --msa
	 */
	return -1;
    }

    if (ZipOut(to) && zip_is_data_out(to->serv->zip_out)) {
	if (DBufLength(&to->sendQ))
	    more_data = 1;
	else {
    int ldata = (to->flags & FLAGS_BURST);

	    msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);
	    if (len == -1) {
		sendto_realops("Zipout error for %s: (%d) %s\n", to->name,
			       ldata, msg);
		return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
	    }
#ifdef HAVE_ENCRYPTION_ON
	    if (IsRC4OUT(to))
		rc4_process_stream(to->serv->rc4_out, msg, len);
#endif
	    /*
	     * silently stick this on the sendq... 
	     */
	    if (!dbuf_put(&to->sendQ, msg, len))
		return dead_link(to, "Buffer allocation error for %s",
				 IRCERR_BUFALLOC);
	}
    }

    while (DBufLength(&to->sendQ) > 0) {
	msg = dbuf_map(&to->sendQ, (size_t *) & len);
	/*
	 * Returns always len > 0 
	 */
	if ((rlen = deliver_it(to, msg, len)) < 0)
	    return dead_link(to, "Write error to %s, closing link (%s)", errno);
	(void) dbuf_delete(&to->sendQ, rlen);
	to->lastsq = (DBufLength(&to->sendQ) >> 10);
	if (rlen < len)
	    /*
	     * ..or should I continue until rlen==0? 
	     */
	    /*
	     * no... rlen==0 means the send returned EWOULDBLOCK... 
	     */
	    break;

	if (more_data && DBufLength(&to->sendQ) == 0) {
    int ldata = (to->flags & FLAGS_BURST);

	    more_data = 0;

	    msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);
	    if (len == -1) {
		sendto_realops("Zipout error for %s: (%d) %s\n", to->name,
			       ldata, msg);
		return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
	    }
#ifdef HAVE_ENCRYPTION_ON
	    if (IsRC4OUT(to))
		rc4_process_stream(to->serv->rc4_out, msg, len);
#endif
	    /*
	     * silently stick this on the sendq... 
	     */
	    if (!dbuf_put(&to->sendQ, msg, len))
		return dead_link(to, "Buffer allocation error for %s",
				 IRCERR_BUFALLOC);
	}
    }

    if ((to->flags & FLAGS_SOBSENT) && IsBurst(to) && DBufLength(&to->sendQ) < 20480) {	/* 20k 
											 */
	if (!(to->flags & FLAGS_BURST)) {
	    to->flags &= (~FLAGS_SOBSENT);
	    sendto_one(to, "BURST %d", DBufLength(&to->sendQ));
	    if (!(to->flags & FLAGS_EOBRECV)) {	/* hey we're the last to
						 * synch.. */
#ifdef HTM_LOCK_ON_NETBURST
		HTMLOCK = NO;
#endif
	    }
	}
    }

    return (IsDead(to)) ? -1 : 0;
}

/*
 * prefix_buffer
 * *
 * * take varargs and dump prefixed message into a buffer
 * * remote: 1 if client is remote, 0 if local
 * * from: the client sending the message
 * * prefix: the prefix as specified (parv[0] usually)
 * * buffer: the buffer to dump this into (NO BOUNDS CHECKING!)
 * * pattern: varargs pattern
 * * vl: varargs variable list with one arg taken already
 */
int
prefix_buffer(int remote, aClient *from, char *prefix, char *buffer,
	      char *pattern, va_list vl)
{
    char *p;			/* temp pointer */
    int msglen;			/* the length of the message we end up with */
    int sidx = 1;		/* start at offset 1 */

    *buffer = ':';

    if (!remote && IsPerson(from)) {
    int flag = 0;
    anUser *user = from->user;

	for (p = from->name; *p; p++)
	    buffer[sidx++] = *p;

	if (user) {
	    if (*user->username) {
		buffer[sidx++] = '!';
		for (p = user->username; *p; p++)
		    buffer[sidx++] = *p;
	    }
	    if (*user->host && !MyConnect(from)) {
		buffer[sidx++] = '@';
#ifdef USE_FAKEHOST
		for (p = IsFake(from) ? from->user->fakehost : from->user->host;
		     *p; p++)
#else
		for (p = from->user->host; *p; p++)
#endif
		    buffer[sidx++] = *p;
		flag = 1;
	    }
	}

	if (!flag && MyConnect(from) && *user->host) {
	    buffer[sidx++] = '@';
#ifdef USE_FAKEHOST
	    for (p = IsFake(from) ? from->user->fakehost : from->user->host; *p;
		 p++)
#else
	    for (p = from->user->host; *p; p++)
#endif
		buffer[sidx++] = *p;
	}
    } else {
	for (p = prefix; *p; p++)
	    buffer[sidx++] = *p;
    }

    msglen = ircvsprintf(&buffer[sidx], pattern + 3, vl);
    msglen += sidx;

    return msglen;
}

int check_fake_direction(aClient *from, aClient *to)
{
    if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) {
	if (IsServer(from)) {
	    sendto_ops
		("Message to %s[%s] dropped from %s (Fake Direction)",
		 to->name, to->from->name, from->name);
	    return -1;
	}

	sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name,
		   to->user->username, to->user->host, from->name,
		   from->user->username, from->user->host, to->from->name);
	sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)",
			   me.name, to->name, me.name, to->name,
			   to->user->username, to->user->host, to->from->name);

	to->flags |= FLAGS_KILLED;
	exit_client(NULL, to, &me, "Ghosted client");

	if (IsPerson(from))
	    sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name,
		       from->name, to->name, to->user->username, to->user->host,
		       to->from);
	return -1;
    }

    return 0;
}

void sendto_fdlist(fdlist * listp, char *pattern, ...)
{
    int len, j, fd;
    va_list vl;

    va_start(vl, pattern);
    len = ircvsprintf(sendbuf, pattern, vl);

    for (fd = listp->entry[j = 1]; j <= listp->last_entry;
	 fd = listp->entry[++j])
	send_message(local[fd], sendbuf, len);
    va_end(vl);
}

void vsendto_fdlist(fdlist * listp, char *pattern, va_list vl)
{
    int len, j, fd;

    len = ircvsprintf(sendbuf, pattern, vl);

    for (fd = listp->entry[j = 1]; j <= listp->last_entry;
	 fd = listp->entry[++j])
	send_message(local[fd], sendbuf, len);
}

/*******************************************
 * Flushing functions (empty queues)
 *******************************************/

/*
 * flush_connections
 * Empty only buffers for clients without FLAGS_BLOCKED
 * dump_connections 
 * Unintelligently try to empty all buffers.
 */
void flush_connections(int fd)
{
    int i;
    aClient *cptr;

    if (fd == me.fd) {
	for (i = highest_fd; i >= 0; i--)
	    if ((cptr = local[i]) && !(cptr->flags & FLAGS_BLOCKED)
		&& DBufLength(&cptr->sendQ) > 0)
		send_queued(cptr);
    } else if (fd >= 0 && (cptr = local[fd])
	       && !(cptr->flags & FLAGS_BLOCKED)
	       && DBufLength(&cptr->sendQ) > 0)
	send_queued(cptr);
}

void dump_connections(int fd)
{
    int i;
    aClient *cptr;

    if (fd == me.fd) {
	for (i = highest_fd; i >= 0; i--)
	    if ((cptr = local[i]) && DBufLength(&cptr->sendQ) > 0)
		send_queued(cptr);
    } else if (fd >= 0 && (cptr = local[fd])
	       && DBufLength(&cptr->sendQ) > 0)
	send_queued(cptr);
}

/*
 * flush an fdlist intelligently 
 */
void flush_fdlist_connections(fdlist * listp)
{
int i, fd;
aClient *cptr;

    for (fd = listp->entry[i = 1]; i <= listp->last_entry;
	 fd = listp->entry[++i])
	if ((cptr = local[fd]) && !(cptr->flags & FLAGS_BLOCKED)
	    && DBufLength(&cptr->sendQ) > 0)
	    send_queued(cptr);
}
