/***********************************************************************
 *   IRC - Internet Relay Chat, src/send_chan.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: send_chan.c,v 1.13 2001/12/05 18:31:50 tr-ircd Exp $ 
 */

#include "struct.h"
#include "channel.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include "numeric.h"
#include "dh.h"
#include "zlink.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendbuf[2048];
static char remotebuf[2048];

static int sentalong[MAXCONNECTIONS];
static int sent_serial;

/*
 * This routine increments our serial number so it will
 * * be unique from anything in sentalong, no need for a memset
 * * except for every MAXINT calls - lucas
 */

/*
 * This should work on any OS where an int is 32 bit, I hope.. 
 */

#define HIGHEST_SERIAL INT_MAX

#define INC_SERIAL if(sent_serial == HIGHEST_SERIAL) \
   { memset(sentalong, 0, sizeof(sentalong)); sent_serial = 0; } \
   sent_serial++;

void
sendto_channel_butone(aClient *one, int cflags, aClient *from, aChannel *chptr,
		      char *pattern, ...)
{
    chanMember *cm;
    aClient *acptr;
    int i;
    int didlocal = 0, didremote = 0;
    va_list vl;
    char *pfix;

    va_start(vl, pattern);

    pfix = va_arg(vl, char *);

    INC_SERIAL for (cm = chptr->members; cm; cm = cm->next) {
	acptr = cm->cptr;
	if ((acptr->from == one)
	    || (cflags && !((cm->flags & cflags) == cflags)))
	    continue;		/* skip the ONE and those who do not have these 
				 * flags */
	i = acptr->from->fd;
	if (IsAnon(acptr))
	    continue;
	if (MyClient(acptr)) {
	    if (!didlocal)
		didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);

	    if (check_fake_direction(from, acptr))
		continue;
	    send_message(acptr, sendbuf, didlocal);
	    sentalong[i] = sent_serial;

	} else {
	    /*
	     * Now check whether a message has been sent to this remote
	     * link already
	     */
	    if (!didremote)
		didremote =
		    prefix_buffer(1, from, pfix, remotebuf, pattern, vl);

	    if (check_fake_direction(from, acptr))
		continue;

	    if (sentalong[i] != sent_serial) {
		send_message(acptr, remotebuf, didremote);
		sentalong[i] = sent_serial;
	    }
	}
    }

    va_end(vl);
    return;
}

/*
 * sendto_common_channels()
 * 
 * Sends a message to all people (inclusing user) on local server who are
 * in same channel with user.
 */
void sendto_common_channels(aClient *from, char *pattern, ...)
{
    Link *channels;
    chanMember *users;
    aClient *cptr;
    va_list vl;
    char *pfix;
    int msglen = 0;

    va_start(vl, pattern);

    pfix = va_arg(vl, char *);

    INC_SERIAL if (from->fd >= 0)
	  sentalong[from->fd] = sent_serial;

    if (from->user) {
	for (channels = from->user->channel; channels;
	     channels = channels->next) {

	    if (IsChanAnon(channels->value.chptr))
		continue;

	    for (users = channels->value.chptr->members; users;
		 users = users->next) {
		cptr = users->cptr;

		if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial)
		    continue;
		if (IsAnon(cptr))
		    continue;

		sentalong[cptr->fd] = sent_serial;
		if (!msglen)
		    msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
		if (check_fake_direction(from, cptr))
		    continue;
		send_message(cptr, sendbuf, msglen);
	    }
	}
    }

    if (MyConnect(from)) {
	if (!msglen)
	    msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
	send_message(from, sendbuf, msglen);
    }

    va_end(vl);
    return;

}

/*
 * send_quit_to_common_channels()
 * 
 * Sends a message to all people (inclusing user) on local server who are
 * in same channel with user if the user can send to this channel.
 */
void send_quit_to_common_channels(aClient *from, char *reason)
{
    Link *channels;
    chanMember *users;
    aClient *cptr;
    int msglen;

    INC_SERIAL if (from->fd >= 0)
	sentalong[from->fd] = sent_serial;
    for (channels = from->user->channel; channels; channels = channels->next) {
	if (!can_send(from, channels->value.chptr, reason)) {
	    msglen =
		sprintf(sendbuf, ":%s!%s@%s QUIT :%s", from->name,
			from->user->username,
#ifdef USE_FAKEHOST
			IsFake(from) ? from->user->fakehost : from->user->host,
#else
			from->user->host,
#endif
			reason);
	    for (users = channels->value.chptr->members; users;
		 users = users->next) {
		cptr = users->cptr;

		if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial)
		    continue;
		if (IsAnon(cptr))
		    continue;
		sentalong[cptr->fd] = sent_serial;
		if (check_fake_direction(from, cptr))
		    continue;
		send_message(cptr, sendbuf, msglen);
	    }
	} else if (!IsChanAnon(channels->value.chptr)) {
	    msglen =
		sprintf(sendbuf, ":%s!%s@%s PART %s", from->name,
			from->user->username,
#ifdef USE_FAKEHOST
			IsFake(from) ? from->user->fakehost : from->user->host,
#else
			from->user->host,
#endif
			channels->value.chptr->chname);
	    for (users = channels->value.chptr->members; users;
		 users = users->next) {
		cptr = users->cptr;

		if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial)
		    continue;
		if (IsAnon(cptr))
		    continue;
		sentalong[cptr->fd] = sent_serial;
		if (check_fake_direction(from, cptr))
		    continue;
		send_message(cptr, sendbuf, msglen);
	    }
	}
    }
    return;
}

/*
 * sendto_channel_butserv
 * 
 * Send a message to all members of a channel that are connected to this
 * server.
 */
void sendto_channel_butserv(aChannel *chptr, aClient *from, char *pattern, ...)
{
    chanMember *cm;
    aClient *acptr;
    va_list vl;
    int didlocal = 0;
    char *pfix;

    if (MyClient(from) && !IsULine(from)) {
	va_start(vl, pattern);
	vsendto_prefix_one(from, from, pattern, vl);
	va_end(vl);
	if (IsChanAnon(chptr) && IsLocalChan(chptr->chname))
	    return;
    }

    if (IsChanAnon(chptr) && IsGlobalChan(chptr->chname))
	return;

    va_start(vl, pattern);
    pfix = va_arg(vl, char *);

    for (cm = chptr->members; cm; cm = cm->next) {
	if (MyConnect(acptr = cm->cptr) && (acptr != from)
	    && !IsAnon(acptr)) {
	    if (!didlocal)
		didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);

	    if (check_fake_direction(from, acptr))
		continue;

	    send_message(acptr, sendbuf, didlocal);

	    /*
	     * vsendto_prefix_one(acptr, from, pattern, vl); -TimeMr14C: see above ;-) 
	     */
	}
    }
    va_end(vl);
    return;
}
